Package: Matching
Version: 4.10-15
Date: 2024-10-14
Title: Multivariate and Propensity Score Matching with Balance
        Optimization
Authors@R: c(
    person("Jasjeet", "Singh Sekhon", role = c("aut","cre"), email = "jas.sekhon@yale.edu"),
    person("Theo", "Saarinen", role = c("aut"), email = "theo_s@berkeley.edu")
    )        
Maintainer: Jasjeet Singh Sekhon <jas.sekhon@yale.edu>
Description: Provides functions for multivariate and propensity score matching 
             and for finding optimal balance based on a genetic search algorithm. 
             A variety of univariate and multivariate metrics to
             determine if balance has been obtained are also provided. For
             details, see the paper by Jasjeet Sekhon 
             (2007, <doi:10.18637/jss.v042.i07>).
Depends: R (>= 2.6.0), MASS (>= 7.2-1), graphics, grDevices, stats
Suggests: parallel, rgenoud (>= 2.12), testthat
License: GPL-3
URL: https://github.com/JasjeetSekhon/Matching
NeedsCompilation: yes
RoxygenNote: 7.2.1
Packaged: 2024-10-14 13:11:12 UTC; theosaa
Author: Jasjeet Singh Sekhon [aut, cre],
  Theo Saarinen [aut]
Repository: CRAN
Date/Publication: 2024-10-14 14:20:02 UTC
Built: R 4.6.0; x86_64-apple-darwin20; 2026-04-26 14:32:33 UTC; unix
Archs: Matching.so.dSYM

import(MASS)
importFrom("stats", "approx", "complete.cases", "ecdf", "ks.test",
           "median", "model.frame", "model.matrix", "model.response",
           "na.pass", "pchisq", "pnorm", "pt", "sd", "terms", "var")
importFrom("utils", "packageDescription")
useDynLib(Matching, .registration = TRUE)
export(balanceUV, ks.boot, Match, Matchby, MatchBalance, qqstats, summary.ks.boot, summary.Match, GenMatch, print.summary.Match, print.summary.ks.boot, print.summary.Matchby)
S3method(summary, balanceUV)
S3method(summary, ks.boot)
S3method(summary, Match)
S3method(summary, Matchby)

S3method(print, summary.Match)
S3method(print, summary.ks.boot)
S3method(print, summary.Matchby)

[![R-CMD-check](https://github.com/JasjeetSekhon/Matching/actions/workflows/check-noncontainerized.yaml/badge.svg)](https://github.com/JasjeetSekhon/Matching/actions/workflows/check-noncontainerized.yaml)

## Matching: Multivariate and Propensity Score Matching Software for Causal Inference

Jasjeet S. Sekhon

## Introduction

This website is for the distribution of "Matching" which is a R package for estimating causal effects by multivariate and propensity score matching. The package provides functions for multivariate and propensity score matching and for finding optimal balance based on a genetic search algorithm. A variety of univariate and multivariate tests to determine if balance has been obtained are also provided. These tests can also be used to determine if an experiment or quasi-experiment is balanced on baseline covariates. 

For an introduction to the package with documentation and examples,
please see "Multivariate and Propensity Score Matching Software with
Automated Balance Optimization: The Matching package for R." Journal
of Statistical Software, 42(7): 1-52. 2011.

## How to install

A version is on CRAN. The latest development version can be installed directly from Github
using [devtools](https://github.com/r-lib/devtools). 

```R
if (!require("devtools")) install.packages("devtools")
devtools::install_github("JasjeetSekhon/Matching")
```

The package contains compiled code, and you must have a development
environment to install the development version. (Use
`devtools::has_devel()` to check whether you do.) If no development
environment exists, Windows users download and install
[Rtools](https://cran.r-project.org/bin/windows/Rtools/) and macOS
users download and install
[Xcode].

#
# Gerber, Alan S. and Donald P. Green. 2000. "The Effects of Canvassing, Telephone Calls, and
# Direct Mail on Voter Turnout: A Field Experiment." American Political Science Review 94: 653-663.
#
# Imai, Kosuke. 2005. "Do Get-Out-The-Vote Calls Reduce Turnout? The Importance of 
# Statistical Methods for Field Experiments".  American Political Science Review 99: 283-300.
#

set.seed(10391)
data(GerberGreenImai)

#replication of Imai's propensity score matching model
pscore.glm<-glm(PHN.C1 ~ PERSONS + VOTE96.1 + NEW + MAJORPTY + AGE +
                WARD + PERSONS:VOTE96.1 + PERSONS:NEW + AGE2, 
                family=binomial(logit), data=GerberGreenImai)

D<-GerberGreenImai$PHN.C1 #treatment phone calls
Y<-GerberGreenImai$VOTED98 #outcome, turnout

cat("\nTHIS MODEL FAILS TO BALANCE AGE\n")
X  <- fitted(pscore.glm)

#propensity score matching estimator
r1  <- Match(Y=Y, Tr=D, X=X, M=3)
summary(r1)

#check for balance before and after matching
mb1  <- MatchBalance(PHN.C1 ~ AGE + AGE2 + PERSONS + VOTE96.1 + NEW + MAJORPTY +
                     WARD + I(PERSONS*VOTE96.1) + I(PERSONS*NEW), match.out=r1,
                     data=GerberGreenImai)

# Replication of Guido Imbens lalonde_exper_04feb2.m file
# See http://elsa.berkeley.edu/~imbens/estimators.shtml
#
# Note that the implications of the 'exact' options differ between the
# two programs

data(lalonde)

X  <- lalonde$age
Z  <- X;             
V  <- lalonde$educ;
Y  <- lalonde$re78/1000;
T  <- lalonde$treat;
w.educ=exp((lalonde$educ-10.1)/2);

res  <- matrix(nrow=1,ncol=3)

rr  <- Match(Y=Y,Tr=T,X=X,Z=Z,V=V,estimand="ATE",M=1,BiasAdj=FALSE,Weight=1,Var.calc=0,
             sample=TRUE);
summary(rr)

res[1,]  <- cbind(1,rr$est,rr$se)


X  <- cbind(lalonde$age, lalonde$educ, lalonde$re74, lalonde$re75)
rr  <- Match(Y=Y,Tr=T,X=X,Z=Z,V=V,estimand="ATE",M=1,BiasAdj=FALSE,Weight=1,Var.calc=0,
             sample=TRUE);
summary(rr)

res  <- rbind(res,cbind(2,rr$est,rr$se))

rr  <- Match(Y=Y,Tr=T,X=X,Z=Z,V=V,estimand="ATE",M=3,BiasAdj=FALSE,Weight=1,Var.calc=0,
             sample=TRUE);
summary(rr)
res  <- rbind(res,cbind(4,rr$est,rr$se))

rr  <- Match(Y=Y,Tr=T,X=X,Z=Z,V=V,estimand="ATT",M=1,BiasAdj=FALSE,Weight=1,Var.calc=0,
             sample=TRUE);
summary(rr)
res  <- rbind(res,cbind(5,rr$est,rr$se))

rr  <- Match(Y=Y,Tr=T,X=X,Z=Z,V=V,estimand="ATC",M=1,BiasAdj=FALSE,Weight=1,Var.calc=0,
             sample=TRUE);
summary(rr)
res  <- rbind(res,cbind(6,rr$est,rr$se))

rr  <- Match(Y=Y,Tr=T,X=X,Z=Z,V=V,estimand="ATE",M=1,BiasAdj=FALSE,Weight=2,Var.calc=0,
             sample=TRUE);
summary(rr)
res  <- rbind(res,cbind(7,rr$est,rr$se))

rr  <- Match(Y=Y,Tr=T,X=X,Z=Z,V=V,estimand="ATE",M=1,BiasAdj=FALSE,Weight=3,Var.calc=0,
             Weight.matrix=diag(4), sample=TRUE);
summary(rr)
res  <- rbind(res,cbind(8,rr$est,rr$se))


rr  <- Match(Y=Y,Tr=T,X=X,Z=X,V=V,estimand="ATE",M=1,BiasAdj=TRUE,Weight=1,Var.calc=0,
             sample=TRUE);
summary(rr)
res  <- rbind(res,cbind(9,rr$est,rr$se))

Z  <- cbind(lalonde$married, lalonde$age)
rr  <- Match(Y=Y,Tr=T,X=X,Z=Z,V=V,estimand="ATE",M=1,BiasAdj=TRUE,Weight=1,Var.calc=0,sample=TRUE);
summary(rr)
res  <- rbind(res,cbind(10,rr$est,rr$se))

V  <- lalonde$age
rr  <- Match(Y=Y,Tr=T,X=X,Z=Z,V=V,estimand="ATE",M=1,BiasAdj=FALSE,Weight=1,Var.calc=0,exact=TRUE,
             sample=TRUE);
summary(rr)
res  <- rbind(res,cbind(11,rr$est,rr$se))

V  <- cbind(lalonde$married, lalonde$u74)
rr  <- Match(Y=Y,Tr=T,X=X,Z=Z,V=V,estimand="ATE",M=1,BiasAdj=FALSE,Weight=1,Var.calc=0,exact=TRUE,
             sample=TRUE);
summary(rr)
res  <- rbind(res,cbind(12,rr$est,rr$se))

rr  <- Match(Y=Y,Tr=T,X=X,Z=Z,V=V,estimand="ATE",M=1,BiasAdj=FALSE,Weight=1,Var.calc=0,sample=FALSE);
summary(rr)
res  <- rbind(res,cbind(13,rr$est,rr$se))

rr  <- Match(Y=Y,Tr=T,X=X,Z=Z,V=V,estimand="ATE",M=1,BiasAdj=FALSE,Weight=1,Var.calc=3,sample=TRUE);
summary(rr)
res  <- rbind(res,cbind(14,rr$est,rr$se))

rr  <- Match(Y=Y,Tr=T,X=X,Z=Z,V=V,estimand="ATE",M=1,BiasAdj=FALSE,Weight=1,Var.calc=0,
             weights=w.educ,sample=TRUE);
summary(rr)
res  <- rbind(res,cbind(15,rr$est,rr$se))


V  <- lalonde$age
Z  <- cbind(lalonde$married, lalonde$age)
X  <- cbind(lalonde$age, lalonde$educ, lalonde$re74, lalonde$re75)
weight  <- w.educ
Weight.matrix  <- diag(4)

rr  <- Match(Y=Y,Tr=T,X=X,Z=Z,V=V,
             sample=FALSE, M=3, estimand="ATT", BiasAdj=TRUE, Weight=3, exact=TRUE,Var.calc=3,
             weights=w.educ, Weight.matrix=Weight.matrix);
summary(rr)
res  <- rbind(res,cbind(75,rr$est,rr$se))


V  <- lalonde$married;
Z  <- cbind(lalonde$age, lalonde$re75);
X  <- cbind(lalonde$age, lalonde$educ, lalonde$re74);

rr  <- Match(Y=Y,Tr=T,X=X,Z=Z,V=V,
             sample=TRUE, M=3, estimand="ATE", BiasAdj=TRUE, Weight=2, exact=TRUE,Var.calc=0,
             weights=w.educ);
summary(rr)
res  <- rbind(res,cbind(76,rr$est,rr$se))

cat("\nResults:\n")
print(res)

# Replication of Dehejia and Wahba psid3 model.  

# Dehejia, Rajeev and Sadek Wahba. 1999.``Causal Effects in
# Non-Experimental Studies: Re-Evaluating the # Evaluation of Training
# Programs.''Journal of the American Statistical Association 94 (448):
# 1053-1062.  

set.seed(10391)
data(lalonde)

#
# Estimate the propensity model
#
glm1  <- glm(treat~age + I(age^2) + educ + I(educ^2) + black +
             hisp + married + nodegr + re74  + I(re74^2) + re75 + I(re75^2) +
             u74 + u75, family=binomial, data=lalonde)

#
#save data objects
#
X  <- glm1$fitted
Y  <- lalonde$re78
Tr  <- lalonde$treat

#
# one-to-one matching with replacement (the "M=1" option).
# Estimating the treatment effect on the treated (the "estimand" option which defaults ATT).
#
rr  <- Match(Y=Y,Tr=Tr,X=X,M=1);
summary(rr)

#
# Let's check for balance
#
mb  <- MatchBalance(treat~age + I(age^2) + educ + I(educ^2) + black +
                    hisp + married + nodegr + re74  + I(re74^2) + re75 + I(re75^2) +
                    u74 + u75, data=lalonde, match.out=rr, nboots=100)

# Jasjeet S. Sekhon <sekhon@berkeley.edu>
# HTTP://sekhon.berkeley.edu/
# UC Berkeley

# Match(): function to estimate treatments using a matching estimator.
# Currently only the ability to estimate average treatment effects
# using the approach of Abadie and Imbens is implemented.  In the
# future, quantile treatment effects will be implemented along with
# the ability to use robust estimation when estimating the propensity
# score. MatchBalance(), and balanceUV() test for balance.

Match  <- function(Y=NULL,Tr,X,Z=X,V=rep(1,length(Y)), estimand="ATT", M=1,
                   BiasAdjust=FALSE,exact=NULL,caliper=NULL, replace=TRUE, ties=TRUE,
                   CommonSupport=FALSE,Weight=1,Weight.matrix=NULL, weights=NULL,
                   Var.calc=0, sample=FALSE, restrict=NULL, match.out=NULL,
                   distance.tolerance=0.00001, tolerance=sqrt(.Machine$double.eps),
                   version="standard")
  {

    BiasAdj  <- as.double(BiasAdjust)
    sample  <- as.double(sample)

    if ( (BiasAdj != 0) & (BiasAdj != 1) )
      {
        warning("User set 'BiasAdjust' to a non-logical value.  Resetting to the default which is FALSE.")        
        BiasAdj <- 0
      }
    
    #we don't need to use a Y
    if (is.null(Y))
      {
        Y = rep(0, length(Tr))
        version <- "fast"

        if(BiasAdj)
          {
            warning("'BiasAdjust' set to FALSE because Y is NULL")
            BiasAdj <- FALSE
          }
      }
    
    Y  <- as.double(Y)
    Tr <- as.double(Tr)
    X  <- as.matrix(X)
    Z  <- as.matrix(Z)
    V  <- as.matrix(V)

    orig.nobs  <- length(Y)
    nobs  <- orig.nobs
    xvars <- ncol(X)    
    orig.tr.nobs <- length(Tr)
    if (orig.tr.nobs != orig.nobs)
      {
        stop("length(Y) != length(Tr)")
      }
    if( orig.tr.nobs != nrow(X))
      {
        stop("length(Tr) != nrow(X)")
      }    
    
    if( orig.nobs != nrow(X))
      {
        stop("length(Y) != nrow(X)")
      }
    if( orig.nobs != nrow(V))
      {
        stop("length(Y) != nrow(V)")
      }
    if( orig.nobs != nrow(Z))
      {
        stop("length(Y) != nrow(Z)")
      }

    if (is.null(weights))
      {
        weights <- rep(1,length(Y))
        weights.flag <- FALSE
      } else {
        weights.flag <- TRUE
        weights <- as.double(weights)
        if( orig.tr.nobs != length(weights))
          {
            stop("length(Tr) != length(weights)")
          }        
      }

    isna  <- sum(is.na(Y)) + sum(is.na(Tr)) + sum(is.na(X)) + sum(is.na(weights)) + sum(is.na(Z)) + sum(is.na(V))
    if (isna!=0)
      {
        stop("Match(): input includes NAs")
        return(invisible(NULL))
      }

    if (sum(Tr !=1 & Tr !=0) > 0) {
      stop("Treatment indicator ('Tr') must be a logical variable---i.e., TRUE (1) or FALSE (0)")
    }
    if (var(Tr)==0) {
      stop("Treatment indicator ('Tr') must contain both treatment and control observations")
    }
    if (distance.tolerance < 0)
      {
        warning("User set 'distance.tolerance' to less than 0.  Resetting to the default which is 0.00001.")
        distance.tolerance <- 0.00001
      }

    #CommonSupport
    if (CommonSupport !=1 & CommonSupport !=0) {
      stop("'CommonSupport' must be a logical variable---i.e., TRUE (1) or FALSE (0)")
    }    
    if(CommonSupport==TRUE)
      {
        tr.min <- min(X[Tr==1,1])
        tr.max <- max(X[Tr==1,1])

        co.min <- min(X[Tr==0,1])
        co.max <- max(X[Tr==0,1])

        if(tr.min >= co.min)
          {
            indx1 <- X[,1] < (tr.min-distance.tolerance)
          } else {
            indx1 <- X[,1] < (co.min-distance.tolerance)            
          }

        if(co.max <= tr.max)
          {        
            indx2 <- X[,1] > (co.max+distance.tolerance)
          } else {
            indx2 <- X[,1] > (tr.max+distance.tolerance)            
          }

        indx3 <- indx1==0 & indx2==0
        Y  <- as.double(Y[indx3])
        Tr <- as.double(Tr[indx3])
        X  <- as.matrix(X[indx3,])
        Z  <- as.matrix(Z[indx3,])
        V  <- as.matrix(V[indx3,])
        weights <- as.double(weights[indx3])

        #let's recalculate these for common support
        orig.nobs  <- length(Y)
        nobs  <- orig.nobs
      }#end of CommonSupport

    #check additional inputs
    if (tolerance < 0)
      {
        warning("User set 'tolerance' to less than 0.  Resetting to the default which is 0.00001.")
        tolerance <- 0.00001
      }
    if (M < 1)
      {
        warning("User set 'M' to less than 1.  Resetting to the default which is 1.")
        M <- 1
      }
    if ( M != round(M) )
      {
        warning("User set 'M' to an illegal value.  Resetting to the default which is 1.")
        M <- 1        
      }
    if (Var.calc < 0)
      {
        warning("User set 'Var.calc' to less than 0.  Resetting to the default which is 0.")
        Var.calc <- 0
      }
    if ( (sample != 0) & (sample != 1) )
      {
        warning("User set 'sample' to a non-logical value.  Resetting to the default which is FALSE.")        
        sample <- 0
      }
    if (Weight != 1 & Weight != 2 & Weight != 3)
      {
        warning("User set 'Weight' to an illegal value.  Resetting to the default which is 1.")        
        Weight <- 1
      }
    if (version!="fast" & version != "standard" & version != "legacy" & version != "Matchby" & version != "MatchbyAI")
      {
        warning("User set 'version' to an illegal value.  Resetting to the default which is 'standard'.")        
        version <- "standard"
      }

    if(version=="Matchby")
      {
        version <- "fast"
        Matchby.call <- TRUE
        MatchbyAI <- FALSE
      } else if(version=="MatchbyAI")
      {
        version <- "standard"
        Matchby.call <- TRUE
        MatchbyAI <- TRUE
      } else {
        Matchby.call <- FALSE
        MatchbyAI <- FALSE
      }

    if (Var.calc !=0 & version=="fast")
      {
        warning("Var.calc cannot be estimate when version=='fast'")
        Var.calc=0
      }
    if (BiasAdj!=FALSE & version=="fast")
      {
        warning("Bias Adjustment cannot be estimated when version=='fast'")
        BiasAdj=0
      }
    if (replace!=FALSE & replace!=TRUE)
      {
        warning("'replace' must be TRUE or FALSE.  Setting to TRUE")
        replace <- TRUE
      }
    if(replace==FALSE)
      {
        ties <- FALSE        
        version="fast"
        if (version=="legacy")
          warning("'version' is set to 'fast' because replace==FALSE")
      }
    if (ties!=FALSE & ties!=TRUE)
      {
        warning("'ties' must be TRUE or FALSE.  Setting to TRUE")
        ties <- TRUE
      }
    if(ties==FALSE)
      {
        version="fast"
        if (version=="legacy")
          warning("'version' is set to 'fast' because ties==FALSE")

        if(BiasAdjust==TRUE) {
          warning("Bias Adjustment can only be estimated when ties==TRUE and replace=TRUE.  Setting BiasAdjust=FALSE")
          BiasAdjust <- FALSE
          BiasAdj  <- 0
        }
      }

    if (!is.null(match.out) & class(match.out) != "Match") {
      warning("match.out object not of class 'Match'")
      return(invisible(NULL))
    }

    ccc  <- tolerance
    cdd  <- distance.tolerance

    orig.treated.nobs  <- sum(Tr==1)
    orig.control.nobs  <- sum(Tr==0)
    orig.wnobs  <- sum(weights)
    orig.weighted.treated.nobs <- sum( weights[Tr==1] )
    orig.weighted.control.nobs <- sum( weights[Tr==0] )    
    weights.orig  <- as.matrix(weights)
    zvars <- ncol(Z);

    estimand.orig <- estimand
    if (estimand=="ATT")
      {
        estimand  <- 0

        if(BiasAdj==1 & orig.treated.nobs<zvars)
          {
            warning("Fewer treated obs than variables in 'Z': BiasAdjust set to FALSE")
            BiasAdj=0            
          }

      } else if(estimand=="ATE") {
        estimand  <- 1

        if(BiasAdj==1 & orig.nobs<zvars)
          {
            warning("Fewer obs than variables in 'Z': BiasAdjust set to FALSE")
            BiasAdj=0            
          }
      } else if(estimand=="ATC") {
        estimand  <- 2

        if(BiasAdj==1 & orig.control.nobs<zvars)
          {
            warning("Fewer control obs than variables in 'Z': BiasAdjust set to FALSE")
            BiasAdj=0            
          }

      } else {
        estimand  <- 0
        warning("User set 'estimand' to an illegal value.  Resetting to the default which is 'ATT'")
      }

    if (!is.null(Weight.matrix))
      {

        if(inherits(Weight.matrix, "GenMatch"))
          {
            Weight.matrix = Weight.matrix$Weight.matrix
          }
        
        if (Weight==2)
          {
            warning("User supplied 'Weight.matrix' is being used even though 'Weight' is not set equal to 3")
          }
        Weight  <- 3
      } else {
        Weight.matrix <- dim(X)[2]
      }

    if(Var.calc > orig.weighted.treated.nobs)
      {
        warning("'Var.calc' > the number of treated obs: 'Var.calc' reset to ",
                orig.weighted.treated.nobs, immediate.=Matchby.call)
        Var.calc <- orig.weighted.treated.nobs
      }
    if(Var.calc > orig.weighted.control.nobs)
      {
        warning("'Var.calc' > the number of control obs: 'Var.calc' reset to ",
                orig.weighted.control.nobs,  immediate.=Matchby.call)
        Var.calc <- orig.weighted.control.nobs
      }
    
    if(orig.nobs > 20000 & version!="fast" & !Matchby.call)
      {
        warning("The version='fast' option is recommended for large datasets if speed is desired.  For additional speed, you may also consider using the ties=FALSE option.", immediate.=TRUE)
      }

    #check the restrict matrix input
    if(!is.null(restrict))
      {
        if(!is.matrix(restrict))
          stop("'restrict' must be a matrix of restricted observations rows and three columns: c(i,j restriction)")

        if(ncol(restrict)!=3 )
          stop("'restrict' must be a matrix of restricted observations rows and three columns: c(i,j restriction)")
      }

    if (!is.null(exact))
      {
        exact = as.vector(exact)
        nexacts = length(exact)
        if ( (nexacts > 1) & (nexacts != xvars) )
          {
            warning("length of exact != ncol(X). Ignoring exact option")
            exact <- NULL
          } else if (nexacts==1 & (xvars > 1) ){
            exact <- rep(exact, xvars)
          }
      }

    if (!is.null(caliper))
      {
        caliper = as.vector(caliper)
        ncalipers = length(caliper)
        if ( (ncalipers > 1) & (ncalipers != xvars) )
          {
            warning("length of caliper != ncol(X). Ignoring caliper option")
            caliper <- NULL
          } else if (ncalipers==1 & (xvars > 1) ){
            caliper <- rep(caliper, xvars)
          }
      }

    if (!is.null(caliper))
      {
        ecaliper <- vector(mode="numeric", length=xvars)
        sweights  <- sum(weights.orig)
        for (i in 1:xvars)
          {
            meanX  <- sum( X[,i]*weights.orig )/sweights
            sdX  <- sqrt(sum( (X[,i]-meanX)^2 )/sweights)
            ecaliper[i]  <- caliper[i]*sdX
          }
      } else {
        ecaliper <- NULL
      }

    if (!is.null(exact))
      {
        if(is.null(caliper))
          {
            max.diff <- abs(max(X)-min(X) + tolerance * 100)
            ecaliper <- matrix(max.diff, nrow=xvars, ncol=1)
          }
        
        for (i in 1:xvars)
          {
            if (exact[i])
              ecaliper[i] <- tolerance;
          }
      }

    if(replace==FALSE)
      {
        #replace==FALE, needs enough observation
        #ATT
        orig.weighted.control.nobs <- sum(weights[Tr!=1])
        if(estimand==0)
          {
            if (orig.weighted.treated.nobs > orig.weighted.control.nobs)
              {
                warning("replace==FALSE, but there are more (weighted) treated obs than control obs.  Some treated obs will not be matched.  You may want to estimate ATC instead.")
              }
          } else if(estimand==1) 
            {
              #ATE
              if (orig.weighted.treated.nobs > orig.weighted.control.nobs)
                {
                  warning("replace==FALSE, but there are more (weighted) treated obs than control obs.  Some treated obs will not be matched.  You may want to estimate ATC instead.")
                }              
              if (orig.weighted.treated.nobs < orig.weighted.control.nobs)
                {
                  warning("replace==FALSE, but there are more (weighted) control obs than treated obs.  Some control obs will not be matched.  You may want to estimate ATT instead.")
                }
            } else 
        {
              #ATC
          if (orig.weighted.treated.nobs < orig.weighted.control.nobs)
            {
              warning("replace==FALSE, but there are more (weighted) control obs than treated obs.  Some obs will be dropped.  You may want to estimate ATC instead")
            }
        }
        
        #we need a restrict matrix if we are going to not do replacement
        if(is.null(restrict))
          {
            restrict <- t(as.matrix(c(0,0,0)))
          }
        if(version!="fast" &  version!="standard")
          {
            warning("reverting to 'standard' version because replace=FALSE")
            version="standard"
          }
      }#end of replace==FALSE

#    if(version=="fast" & is.null(ecaliper) & sum(weights==1)==orig.nobs)
    if(version=="fast" | version=="standard")
      {
        if(!is.null(match.out))
          {
            ret <- RmatchLoop(Y=Y, Tr=Tr, X=X, Z=Z, V=V, All=estimand, M=M, BiasAdj=BiasAdj,
                              Weight=Weight, Weight.matrix=Weight.matrix, Var.calc=Var.calc,
                              weight=weights, SAMPLE=sample, ccc=ccc, cdd=cdd,
                              ecaliper=ecaliper, exact=exact, caliper=caliper,
                              restrict=restrict, MatchLoopC.indx=match.out$MatchLoopC,
                              weights.flag=weights.flag,  replace=replace, ties=ties,
                              version=version, MatchbyAI=MatchbyAI)
          } else {
            ret <- RmatchLoop(Y=Y, Tr=Tr, X=X, Z=Z, V=V, All=estimand, M=M, BiasAdj=BiasAdj,
                              Weight=Weight, Weight.matrix=Weight.matrix, Var.calc=Var.calc,
                              weight=weights, SAMPLE=sample, ccc=ccc, cdd=cdd,
                              ecaliper=ecaliper, exact=exact, caliper=caliper,
                              restrict=restrict, weights.flag=weights.flag,  replace=replace, ties=ties,
                              version=version, MatchbyAI=MatchbyAI)   
          }
      } else {
        ret <- Rmatch(Y=Y, Tr=Tr, X=X, Z=Z, V=V, All=estimand, M=M, BiasAdj=BiasAdj,
                      Weight=Weight, Weight.matrix=Weight.matrix, Var.calc=Var.calc,
                      weight=weights, SAMPLE=sample, ccc=ccc, cdd=cdd,
                      ecaliper=ecaliper, restrict=restrict)
      }

    if(is.null(ret$est))
      {
        if(!Matchby.call)
          {
            if(ret$valid < 1)
              {
                if (ret$sum.caliper.drops > 0) {
                  warning("'Match' object contains no valid matches (probably because of the caliper or the exact option).") 
                } else {
                  warning("'Match' object contains no valid matches")
                }
              } else {
                if (ret$sum.caliper.drops > 0) {
                  warning("'Match' object contains only 1 valid match (probably because of the caliper or the exact option).") 
                } else {
                  warning("'Match' object contains only one valid match")
                }            
              }
          } #endof if(!Matchby.call)

        z <- NA
        class(z)  <- "Match"    
        return(z)
      }
    
    indx <-  cbind(ret$art.data[,1],  ret$art.data[,2],  ret$W)

    index.treated  <- indx[,1]
    index.control  <- indx[,2]
    weights        <- indx[,3]
    sum.caliper.drops <- ret$sum.caliper.drops

   #RESET INDEX.TREATED        
    indx  <- as.matrix(cbind(index.treated,index.control))
    if (estimand==0) {
      #"ATT"
      index.treated  <- indx[,1]
      index.control  <- indx[,2]
    } else if(estimand==1) {
      #"ATE"
      tmp.index.treated  <- indx[,1]
      tmp.index.control  <- indx[,2]
      
      tl  <- length(tmp.index.treated)
      index.treated <- vector(length=tl, mode="numeric")
      index.control <- vector(length=tl, mode="numeric")
      trt  <- Tr[tmp.index.treated]==1
      for (i in 1:tl)
        {
          if (trt[i]) {
            index.treated[i]  <- tmp.index.treated[i]
            index.control[i]  <- tmp.index.control[i]
          } else {
            index.treated[i]  <- tmp.index.control[i]
            index.control[i]  <- tmp.index.treated[i]
          }
        }
    } else if(estimand==2) {
      #"ATC"
      index.treated  <- indx[,2]
      index.control  <- indx[,1]
    }        


    mdata  <- list()
    mdata$Y  <- c(Y[index.treated],Y[index.control])
    mdata$Tr <- c(Tr[index.treated],Tr[index.control])
    mdata$X  <- rbind(X[index.treated,],X[index.control,])
    mdata$orig.weighted.treated.nobs <- orig.weighted.treated.nobs

    #naive standard errors
    mest  <- sum((Y[index.treated]-Y[index.control])*weights)/sum(weights)
    v1  <- Y[index.treated] - Y[index.control]
    varest  <- sum( ((v1-mest)^2)*weights)/(sum(weights)*sum(weights))
    se.standard  <- sqrt(varest)

    wnobs <- sum(weights)

    if(estimand==0)
      {
        #ATT
        actual.drops <- orig.weighted.treated.nobs-wnobs
      } else if (estimand==1)
        {
          #ATE
          actual.drops <- orig.wnobs-wnobs
        } else {
          #ATC
          actual.drops <- (orig.wnobs-orig.weighted.treated.nobs)-wnobs 
        }

    #What obs were dropped?
    index.dropped <-  NULL #nothing was dropped
    if (sum.caliper.drops > 0 )
      {
        if(estimand.orig=="ATT")
          {
            matched.index <- which(Tr==1)
            matched <- !(matched.index %in% index.treated)        
            
          } else if(estimand.orig=="ATC")
            {
              matched.index <- which(Tr==0)
              matched <- !(matched.index %in% index.control)
            } else if(estimand.orig=="ATE")
              {
                matched.index <- 1:length(Tr)
                matched <- !(matched.index %in% c(index.treated,index.control))
              } 
        index.dropped <- matched.index[matched]  #obs not matched
      } #end of sum.caliper.drops > 0
            
    z  <- list(est=ret$est, se=ret$se, est.noadj=mest, se.standard=se.standard,
               se.cond=ret$se.cond, 
               mdata=mdata, 
               index.treated=index.treated, index.control=index.control, index.dropped=index.dropped,
               weights=weights, orig.nobs=orig.nobs, orig.wnobs=orig.wnobs,
               orig.treated.nobs=orig.treated.nobs,
               nobs=nobs, wnobs=wnobs,
               caliper=caliper, ecaliper=ecaliper, exact=exact,
               ndrops=actual.drops, ndrops.matches=sum.caliper.drops, 
               MatchLoopC=ret$MatchLoopC, version=version,
               estimand=estimand.orig)

    if(MatchbyAI)
      {
        z$YCAUS   <- ret$YCAUS
        z$ZCAUS   <- ret$ZCAUS
        z$Kcount  <- ret$Kcount
        z$KKcount <- ret$KKcount
        z$Sigs    <- ret$Sigs
      }

    class(z)  <- "Match"    
    return(z)
  } #end of Match

summary.Match  <- function(object, ..., full=FALSE, digits=5)
  {
    if(!is.list(object)) {
      warning("'Match' object contains less than two valid matches.  Cannot proceed.")
      return(invisible(NULL))
    }
    
    if (class(object) != "Match") {
      warning("Object not of class 'Match'")
      return(invisible(NULL))
    }

    if(object$version!="fast")
      {
        cat("\n")
        cat("Estimate... ",format(object$est,digits=digits),"\n")
        cat("AI SE...... ",format(object$se,digits=digits),"\n")
        cat("T-stat..... ",format(object$est/object$se,digits=digits),"\n")
        cat("p.val...... ",format.pval((1-pnorm(abs(object$est/object$se)))*2,digits=digits),"\n")
        cat("\n")
      } else {
        cat("\n")
        cat("Estimate... ",format(object$est,digits=digits),"\n")
        cat("SE......... ",format(object$se.standard,digits=digits),"\n")
        cat("T-stat..... ",format(object$est/object$se.standard,digits=digits),"\n")
        cat("p.val...... ",format.pval((1-pnorm(abs(object$est/object$se.standard)))*2,digits=digits),"\n")
        cat("\n")        
      }

    if(full)
      {
        cat("Est noAdj.. ",format(object$est.noadj,digits=digits),"\n")        
        cat("SE......... ",format(object$se.standard,digits=digits),"\n")
        cat("T-stat..... ",format(object$est.noadj/object$se.standard,digits=digits),"\n")
        cat("p.val...... ",format.pval((1-pnorm(abs(object$est.noadj/object$se.standard)))*2,digits=digits),"\n")
        cat("\n")
      }


    if(object$orig.wnobs!=object$orig.nobs)
      cat("Original number of observations (weighted)... ", round(object$orig.wnobs, 3),"\n")
    cat("Original number of observations.............. ", object$orig.nobs,"\n")
    if(object$mdata$orig.weighted.treated.nobs!=object$orig.treated.nobs)
      cat("Original number of treated obs (weighted).... ", round(object$mdata$orig.weighted.treated.nobs, 3),"\n")
    if(object$estimand!="ATC")
      {
        cat("Original number of treated obs............... ", object$orig.treated.nobs,"\n")
      } else {
        cat("Original number of control obs............... ",
            object$orig.nobs-object$orig.treated.nobs,"\n")        
      }
    cat("Matched number of observations............... ", round(object$wnobs, 3),"\n")
    cat("Matched number of observations  (unweighted). ", length(object$index.treated),"\n")

    cat("\n")
    if(!is.null(object$exact))
      {
        cat("Number of obs dropped by 'exact' or 'caliper' ", object$ndrops.matches, "\n")
        if (object$ndrops.matches!=round(object$ndrops))
          cat("Weighted #obs dropped by 'exact' or 'caliper' ", round(object$ndrops, 3),"\n")        
        cat("\n")        
      }else if(!is.null(object$caliper))
      {
        cat("Caliper (SDs)........................................  ",object$caliper,"\n")            
        cat("Number of obs dropped by 'exact' or 'caliper' ", object$ndrops.matches, "\n")
        if (object$ndrops.matches!=round(object$ndrops))
          cat("Weighted #obs dropped by 'exact' or 'caliper' ", round(object$ndrops, 3),"\n")        
        cat("\n")
      }
    
    z <- list()
    class(z) <- "summary.Match"
    return(invisible(z))
  } #end of summary.Match

print.summary.Match <- function(x, ...)
  {
    invisible(NULL)
  }


Rmatch <- function(Y, Tr, X, Z, V, All, M, BiasAdj, Weight, Weight.matrix, Var.calc, weight,
                   SAMPLE, ccc, cdd, ecaliper=NULL, restrict=NULL)
  {
    sum.caliper.drops <- 0
    X.orig <- X

#are we using the restriction matrix?
    if(is.matrix(restrict)) {
      restrict.trigger <- TRUE
    } else {
      restrict.trigger <- FALSE
    }
    
# if SATC is to be estimated the treatment indicator is reversed  
    if (All==2) {
      Tr <- 1-Tr
    }

    

# check on the number of matches, to make sure the number is within the limits
# feasible given the number of observations in both groups.
    if (All==1)
      {
        M <- min(M,min(sum(Tr),sum(1-Tr)));        
      } else {
        M <- min(M,sum(1-Tr));
      }

# two slippage parameters that are used to determine whether distances are equal
# distances less than ccc or cdd are interpeted as zero.
# these are passed in.  ccc, cdd


# I. set up
# I.a. vector for which observations we want the average effect
# iot_t is the vector with weights in the average treatment effects
# iot_c is the vector of indicators for potential controls

    if (All==1)
      {
        iot.t <- weight;
        iot.c <- as.matrix(rep(1,length(Tr)))
      } else {
        iot.t <- Tr*weight;
        iot.c <- 1-Tr    
      }

# I.b. determine sample and covariate vector sizes
    N  <- nrow(X)
    Kx <- ncol(X)
    Kz <- ncol(Z)

# K covariates
# N observations
#    Nt <- sum(Tr)
#    Nc <- sum(1-Tr)
#    on <- as.matrix(rep(1,N))

# I.c. normalize regressors to have mean zero and unit variance.
# If the standard deviation of a variable is zero, its normalization
# leads to a variable with all zeros.
# The matrix AA enables one to transform the user supplied weight matrix 
# to take account of this transformation.  BUT THIS IS NOT USED!!
# Mu_X and Sig_X keep track of the original mean and variances
#    AA    <- diag(Kx)
    Mu.X  <- matrix(0, Kx, 1)
    Sig.X <- matrix(0, Kx, 1)

    for (k in 1:Kx)
      {
        Mu.X[k,1] <- sum(X[,k]*weight)/sum(weight)
        eps <- X[,k]-Mu.X[k,1]
        Sig.X[k,1] <- sqrt(max(ccc, sum(X[,k]*X[,k]*weight))/sum(weight)-Mu.X[k,1]^2)
        Sig.X[k,1] <- Sig.X[k,1]*sqrt(N/(N-1))
        if(Sig.X[k,1] < ccc)
          Sig.X[k,1] <- ccc
        X[,k]=eps/Sig.X[k,1]
#        AA[k,k]=Sig.X[k,1]
      } #end of k loop

#    Nv <- nrow(V)
    Mv <- ncol(V)
    Mu.V  <- matrix(0, Mv, 1)
    Sig.V <- matrix(0, Mv, 1)

    for (j in 1:Mv)
      {
        Mu.V[j,1]= ( t(V[,j])%*%weight ) /sum(weight)
#        dv <- V[,j]-Mu.V[j,1]
        sv <- sum(V[,j]*V[,j]*weight)/sum(weight)-Mu.V[j,1]^2
        if (sv > 0)
          {
            sv <- sqrt(sv)
          } else {
            sv <- 0
          }
        sv <- sv * sqrt(N/(N-1))
        Sig.V[j,1] <- sv
      } #end of j loop

# I.d. define weight matrix for metric, taking into account normalization of
# regressors.
# If the eigenvalues of the regressors are too close to zero the Mahalanobis metric
# is not used and we revert back to the default inverse of variances.
    if (Weight==1)
      {
        Weight.matrix=diag(Kx)
      } else if (Weight==2) {
        if (min (eigen( t(X)%*%X/N, only.values=TRUE)$values) > ccc)
          {
            Weight.matrix= solve(t(X)%*%X/N) 
          } else {
            Weight.matrix <- diag(Kx)
          }
      }
      # DO NOT RESCALE THE Weight.matrix!!
      #else if (Weight==3)
      #  {
      #    Weight.matrix <- AA %*% Weight.matrix %*% AA
      #  }

#    if (exact==1)
#      {
#        Weight.matrix <- cbind(Weight.matrix, matrix(0,nrow=Kx,ncol=Mv))
#        Weight.matrix <- rbind(Weight.matrix, cbind(matrix(0,nrow=Mv,ncol=Kx),
#                               1000*solve(diag(as.vector(Sig.V*Sig.V), nrow=length(Sig.V)))))
#        Weight.matrix <- as.matrix(Weight.matrix)
#        X <- cbind(X,V)
#        Mu.X  <- rbind(Mu.X, matrix(0, nrow(Mu.V), 1))
#        Sig.X <- rbind(Sig.X, matrix(1, nrow(Sig.V), 1))
#      } #end of exact

    Nx <- nrow(X)
    Kx <- ncol(X)

    if ( min(eigen(Weight.matrix, only.values=TRUE)$values) < ccc )
      Weight.matrix <- Weight.matrix + diag(Kx)*ccc

    # I.fg. initialize matrices before looping through sample
    YCAUS  <- as.matrix(rep(0, N))
    SCAUS  <- as.matrix(rep(0, N))
    XCAUS  <- matrix(0, nrow=N, ncol=Kx)
    ZCAUS  <- matrix(0, nrow=N, ncol=Kz)
    Kcount <- as.matrix(rep(0, N))
    KKcount <- as.matrix(rep(0, N))
    MMi     <- as.matrix(rep(0, N))

# II. Loop through all observations that need to be matched.
    INN <- as.matrix(1:N)
    ww <- chol(Weight.matrix) # so that ww*ww=w.m
#    TT <- as.matrix(1:N)

    # initialize some data objects
    DD <- NULL
    I  <- NULL
    IM <- NULL
    IT <- NULL
    IX <- NULL
    IZ <- NULL
    IY <- NULL
    W  <- NULL
    ADist <- NULL
    WWi <- NULL
    Xt <- NULL
    Zt <- NULL
    Yt <- NULL
    Xc <- NULL
    Zc <- NULL
    Yc <- NULL

    for (i in 1:N)
      {
        #treatment indicator for observation to be matched        
        TREATi <- Tr[i]

        # proceed with all observations if All==1
        # but only with treated observations if All=0        
        if ( (TREATi==1 & All!=1) | All==1 )
          {
            # covariate value for observation to be matched                        
            xx <- t(as.matrix(X[i,]))

            # covariate value for observation to be matched            
            zz <- t(as.matrix(Z[i,]))

            # outcome value for observation to be matched                        
            yy <- Y[i]

            #JSS: check *
            foo <- as.matrix(rep(1, Nx))
            DX <- (X - foo %*% xx) %*% t(ww)

            if (Kx>1)
              {
                #JSS
                foo <- t(DX*DX)
                Dist <- as.matrix(apply(foo, 2, sum))
              } else {
                Dist <- as.matrix(DX*DX)
              } #end of Kx

            # Dist distance to observation to be matched
            # is N by 1 vector

            #use the restriction matrix
            if (restrict.trigger)
              {
                for (j in 1:nrow(restrict))
                  {
                    if (restrict[j,1]==i)
                      {
                        if (restrict[j,3] < 0) {
                          Dist[restrict[j,2]] = .Machine$double.xmax
                        } else {
                          Dist[restrict[j,2]] = restrict[j,3]
                        }
                      } else if (restrict[j,2]==i) {
                        if (restrict[j,3] < 0) {
                          Dist[restrict[j,1]] = .Machine$double.xmax
                        } else {
                          Dist[restrict[j,1]] = restrict[j,3]
                        }
                      }
                  }
              } #end if restrict.trigger

            # set of potential matches (all observations with other treatment)
            # JSS, note:logical vector
            POTMAT <- Tr == (1-TREATi) 

            # X's for potential matches
#            XPOT <- X[POTMAT,]
            DistPot <- Dist[POTMAT,1]

#            TTPotMat <- TT[POTMAT,1]
            weightPot <- as.matrix(weight[POTMAT,1])

            # sorted distance of potential matches            
            S <- sort(DistPot)
            L <- order(DistPot)
            weightPot.sort <- weightPot[L,1]
            weightPot.sum <- cumsum(weightPot.sort)
            tt <- 1:(length(weightPot.sum))
            MMM <- min(tt[weightPot.sum>=M])

            # distance at last match
            Distmax <- S[MMM]

            # selection of actual matches 
            #logical index
            if (restrict.trigger)
              {
                ACTMAT <- POTMAT & ( (Dist <= (Distmax+cdd))  & (Dist < .Machine$double.xmax) )

                if (sum(ACTMAT) < 1)
                  next;                
              } else {
                ACTMAT <- POTMAT & ( Dist <= (Distmax+cdd) )
              }
            
            Ii <- i * matrix(1, nrow=sum(ACTMAT), ncol=1)
            IMi <- as.matrix(INN[ACTMAT,1])
            
            if(!is.null(ecaliper))
              {
                for (j in 1:length(Ii))
                  {
                    for( x in 1:Kx)
                      {
                        diff <- abs(X.orig[i,x] - X.orig[IMi[j], x])
                        if (diff > ecaliper[x])
                          {
#                            print(diff)

                            ACTMAT[IMi[j]] <- FALSE
                            sum.caliper.drops <- sum.caliper.drops+1
                            
                            break
                          }
                      } #x loop
                  } #j loop

                if (sum(ACTMAT) < 1)
                  next;
              } #ecaliper check

            # distance to actual matches            
            ACTDIST <- as.matrix(Dist[ACTMAT,1])

            # counts how many times each observation is matched.
            Kcount <- Kcount + weight[i] * weight*ACTMAT/sum(ACTMAT*weight)
            
            KKcount <- KKcount+weight[i,1] * weight*weight*ACTMAT /
              (sum(ACTMAT*weight)*sum(ACTMAT*weight))

            # counts how many times each observation is matched.
            # counts number of matches for observation i
            # Unless there are ties this should equal M

            Mi <- sum(weight*ACTMAT)            
            MMi[i,1] <- Mi
            Wi <- as.matrix(weight[ACTMAT,1])

            # mean of Y's for actual matches
            ymat <- t(Y[ACTMAT,1]) %*% Wi/Mi

            # mean of X's for actual matches
            # mean of Z's for actual matches
            if (length(Wi)>1)
              {
                xmat <- t(t(X[ACTMAT,]) %*% Wi/Mi)
                zmat <- t(t(Z[ACTMAT,]) %*% Wi/Mi)
              } else {
                xmat <- t(X[ACTMAT,]) * as.double(Wi)/Mi
                zmat <- t(Z[ACTMAT,]) * as.double(Wi)/Mi
              }

            # estimate causal effect on y for observation i
            YCAUS[i,1] <- TREATi %*% (yy-ymat)+(1-TREATi) %*% (ymat-yy)

            # difference between x and actual matches for observation i
            XCAUS[i,] <- TREATi %*% (xx-xmat)+(1-TREATi) %*% (xmat-xx)

            ZCAUS[i,] <- TREATi %*% (zz-zmat)+(1-TREATi) %*% (zmat-zz)

            # collect results
            I  <- rbind(I, i * matrix(1, nrow=sum(ACTMAT), ncol=1))
            DD <- rbind(DD, ACTDIST)
            IM <- rbind(IM, as.matrix(INN[ACTMAT,1]))
            IT <- rbind(IT, TREATi * as.matrix(rep(1, sum(ACTMAT))))
            IX <- rbind(IX, as.matrix(rep(1, sum(ACTMAT))) %*% xx)
            IZ <- rbind(IZ, as.matrix(rep(1, sum(ACTMAT))) %*% zz)
            IY <- rbind(IY, as.matrix(rep(1, sum(ACTMAT))) * yy)

            # weight for matches
            W <- as.matrix(c(W, weight[i,1] * Wi/Mi))
            ADist <- as.matrix(c(ADist, ACTDIST))
            WWi   <- as.matrix(c(WWi, Wi))

            if (TREATi==1)
              {

                if (ncol(X) > 1)
                  {
                    # covariates for treated
                    Xt <- rbind(Xt, as.matrix(rep(1, sum(ACTMAT))) %*% xx)                    
                    # covariate for matches
                    Xc <- rbind(Xc, X[ACTMAT,])
                    
                  } else {
                    # covariates for treated
                    Xt <- as.matrix(c(Xt, as.matrix(rep(1, sum(ACTMAT))) %*% xx))
                    # covariate for matches
                    Xc <- as.matrix(c(Xc, X[ACTMAT,]))
                  }

                if (ncol(Z) > 1)
                  {
                    # covariates for treated                
                    Zt <- rbind(Zt, as.matrix(rep(1, sum(ACTMAT))) %*% zz)
                    # covariate for matches
                    Zc <- rbind(Zc, Z[ACTMAT,])
                  } else {
                    # covariates for treated                
                    Zt <- as.matrix(c(Zt, as.matrix(rep(1, sum(ACTMAT))) %*% zz))
                    # covariate for matches                
                    Zc <- as.matrix(c(Zc, Z[ACTMAT,]))
                  }

                # outcome for treated                              
                Yt <- as.matrix(c(Yt, yy * as.matrix(rep(1, sum(ACTMAT))) ))
                # outcome for matches
                Yc <- as.matrix(c(Yc, Y[ACTMAT,1]))
              } else {

                if (ncol(X) > 1)
                  {
                    # covariates for controls
                    Xc <- rbind(Xc, as.matrix(rep(1, sum(ACTMAT))) %*% xx)
                    # covariate for matches
                    Xt <- rbind(Xt, X[ACTMAT,])
                  } else {
                    # covariates for controls                
                    Xc <- as.matrix(c(Xc, as.matrix(rep(1, sum(ACTMAT))) %*% xx))
                    # covariate for matches
                    Xt <- as.matrix(c(Xt, X[ACTMAT,]))
                  }

                if (ncol(Z) > 1)
                  {
                    # covariates for controls                
                    Zc <- rbind(Zc, as.matrix(rep(1, sum(ACTMAT))) %*% zz)
                    # covariate for matches                
                    Zt <- rbind(Zt, Z[ACTMAT,])
                  } else {
                    # covariates for controls                
                    Zc <- as.matrix(c(Zc, as.matrix(rep(1, sum(ACTMAT))) %*% zz))
                    # covariate for matches                
                    Zt <- as.matrix(c(Zt, Z[ACTMAT,]))
                  }
                # outcome for controls                               
                Yc <- as.matrix(c(Yc, as.matrix(rep(1, sum(ACTMAT))) * yy))
                # outcome for matches
                Yt <- as.matrix(c(Yt, Y[ACTMAT,1]))

              } #end of TREATi
          } #end of if
      } #i loop

    # transform matched covariates back for artificial data set
    Xt.u <- Xt
    Xc.u <- Xc
    IX.u <- IX
    for (k in 1:Kx)
      {
        Xt.u[,k] <- Mu.X[k,1]+Sig.X[k,1] * Xt.u[,k]
        Xc.u[,k] <- Mu.X[k,1]+Sig.X[k,1] * Xc.u[,k]
        IX.u[,k] <- Mu.X[k,1]+Sig.X[k,1] * IX.u[,k]
      }
    if (All!=1)
      {
        I  <- as.matrix(I[IT==1,])
        IM <- as.matrix(IM[IT==1,])
        IT <- as.matrix(IT[IT==1,])
        IY <- as.matrix(IY[IT==1,])
        Yc <- as.matrix(Yc[IT==1,])
        Yt <- as.matrix(Yt[IT==1,])
        W  <- as.matrix(W[IT==1,])
        ADist <- as.matrix(ADist[IT==1,])
        WWi   <- as.matrix(WWi[IT==1,])
        IX.u  <- as.matrix(IX.u[IT==1,])
        Xc.u  <- as.matrix(Xc.u[IT==1,])
        Xt.u  <- as.matrix(Xt.u[IT==1,])
        Xc    <- as.matrix(Xc[IT==1,])
        Xt <- as.matrix(Xt[IT==1,])
        Zc <- as.matrix(Zc[IT==1,])
        Zt <- as.matrix(Zt[IT==1,])
        IZ <- as.matrix(IZ[IT==1,])
      } #end of if

    if (length(I) < 1)
      {
        return(list(sum.caliper.drops=sum.caliper.drops,valid=0))
      } else if(length(I) < 2)
        {
          return(list(sum.caliper.drops=sum.caliper.drops,valid=1))          
        }

    if (BiasAdj==1)
      {
        # III. Regression of outcome on covariates for matches
        if (All==1)
          {
            # number of observations
            NNt <- nrow(Z)
            # add intercept        
            ZZt <- cbind(rep(1, NNt), Z)
            # number of covariates
            Nx <- nrow(ZZt)
            Kx <- ncol(ZZt)
            xw <- ZZt*(sqrt(Tr*Kcount) %*% t(as.matrix(rep(1,Kx))))
            
            foo <- min(eigen(t(xw)%*%xw, only.values=TRUE)$values)
            foo <- as.double(foo<=ccc)
            foo2 <- apply(xw, 2, sd)

            options(show.error.messages = FALSE)
            wout <- NULL
            try(wout <- solve( t(xw) %*% xw + diag(Kx) * ccc * (foo) * max(foo2)) %*%
                (t(xw) %*% (Y*sqrt(Tr*Kcount))))
            if(is.null(wout))
              {
                wout2 <- NULL
                try(wout2 <- ginv( t(xw) %*% xw + diag(Kx) * ccc * (foo) * max(foo2)) %*%
                    (t(xw) %*% (Y*sqrt(Tr*Kcount))))
                if(!is.null(wout2))
                  {
                    wout <-wout2
                    warning("using generalized inverse to calculate Bias Adjustment probably because of singular 'Z'")
                  }
              }
            options(show.error.messages = TRUE)
            if(is.null(wout))
              {
                warning("unable to calculate Bias Adjustment probably because of singular 'Z'")
                BiasAdj <- 0
              } else {
                NW <- nrow(wout)
#                KW <- ncol(wout)
                Alphat <- wout[2:NW,1]
              }
          } else {
            Alphat <- matrix(0, nrow=Kz, ncol=1)
          } #end if ALL
      }

    if(BiasAdj==1)
      {
        # III.b.  Controls
        NNc <- nrow(Z)
        ZZc <- cbind(matrix(1, nrow=NNc, ncol=1),Z)
        Nx <- nrow(ZZc)
        Kx <- ncol(ZZc)
        
        xw <- ZZc*(sqrt((1-Tr)*Kcount) %*% matrix(1, nrow=1, ncol=Kx))
        
        foo <- min(eigen(t(xw)%*%xw, only.values=TRUE)$values)
        foo <- as.double(foo<=ccc)
        foo2 <- apply(xw, 2, sd)

        options(show.error.messages = FALSE)
        wout <- NULL        
        try(wout <- solve( t(xw) %*% xw + diag(Kx) * ccc * (foo) * max(foo2)) %*%
            (t(xw) %*% (Y*sqrt((1-Tr)*Kcount))))
        if(is.null(wout))
          {
            wout2 <- NULL
            try(wout2 <- ginv( t(xw) %*% xw + diag(Kx) * ccc * (foo) * max(foo2)) %*%
                (t(xw) %*% (Y*sqrt((1-Tr)*Kcount))))
            if(!is.null(wout2))
              {
                wout <-wout2
                warning("using generalized inverse to calculate Bias Adjustment probably because of singular 'Z'")
              }
          }        
        options(show.error.messages = TRUE)
        if(is.null(wout))
          {
            warning("unable to calculate Bias Adjustment probably because of singular 'Z'")
            BiasAdj <- 0
          } else {
            NW <- nrow(wout)
#            KW <- ncol(wout)
            Alphac <- as.matrix(wout[2:NW,1])
          }
      }
    

    if(BiasAdj==1)
      {
        # III.c. adjust matched outcomes using regression adjustment for bias adjusted matching estimator

        SCAUS <- YCAUS-Tr*(ZCAUS %*% Alphac)-(1-Tr)*(ZCAUS %*% Alphat)
        # adjusted control outcome
        Yc.adj <- Yc+BiasAdj * (IZ-Zc) %*% Alphac
        # adjusted treated outcome
        Yt.adj <- Yt+BiasAdj*(IZ-Zt) %*% Alphat
        Tau.i <- Yt.adj - Yc.adj
      } else {
        Yc.adj <- Yc
        Yt.adj <- Yt
        Yt.adj <- Yt
        Tau.i <- Yt.adj - Yc.adj        
      }

    art.data <- cbind(I,IM,IT,DD,IY,Yc,Yt,W,WWi,ADist,IX.u,Xc.u,Xt.u,
                      Yc.adj,Yt.adj,Tau.i)


    # III. If conditional variance is needed, initialize variance vector
    # and loop through all observations

    Nx <- nrow(X)
    Kx <- ncol(X)

#   ww <- chol(Weight.matrix)
#   NN <- as.matrix(1:N)
    if (Var.calc>0)
      {
        Sig <- matrix(0, nrow=N, ncol=1)
        # overall standard deviation of outcome
        # std <- sd(Y)
        for (i in 1:N)
          {
            # treatment indicator observation to be matched
            TREATi <- Tr[i,1]
            # covariate value for observation to be matched
            xx <- X[i,]
            # potential matches are all observations with the same treatment value
            POTMAT <- (Tr==TREATi)
            POTMAT[i,1] <- 0
            weightPOT <- as.matrix(weight[POTMAT==1,1])
            DX <- (X - matrix(1, Nx,1) %*% xx) %*% t(ww)
            if (Kx>1)
              {
                foo <- apply(t(DX*DX), 2, sum)
                Dist <- as.matrix(foo)
              } else {
                Dist <- DX*DX
              }

            # distance to observation to be matched

            # Distance vector only for potential matches
            DistPot <- Dist[POTMAT==1,1]
            # sorted distance of potential matches
            S <- sort(DistPot)
            L <- order(DistPot)
            weightPOT.sort <- weightPOT[L,1]
            weightPOT.sum <- cumsum(weightPOT.sort)
            tt <-  1:(length(weightPOT.sum))
            MMM <- min(tt[weightPOT.sum >= Var.calc])
            MMM <- min(MMM,length(S))
            Distmax=S[MMM]

            # distance of Var_calc-th closest match
            ACTMAT <- (POTMAT==1) & (Dist<= (Distmax+ccc))

            # indicator for actual matches, that is all potential
            # matches closer than, or as close as the Var_calc-th
            # closest

            Yactmat <- as.matrix(c(Y[i,1], Y[ACTMAT,1]))
            weightactmat <- as.matrix(c(weight[i,1], weight[ACTMAT,1]))
            fm <- t(Yactmat) %*% weightactmat/sum(weightactmat)
            sm <- sum(Yactmat*Yactmat*weightactmat)/sum(weightactmat)
            sigsig <- (sm-fm %*% fm)*sum(weightactmat)/(sum(weightactmat)-1)
            
            # standard deviation of actual matches
            Sig[i,1] <- sqrt(sigsig)
          }# end of i loop
        #variance estimate 
        Sigs <- Sig*Sig
      } #end of var.calc > 0

    # matching estimator
    est <- t(W) %*% Tau.i/sum(W)
#    est.t <- sum((iot.t*Tr+iot.c*Kcount*Tr)*Y)/sum(iot.t*Tr+iot.c*Kcount*Tr)
#    est.c <- sum((iot.t*(1-Tr)+iot.c*Kcount*(1-Tr))*Y)/sum(iot.t*(1-Tr)+iot.c*Kcount*(1-Tr))

    if (Var.calc==0)
      {
        eps <- Tau.i - as.double(est)
        eps.sq <- eps*eps
        Sigs <- 0.5 * matrix(1, N, 1) %*% (t(eps.sq) %*% W)/sum(W)
#        sss <- sqrt(Sigs[1,1])
      } #end of Var.calc==0

    SN <- sum(iot.t)
    var.sample <- sum((Sigs*(iot.t+iot.c*Kcount)*(iot.t+iot.c*Kcount))/(SN*SN))

    if (All==1)
      {
        var.pop <- sum((Sigs*(iot.c*Kcount*Kcount+2*iot.c*Kcount-iot.c*KKcount))/(SN*SN))
      } else {
        var.pop=sum((Sigs*(iot.c*Kcount*Kcount-iot.c*KKcount))/(SN*SN))
      }

    if (BiasAdj==1)
      {
        dvar.pop <- sum(iot.t*(SCAUS-as.double(est))*(SCAUS-as.double(est)))/(SN*SN)
      } else {
        dvar.pop <- sum(iot.t*(YCAUS-as.double(est))*(YCAUS-as.double(est)))/(SN*SN)
      }

    var.pop <- var.pop + dvar.pop

    if (SAMPLE==1)
      {
        var <- var.sample
      } else {
        var <- max(var.sample, var.pop)
        var <- var.pop
      }

    var.cond <- max(var.sample,var.pop)-var.sample
    se <- sqrt(var)
    se.cond <- sqrt(var.cond)

#    Sig <- sqrt(Sigs)
#    aug.data <- cbind(Y,Tr,X,Z,Kcount,Sig,weight)

    if (All==2)
      est <- -1*est

#    if (exact==1)
#      {
#        Vt.u <- Xt.u[,(Kx-Mv+1):Kx]
#        Vc.u <- Xc.u[,(Kx-Mv+1):Kx]
#        Vdif <- abs(Vt.u-Vc.u)
#
#        if (Mv>1)
#          Vdif <- as.matrix(apply(t(Vdif), 2, sum))
#
#        em[1,1] <- length(Vdif)
#        em[2,1] <- sum(Vdif>0.000001)
#      }#end of exact==1

    return(list(est=est, se=se, se.cond=se.cond, W=W,
                sum.caliper.drops=sum.caliper.drops,
                art.data=art.data))
  }# end of Rmatch


#
# See Rosenbaum and Rubin (1985) and Smith and Todd Rejoinder (JofEconometrics) p.9
#
sdiff.pooled  <- function(Tr, Co, weights=rep(1,length(Co)),
                          weights.Tr=rep(1,length(Tr)),
                          weights.Co=rep(1,length(Co)),
                          match=FALSE)
  {
    if (!match)
      {
        obs.Tr <- sum(weights.Tr)
        obs.Co <- sum(weights.Co)
#        obs.total <- obs.Tr+obs.Co
        
        mean.Tr <- sum(Tr*weights.Tr)/obs.Tr
        mean.Co <- sum(Co*weights.Co)/obs.Co
        diff  <- mean.Tr - mean.Co

#match Rubin
#        mean.total <- sum(Tr*weights.Tr)/obs.total + sum(Co*weights.Co)/obs.total
#        var.total  <- sum( ( (Tr - mean.total)^2 )*weights.Tr)/(obs.total-1) +
#          sum( ( (Co - mean.total)^2 )*weights.Co)/(obs.total-1)
        var.pooled <- ( sum( ( (Tr - mean.Tr)^2)*weights.Tr)/(obs.Tr-1) +
          sum( ( (Co - mean.Co)^2 )*weights.Co)/(obs.Co-1) )/2 
        

        if(var.pooled==0 & diff==0)
          {
            sdiff <- 0
          } else {
            sdiff <- 100*diff/sqrt( var.pooled )
          }
        
      } else{
        diff  <- sum( (Tr-Co)*weights ) /sum(weights)
        mean.Tr <- sum(Tr*weights)/sum(weights)
        mean.Co <- sum(Co*weights)/sum(weights)

#match Rubin
        obs <- sum(weights)

#       obs for total        
#       obs = sum(weights)*2
#        mean.total <- (mean.Tr + mean.Co)/2
#        var.total  <- sum( ( (Tr - mean.total)^2 )*weights)/(obs-1) +
#          sum( ( (Co - mean.total)^2 )*weights)/(obs-1)
        
        var.pooled  <- ( sum( ( (Tr - mean.Tr)^2 )*weights)/(obs-1) +
          sum( ( (Co - mean.Co)^2 )*weights)/(obs-1) )/2

        if(var.pooled==0 & diff==0)
          {
            sdiff <- 0
          } else {
            sdiff <- 100*diff/sqrt(var.pooled)
          }        
      }

    return(sdiff)
  }


#
# STANDARDIZED BY THE VARIANCE OF THE TREATMENT GROUP
# See sdiff.rubin for Rosenbaum and Rubin (1985) and Smith and Todd Rejoinder (JofEconometrics) p.9
#
sdiff  <- function(Tr, Co, weights=rep(1,length(Co)),
                   weights.Tr=rep(1,length(Tr)),
                   weights.Co=rep(1,length(Co)),
                   match=FALSE,
                   estimand="ATT")

  {
    if (!match)
      {
        obs.Tr <- sum(weights.Tr)
        obs.Co <- sum(weights.Co)
        
        mean.Tr <- sum(Tr*weights.Tr)/obs.Tr
        mean.Co <- sum(Co*weights.Co)/obs.Co
        diff  <- mean.Tr - mean.Co

        if(estimand=="ATC")
          {
            var  <- sum( ( (Co - mean.Co)^2 )*weights.Co)/(obs.Co-1)
          } else {
            var  <- sum( ( (Tr - mean.Tr)^2 )*weights.Tr)/(obs.Tr-1)
          }

        if(var==0 & diff==0)
          {
            sdiff=0
          } else {
            sdiff <- 100*diff/sqrt( (var) )
          }
      } else{
        mean.Tr <- sum(Tr*weights)/sum(weights)
        mean.Co <- sum(Co*weights)/sum(weights)
        diff  <- mean.Tr - mean.Co

        if(estimand=="ATC")
          {
            var  <- sum( ( (Co - mean.Co)^2 )*weights)/(sum(weights)-1)
          } else {
            var  <- sum( ( (Tr - mean.Tr)^2 )*weights)/(sum(weights)-1)
          }

        if(var==0 & diff==0)
          {
            sdiff <- 0
          } else {
            sdiff <- 100*diff/sqrt( (var) )
          }        
      }

    return(sdiff)
  }

# function runs sdiff and t.test
#
balanceUV  <- function(Tr, Co, weights=rep(1,length(Co)), exact=FALSE, ks=FALSE, nboots=1000,
                       paired=TRUE, match=FALSE,
                       weights.Tr=rep(1,length(Tr)), weights.Co=rep(1,length(Co)),
                       estimand="ATT")
  {
    ks.out  <- NULL

    if (!match)
      {
        sbalance.pooled  <- sdiff.pooled(Tr=Tr, Co=Co,
                                       weights.Tr=weights.Tr,
                                       weights.Co=weights.Co,
                                       match=FALSE)

        sbalance.constvar  <- sdiff(Tr=Tr, Co=Co,
                                    weights.Tr=weights.Tr,
                                    weights.Co=weights.Co,
                                    match=FALSE,
                                    estimand=estimand)        

        obs.Tr <- sum(weights.Tr)
        obs.Co <- sum(weights.Co)        
        
        mean.Tr <- sum(Tr*weights.Tr)/obs.Tr
        mean.Co <- sum(Co*weights.Co)/obs.Co
        var.Tr  <- sum( ( (Tr - mean.Tr)^2 )*weights.Tr)/(obs.Tr-1)
        var.Co  <- sum( ( (Co - mean.Co)^2 )*weights.Co)/(obs.Co-1)
        var.ratio  <- var.Tr/var.Co

        qqsummary <- qqstats(x=Tr, y=Co, standardize=TRUE)
        qqsummary.raw <- qqstats(x=Tr, y=Co, standardize=FALSE)

        tt  <- Mt.test.unpaired(Tr, Co,
                                weights.Tr=weights.Tr,
                                weights.Co=weights.Co)

        if (ks)
          ks.out <- ks.boot(Tr, Co,nboots=nboots)
      } else {
        sbalance.pooled  <- sdiff(Tr=Tr, Co=Co, weights=weights, match=TRUE)
        sbalance.constvar  <- sdiff(Tr=Tr, Co=Co, weights=weights, match=TRUE, estimand=estimand)
        
        mean.Tr  <- sum(Tr*weights)/sum(weights);
        mean.Co  <- sum(Co*weights)/sum(weights);
        var.Tr  <- sum( ( (Tr - mean.Tr)^2 )*weights)/sum(weights);
        var.Co  <- sum( ( (Co - mean.Co)^2 )*weights)/sum(weights);
        var.ratio  <- var.Tr/var.Co

        qqsummary <- qqstats(x=Tr, y=Co, standardize=TRUE)
        qqsummary.raw <- qqstats(x=Tr, y=Co, standardize=FALSE)

        if(paired)
          {
            tt  <- Mt.test(Tr, Co, weights)
          } else {
            tt  <- Mt.test.unpaired(Tr, Co, weights.Tr=weights, weights.Co=weights)
          }
        
        if (ks)
          ks.out  <- ks.boot(Tr, Co, nboots=nboots)
      }        

    ret  <- list(sdiff=sbalance.constvar,
                 sdiff.pooled=sbalance.pooled,
                 mean.Tr=mean.Tr,mean.Co=mean.Co,
                 var.Tr=var.Tr,var.Co=var.Co, p.value=tt$p.value,
                 var.ratio=var.ratio,
                 ks=ks.out, tt=tt,
                 qqsummary=qqsummary,
                 qqsummary.raw=qqsummary.raw)

    class(ret) <-  "balanceUV"
    return(ret)
  } #balanceUV

summary.balanceUV  <- function(object, ..., digits=5)
  {

    if (!inherits(object,  "balanceUV")) {
      warning("Object not of class 'balanceUV'")
      return(NULL)
    }

    cat("mean treatment........", format(object$mean.Tr, digits=digits),"\n")
    cat("mean control..........", format(object$mean.Co, digits=digits),"\n")
    cat("std mean diff.........", format(object$sdiff, digits=digits),"\n\n")    

    cat("mean raw eQQ diff.....", format(object$qqsummary.raw$meandiff, digits=digits),"\n")
    cat("med  raw eQQ diff.....", format(object$qqsummary.raw$mediandiff, digits=digits),"\n")
    cat("max  raw eQQ diff.....", format(object$qqsummary.raw$maxdiff, digits=digits),"\n\n")

    cat("mean eCDF diff........", format(object$qqsummary$meandiff, digits=digits),"\n")
    cat("med  eCDF diff........", format(object$qqsummary$mediandiff, digits=digits),"\n")
    cat("max  eCDF diff........", format(object$qqsummary$maxdiff, digits=digits),"\n\n")

    cat("var ratio (Tr/Co).....", format(object$var.ratio, digits=digits),"\n")
    cat("T-test p-value........", format.pval(object$tt$p.value,digits=digits), "\n")            
    if (!is.null(object$ks))
      {
        if(!is.na(object$ks$ks.boot.pvalue))
          {
            cat("KS Bootstrap p-value..", format.pval(object$ks$ks.boot.pvalue, digits=digits), "\n")
          }
        cat("KS Naive p-value......", format(object$ks$ks$p.value, digits=digits), "\n")                        
        cat("KS Statistic..........", format(object$ks$ks$statistic, digits=digits), "\n")
      }
    cat("\n")
  } #end of summary.balanceUV


#print Before and After balanceUV together
PrintBalanceUV  <- function(BeforeBalance, AfterBalance, ..., digits=5)
  {

    if (!inherits(BeforeBalance,  "balanceUV")) {
      warning("BeforeBalance not of class 'balanceUV'")
      return(NULL)
    }

    if (!inherits(AfterBalance, "balanceUV")) {
      warning("AfterBalance not of class 'balanceUV'")
      return(NULL)
    }

    space.size <- digits*2
#    space <- rep(" ",space.size)

    cat("                      ", "Before Matching", "\t \t After Matching\n")
    cat("mean treatment........", format(BeforeBalance$mean.Tr, digits=digits, width=space.size), "\t \t",
        format(AfterBalance$mean.Tr, digits=digits, width=space.size), 
        "\n")
    cat("mean control..........", format(BeforeBalance$mean.Co, digits=digits, width=space.size), "\t \t",
        format(AfterBalance$mean.Co, digits=digits, width=space.size),
        "\n")
    cat("std mean diff.........", format(BeforeBalance$sdiff, digits=digits, width=space.size), "\t \t",
        format(AfterBalance$sdiff, digits=digits, width=space.size),
        "\n\n")
    
    cat("mean raw eQQ diff.....", format(BeforeBalance$qqsummary.raw$meandiff, digits=digits, width=space.size), "\t \t",
        format(AfterBalance$qqsummary.raw$meandiff, digits=digits, width=space.size),
        "\n")
    cat("med  raw eQQ diff.....", format(BeforeBalance$qqsummary.raw$mediandiff, digits=digits, width=space.size), "\t \t",
        format(AfterBalance$qqsummary.raw$mediandiff, digits=digits, width=space.size),
        "\n")
    cat("max  raw eQQ diff.....", format(BeforeBalance$qqsummary.raw$maxdiff, digits=digits, width=space.size), "\t \t",
        format(AfterBalance$qqsummary.raw$maxdiff, digits=digits, width=space.size),
        "\n\n")

    cat("mean eCDF diff........", format(BeforeBalance$qqsummary$meandiff, digits=digits, width=space.size), "\t \t",
        format(AfterBalance$qqsummary$meandiff, digits=digits, width=space.size),
        "\n")
    cat("med  eCDF diff........", format(BeforeBalance$qqsummary$mediandiff, digits=digits, width=space.size), "\t \t",
        format(AfterBalance$qqsummary$mediandiff, digits=digits, width=space.size),
        "\n")
    cat("max  eCDF diff........", format(BeforeBalance$qqsummary$maxdiff, digits=digits, width=space.size), "\t \t",
        format(AfterBalance$qqsummary$maxdiff, digits=digits, width=space.size),
        "\n\n")

    cat("var ratio (Tr/Co).....", format(BeforeBalance$var.ratio, digits=digits, width=space.size), "\t \t",
        format(AfterBalance$var.ratio, digits=digits, width=space.size),
        "\n")
    cat("T-test p-value........", format(format.pval(BeforeBalance$tt$p.value,digits=digits), justify="right", width=space.size), "\t \t",
        format(format.pval(AfterBalance$tt$p.value,digits=digits), justify="right", width=space.size),
        "\n")            
    if (!is.null(BeforeBalance$ks))
      {
        if(!is.na(BeforeBalance$ks$ks.boot.pvalue))
          {
            cat("KS Bootstrap p-value..", format(format.pval(BeforeBalance$ks$ks.boot.pvalue, digits=digits),  justify="right",width=space.size), "\t \t",
                format(format.pval(AfterBalance$ks$ks.boot.pvalue, digits=digits),  justify="right", width=space.size),
                "\n")
          }
        cat("KS Naive p-value......", format(format.pval(BeforeBalance$ks$ks$p.value, digits=digits),  justify="right",width=space.size), "\t \t",
            format(format.pval(AfterBalance$ks$ks$p.value, digits=digits),  justify="right",width=space.size),
            "\n")                        
        cat("KS Statistic..........", format(BeforeBalance$ks$ks$statistic, digits=digits, width=space.size), "\t \t",
            format(AfterBalance$ks$ks$statistic, digits=digits, width=space.size),
            "\n")
      }
    cat("\n")        
  } #end of PrintBalanceUV


#removed as of 0.99-7 (codetools)
#McNemar  <- function(Tr, Co, weights=rep(1,length(Tr)))

McNemar2 <- function (Tr, Co, correct = TRUE, weights=rep(1,length(Tr)))
{
  x  <- Tr
  y  <- Co
  if (is.matrix(x)) {
    stop("this version of McNemar cannot handle x being a matrix")
  } else {
    if (is.null(y)) 
      stop("if x is not a matrix, y must be given")
    if (length(x) != length(y)) 
      stop("x and y must have the same length")
    DNAME <- paste(deparse(substitute(x)), "and", deparse(substitute(y)))
    OK <- complete.cases(x, y)
    x <- factor(x[OK])
    y <- factor(y[OK])
    r <- nlevels(x)
    if ((r < 2) || (nlevels(y) != r))
      {
        stop("x and y must have the same number of levels (minimum 2)")
      }
  }
  tx <- table(x, y)
  facs  <- levels(x)
  txw  <- tx
  for(i in 1:r)
    {
      for(j in 1:r)
        {
          indx  <- x==facs[i] & y==facs[j]
          txw[i,j]  <- sum(weights[indx]);
        }
    }
  pdiscordant  <- sum( ( (x!=y)*weights )/sum(weights) )
  x  <- txw
  
  PARAMETER <- r * (r - 1)/2
  METHOD <- "McNemar's Chi-squared test"
  if (correct && (r == 2) && any(x - t(x))) {
    y <- (abs(x - t(x)) - 1)
    METHOD <- paste(METHOD, "with continuity correction")
  } else y <- x - t(x)
  x <- x + t(x)
  STATISTIC <- sum(y[upper.tri(x)]^2/x[upper.tri(x)])
  PVAL <- pchisq(STATISTIC, PARAMETER, lower.tail = FALSE)
  names(STATISTIC) <- "McNemar's chi-squared"
  names(PARAMETER) <- "df"

  RVAL <- list(statistic = STATISTIC, parameter = PARAMETER, 
               p.value = PVAL, method = METHOD, data.name = DNAME,
               pdiscordant=pdiscordant)
  class(RVAL) <- "htest"
  return(RVAL)
}

ks<-function(x,y,w=F,sig=T){
#  Compute the Kolmogorov-Smirnov test statistic
#
# Code for the Kolmogorov-Smirnov test is adopted from the Splus code
# written by Rand R. Wilcox for his book titled "Introduction to
# Robust Estimation and Hypothesis Testing." Academic Press, 1997.
#
#  Also see 
#@book( knuth1998,
#  author=       {Knuth, Donald E.},
#  title=        {The Art of Computer Programming, Vol. 2: Seminumerical Algorithms},
#  edition=      "3rd",
#  publisher=    "Addison-Wesley", address= "Reading: MA", 
#  year=         1998
#)
# and Wilcox 1997 
#
#  w=T computes the weighted version instead. 
#  sig=T indicates that the exact significance level is to be computed.
#  If there are ties, the reported significance level is exact when
#  using the unweighted test, but for the weighted test the reported
#  level is too high.
#
#  This function uses the functions ecdf, kstiesig, kssig and kswsig
#
#  This function returns the value of the test statistic, the approximate .05
#  critical value, and the exact significance level if sig=T.
#
#  Missing values are automatically removed
#

  ecdf<-function(x,val){
  #  compute empirical cdf for data in x evaluated at val
  #  That is, estimate P(X <= val) 
  #
    ecdf<-length(x[x<=val])/length(x)
    ecdf
  }#end ecdf
  
  x<-x[!is.na(x)]
  y<-y[!is.na(y)]
  w<-as.logical(w)
  sig<-as.logical(sig)
  tie<-logical(1)
  tie<-F
  siglevel<-NA
  z<-sort(c(x,y))  # Pool and sort the observations
  for (i in 2:length(z))if(z[i-1]==z[i])tie<-T #check for ties
  v<-1   # Initializes v
  for (i in 1:length(z))v[i]<-abs(ecdf(x,z[i])-ecdf(y,z[i]))
  ks<-max(v)
  crit<-1.36*sqrt((length(x)+length(y))/(length(x)*length(y))) # Approximate
#                                                       .05 critical value
  if(!w && sig && !tie)siglevel<-kssig(length(x),length(y),ks)
  if(!w && sig && tie)siglevel<-kstiesig(x,y,ks)
  if(w){
    crit<-(max(length(x),length(y))-5)*.48/95+2.58+abs(length(x)-length(y))*.44/95
    if(length(x)>100 || length(y)>100){
      print("When either sample size is greater than 100,")
      print("the approximate critical value can be inaccurate.")
      print("It is recommended that the exact significance level be computed.")
    }
    for (i in 1:length(z)){
      temp<-(length(x)*ecdf(x,z[i])+length(y)*ecdf(y,z[i]))/length(z)
      temp<-temp*(1.-temp)
      v[i]<-v[i]/sqrt(temp)
    }
    v<-v[!is.na(v)]
    ks<-max(v)*sqrt(length(x)*length(y)/length(z))
    if(sig)siglevel<-kswsig(length(x),length(y),ks)
    if(tie && sig){
      print("Ties were detected. The reported significance level")
      print("of the weighted Kolmogorov-Smirnov test statistic is not exact.")
    }}
                                        #round off siglevel in a nicer way
  if(is.double(ks) & is.double(crit) & !is.na(ks) & !is.na(crit))
    {
      if (is.na(siglevel) & ks < crit)
        {
          siglevel  <- 0.99999837212332
        }
      if (is.double(siglevel) & !is.na(siglevel))
        {
          if (siglevel < 0)
            siglevel  <- 0
        }
    }
  list(test=ks,critval=crit,pval=siglevel)
}


kssig<-function(m,n,val){
#
#    Compute significance level of the  Kolmogorov-Smirnov test statistic
#    m=sample size of first group
#    n=sample size of second group
#    val=observed value of test statistic
#
  cmat<-matrix(0,m+1,n+1)
  umat<-matrix(0,m+1,n+1)
  for (i in 0:m){
    for (j in 0:n)cmat[i+1,j+1]<-abs(i/m-j/n)
  }
  cmat<-ifelse(cmat<=val,1e0,0e0)
  for (i in 0:m){
    for (j in 0:n)if(i*j==0)umat[i+1,j+1]<-cmat[i+1,j+1]
    else umat[i+1,j+1]<-cmat[i+1,j+1]*(umat[i+1,j]+umat[i,j+1])
  }
  term<-lgamma(m+n+1)-lgamma(m+1)-lgamma(n+1)
  kssig<-1.-umat[m+1,n+1]/exp(term)
  return(kssig)
}

kstiesig<-function(x,y,val){
#
#    Compute significance level of the  Kolmogorov-Smirnov test statistic
#    for the data in x and y.
#    This function allows ties among the  values.
#    val=observed value of test statistic
#
m<-length(x)
n<-length(y)
z<-c(x,y)
z<-sort(z)
cmat<-matrix(0,m+1,n+1)
umat<-matrix(0,m+1,n+1)
for (i in 0:m){
for (j in 0:n){
if(abs(i/m-j/n)<=val)cmat[i+1,j+1]<-1e0
k<-i+j
if(k > 0 && k<length(z) && z[k]==z[k+1])cmat[i+1,j+1]<-1
}
}
for (i in 0:m){
for (j in 0:n)if(i*j==0)umat[i+1,j+1]<-cmat[i+1,j+1]
else umat[i+1,j+1]<-cmat[i+1,j+1]*(umat[i+1,j]+umat[i,j+1])
}
term<-lgamma(m+n+1)-lgamma(m+1)-lgamma(n+1)
kstiesig<-1.-umat[m+1,n+1]/exp(term)
kstiesig
}



kswsig<-function(m,n,val){
#
#    Compute significance level of the weighted
#    Kolmogorov-Smirnov test statistic
#
#    m=sample size of first group
#    n=sample size of second group
#    val=observed value of test statistic
#
  mpn<-m+n
  cmat<-matrix(0,m+1,n+1)
  umat<-matrix(0,m+1,n+1)
  for (i in 1:m-1){
    for (j in 1:n-1)cmat[i+1,j+1]<-abs(i/m-j/n)*sqrt(m*n/((i+j)*(1-(i+j)/mpn)))
  }
  cmat<-ifelse(cmat<=val,1,0)
  for (i in 0:m){
    for (j in 0:n)if(i*j==0)umat[i+1,j+1]<-cmat[i+1,j+1]
    else umat[i+1,j+1]<-cmat[i+1,j+1]*(umat[i+1,j]+umat[i,j+1])
  }
  term<-lgamma(m+n+1)-lgamma(m+1)-lgamma(n+1)
  kswsig<-1.-umat[m+1,n+1]/exp(term)
  kswsig
}

Mks.test.handler <- function(w)
  {
    #suppress the following warning:
    #In ks.test() :
    # p-values will be approximate in the presence of ties
    if( any( grepl( "ties", w) ) )
      invokeRestart( "muffleWarning" )

    #invoke as:
    #withCallingHandlers( ks.test(round(x1), round(x2)), warning = Mks.test.handler )
  }

Mks.test <- function(...)
  {
    RVAL <- withCallingHandlers( ks.test(...), warning = Mks.test.handler )
    return(RVAL)
  }

Mt.test  <- function(Tr, Co, weights)
  {
    v1  <- Tr-Co
    estimate  <- sum(v1*weights)/sum(weights)
    var1  <- sum( ((v1-estimate)^2)*weights )/( sum(weights)*sum(weights) )

    parameter  <- Inf

    #get rid of NA for t.test!
    if (estimate==0 & var1==0)
      {
        statistic = 0        
        p.value = 1
      }  else {
        statistic  <- estimate/sqrt(var1)
        
        p.value    <- (1-MATCHpt(abs(statistic), df=sum(weights)-1))*2
      }

    z  <- list(statistic=statistic, parameter=parameter, p.value=p.value,
               estimate=estimate)
    return(z)
  } #end of Mt.test

Mt.test.unpaired  <- function(Tr, Co,
                              weights.Tr=rep(1,length(Tr)),
                              weights.Co=rep(1,length(Co)))
  {
    obs.Tr <- sum(weights.Tr)
    obs.Co <- sum(weights.Co)
    
    mean.Tr <- sum(Tr*weights.Tr)/obs.Tr
    mean.Co <- sum(Co*weights.Co)/obs.Co
    estimate <- mean.Tr-mean.Co
    var.Tr  <- sum( ( (Tr - mean.Tr)^2 )*weights.Tr)/(obs.Tr-1)
    var.Co  <- sum( ( (Co - mean.Co)^2 )*weights.Co)/(obs.Co-1)
    dim <- sqrt(var.Tr/obs.Tr + var.Co/obs.Co)

    parameter  <- Inf

    #get rid of NA for t.test!
    if (estimate==0 & dim==0)
      {
        statistic = 0
        p.value = 1
      }  else {
        statistic  <- estimate/dim

        a1 <- var.Tr/obs.Tr
        a2 <- var.Co/obs.Co
        dof <- ((a1 + a2)^2)/( (a1^2)/(obs.Tr - 1) + (a2^2)/(obs.Co - 1) )    
        
        p.value    <- (1-MATCHpt(abs(statistic), df=dof))*2
      }

    z  <- list(statistic=statistic, parameter=parameter, p.value=p.value,
               estimate=estimate)
    return(z)
  } #end of Mt.test.unpaired

MatchBalance <- function(formul, data=NULL, match.out=NULL, ks=TRUE, 
                         nboots=500, weights=NULL,
                         digits=5, paired=TRUE, print.level=1)
  {
    if(!is.list(match.out) & !is.null(match.out)) {
      warning("'Match' object contains no valid matches")
      match.out  <- NULL
    }

    if ((!inherits(match.out, "Match")) & (!inherits(match.out, "Matchby")) & (!is.null(match.out)) ) {
      warning("Object not of class 'Match'")
      match.out  <- NULL
    }

    orig.na.action <- as.character(options("na.action"))
    options("na.action"=na.pass)            
    if (is.null(data))
      {
        xdata <- as.data.frame(get.xdata(formul,datafr=environment(formul)))
        Tr <- as.double(get.ydata(formul,datafr=environment(formul)))

      } else {
        data  <- as.data.frame(data)

        xdata  <- as.data.frame(get.xdata(formul, data))
        Tr  <- as.double(get.ydata(formul, data))
      }
    options("na.action"=orig.na.action)

    if (is.null(weights))
      weights <- rep(1,length(Tr))

    if(!is.numeric(weights))
      stop("'weights' must be a numeric vector")

    if( sum(is.na(xdata))!=0 | sum(is.na(Tr))!=0)
      {

        if(orig.na.action!="na.omit" & orig.na.action!="na.exclude" & orig.na.action!="na.fail")
          warning("'na.action' should be set to 'na.omit', 'na.exclude' or 'na.fail' see 'help(na.action)'")

        if (orig.na.action=="na.fail")
          {
            stop("NA's found in data input.")            
          } else {
            warning("NA's found in data input.  IT IS HIGHLY RECOMMENDED THAT YOU TEST IF THE MISSING VALUES ARE BALANCED INSTEAD OF JUST DELETING THEM.")
          }

        indx1 <- apply(is.na(xdata),1,sum)==0 & is.na(Tr)==0
        Tr <- Tr[indx1]
        xdata = xdata[indx1,]
        weights <- weights[indx1]
      } #end of na

    if (sum(Tr !=1 & Tr !=0) > 0) {
      stop("Treatment indicator must be a logical variable---i.e., TRUE (1) or FALSE (0)")
    }

    nvars  <- ncol(xdata)
    names.xdata  <- names(xdata)

    findx  <- 1
    if (sum(xdata[,1]==rep(1,nrow(xdata)))==nrow(xdata))
      {
        findx  <- 2
      }

    if(nboots < 10 & nboots > 0)
      nboots <- 10
    
    if (ks)
      {
        ks.bm <- KSbootBalanceSummary(index.treated=(Tr==0),
                                      index.control=(Tr==1),
                                      X=xdata[,findx:nvars],
                                      nboots=nboots)

        if (!is.null(match.out))
          {
            ks.am <- KSbootBalanceSummary(index.treated=match.out$index.treated,
                                          index.control=match.out$index.control,
                                          X=xdata[,findx:nvars],
                                          nboots=nboots)
          }
      } 

    BeforeMatchingBalance <- list()
    AfterMatchingBalance <- list()

    BMsmallest.p.value <- 1
    BMsmallest.number <- 1
    BMsmallest.name <- names.xdata[findx]

    AMsmallest.p.value <- NULL
    AMsmallest.number <- NULL
    AMsmallest.name <- NULL
    
    if (!is.null(match.out))
      {
        AMsmallest.p.value <- 1
        AMsmallest.number <- 1
        AMsmallest.name <- names.xdata[findx]        
      }

    for( i in findx:nvars)
      {
        count <- i-findx+1
        if(print.level > 0)
          cat("\n***** (V",count,") ", names.xdata[i]," *****\n",sep="")
        
        ks.do  <- FALSE
        is.dummy  <- length(unique( xdata[,i] )) < 3
        if (ks & !is.dummy)
          ks.do  <- TRUE

        BeforeMatchingBalance[[count]]  <-  balanceUV(xdata[,i][Tr==1], xdata[,i][Tr==0], nboots=0,
                                                      weights.Tr=weights[Tr==1], weights.Co=weights[Tr==0],
                                                      match=FALSE)
        
        if (BeforeMatchingBalance[[count]]$tt$p.value < BMsmallest.p.value)
          {
            BMsmallest.p.value <- BeforeMatchingBalance[[count]]$tt$p.value
            BMsmallest.number <- count
            BMsmallest.name <- names.xdata[i]            
          } else if (BeforeMatchingBalance[[count]]$tt$p.value == BMsmallest.p.value)
            {
              BMsmallest.number <- c(BMsmallest.number,count)
              BMsmallest.name <- c(BMsmallest.name,names.xdata[i])
            }
        
        if (ks.do)
          {
            BeforeMatchingBalance[[count]]$ks <- list()
            BeforeMatchingBalance[[count]]$ks$ks <- list()
            BeforeMatchingBalance[[count]]$ks$ks$p.value <- ks.bm$ks.naive.pval[count]
            BeforeMatchingBalance[[count]]$ks$ks$statistic <- ks.bm$ks.stat[count]              
            if (nboots > 0)
              {
                BeforeMatchingBalance[[count]]$ks$ks.boot.pvalue <- ks.bm$ks.boot.pval[count]

                if (ks.bm$ks.boot.pval[count] < BMsmallest.p.value)
                  {
                    BMsmallest.p.value <- ks.bm$ks.boot.pval[count]
                    BMsmallest.number <- count
                    BMsmallest.name <- names.xdata[i]            
                  } else if ( (ks.bm$ks.boot.pval[count] == BMsmallest.p.value) & (sum(BMsmallest.number==count)==0) )
                    {
                      BMsmallest.number <- c(BMsmallest.number,count)
                      BMsmallest.name <- c(BMsmallest.name,names.xdata[i])
                    }
              } else {
                BeforeMatchingBalance[[count]]$ks$ks.boot.pvalue <- NA

                if (ks.bm$ks.naive.pval[count] < BMsmallest.p.value)
                  {
                    BMsmallest.p.value <- ks.bm$ks.naive.pval[count]
                    BMsmallest.number <- count
                    BMsmallest.name <- names.xdata[i]            
                  } else if ( (ks.bm$ks.naive.pval[count] == BMsmallest.p.value) & (sum(BMsmallest.number==count)==0) )
                    {
                      BMsmallest.number <- c(BMsmallest.number,count)
                      BMsmallest.name <- c(BMsmallest.name,names.xdata[i])
                    }              
              }
            
          } else {
            BeforeMatchingBalance[[count]]$ks <- NULL
          }
        
        if (!is.null(match.out))
          {
            AfterMatchingBalance[[count]]  <- balanceUV(xdata[,i][match.out$index.treated],
                                                        xdata[,i][match.out$index.control],
                                                        weights=match.out$weights, nboots=0,
                                                        paired=paired, match=TRUE)
            
            if (AfterMatchingBalance[[count]]$tt$p.value < AMsmallest.p.value)
              {
                AMsmallest.p.value <- AfterMatchingBalance[[count]]$tt$p.value
                AMsmallest.number <- count
                AMsmallest.name <- names.xdata[i]            
              } else if ( (AfterMatchingBalance[[count]]$tt$p.value == AMsmallest.p.value) & (sum(AMsmallest.number==count)==0) )
                    {
                      AMsmallest.number <- c(AMsmallest.number,count)
                      AMsmallest.name <- c(AMsmallest.name,names.xdata[i])
                    }
            
            if (ks.do)
              {                
                AfterMatchingBalance[[count]]$ks <- list()
                AfterMatchingBalance[[count]]$ks$ks <- list()
                AfterMatchingBalance[[count]]$ks$ks$p.value <- ks.am$ks.naive.pval[count]
                AfterMatchingBalance[[count]]$ks$ks$statistic <- ks.am$ks.stat[count]
                if (nboots > 0)
                  {
                    AfterMatchingBalance[[count]]$ks$ks.boot.pvalue <- ks.am$ks.boot.pval[count]

                    if (ks.am$ks.boot.pval[count] < AMsmallest.p.value)
                      {
                        AMsmallest.p.value <- ks.am$ks.boot.pval[count]
                        AMsmallest.number <- count
                        AMsmallest.name <- names.xdata[i]            
                      } else if ( (ks.am$ks.boot.pval[count] == AMsmallest.p.value) & (sum(AMsmallest.number==count)==0) )
                        {
                          AMsmallest.number <- c(AMsmallest.number,count)
                          AMsmallest.name <- c(AMsmallest.name,names.xdata[i])
                        }
                  } else {
                    AfterMatchingBalance[[count]]$ks$ks.boot.pvalue <- NA

                    if (ks.am$ks.naive.pval[count] < AMsmallest.p.value)
                      {
                        AMsmallest.p.value <- ks.am$ks.naive.pval[count]
                        AMsmallest.number <- count
                        AMsmallest.name <- names.xdata[i]            
                      } else if ( (ks.am$ks.naive.pval[count] == AMsmallest.p.value) & (sum(AMsmallest.number==count)==0) )
                        {
                          AMsmallest.number <- c(AMsmallest.number,count)
                          AMsmallest.name <- c(AMsmallest.name,names.xdata[i])
                        }              
                  }                    
              } else {
                AfterMatchingBalance[[count]]$ks <- NULL
              }
            if(print.level > 0)
              PrintBalanceUV(BeforeMatchingBalance[[count]], AfterMatchingBalance[[count]], digits=digits)
          } else {
            if(print.level > 0)
              {
                cat("before matching:\n")
                summary(BeforeMatchingBalance[[count]], digits=digits)
              }
          } #end of if match.out
      } #end of for loop
    
    if(print.level & ( (nvars-findx+1) > 1))
      {
        cat("\n")

        if (BMsmallest.p.value < 1)
          {
            cat("Before Matching Minimum p.value:", format.pval(BMsmallest.p.value, digits=digits),"\n")
            cat("Variable Name(s):",BMsmallest.name, " Number(s):",BMsmallest.number,"\n\n")
          } else {
            cat("Before Matching Minimum p.value: 1\n\n")
          }

        if (!is.null(match.out))
          {
            if(AMsmallest.p.value < 1)
              {
                cat("After Matching Minimum p.value:", format.pval(AMsmallest.p.value, digits=digits),"\n")
                cat("Variable Name(s):",AMsmallest.name, " Number(s):",AMsmallest.number,"\n\n")
              } else {
                cat("After Matching Minimum p.value: 1\n\n")
              }
          } #end of !is.null(match.out)
      }#end of print.level & (nvars > 1)
  
    return(invisible(list(BeforeMatching=BeforeMatchingBalance,
                          AfterMatching=AfterMatchingBalance,
                          BMsmallest.p.value=BMsmallest.p.value,
                          BMsmallestVarName=BMsmallest.name,
                          BMsmallestVarNumber=BMsmallest.number,
                          AMsmallest.p.value=AMsmallest.p.value,
                          AMsmallestVarName=AMsmallest.name,
                          AMsmallestVarNumber=AMsmallest.number)))
  } #end of MatchBalance


get.xdata <- function(formul, datafr) {
  t1 <- terms(formul, data=datafr);
  if (length(attr(t1, "term.labels"))==0 & attr(t1, "intercept")==0) {
    m <- NULL;  # no regressors specified for the model matrix
  } else {
    m <- model.matrix(formul, data=datafr, drop.unused.levels = TRUE)
  }
  return(m);
}


# get.ydata:
# Return response vector corresponding to the formula in formul
# 
get.ydata <- function(formul, datafr) {
  t1 <- terms(formul, data=datafr);
  if (length(attr(t1, "response"))==0) {
    m <- NULL;  # no response variable specified
  }  else {
    m <- model.response(model.frame(formul, data=datafr))
  }
  return(m);
}

#
# bootstrap ks test implemented.  Fast version
#

ks.boot  <- function(Tr, Co, nboots=1000, alternative = c("two.sided", "less", "greater"), print.level=0)
  {
    alternative <- match.arg(alternative)
    tol <- sqrt(.Machine$double.eps)
    Tr <- Tr[!is.na(Tr)]
    Co <- Co[!is.na(Co)]

    w    <- c(Tr, Co)
    obs  <- length(w)
    n.x <- length(Tr)
    n.y <- length(Co)
    cutp <- n.x
    ks.boot.pval <- NULL
    bbcount <- 0

    if (nboots < 10)
      {
        nboots  <- 10
        warning("At least 10 'nboots' must be run; seting 'nboots' to 10")
      }

    if (nboots < 500)
      warning("For publication quality p-values it is recommended that 'nboots'\n be set equal to at least 500 (preferably 1000)") 
    
    fs.ks  <- Mks.test(Tr, Co, alternative=alternative)    

    if (alternative=="two.sided")
      {
        if (print.level > 0)
          cat("ks.boot: two.sided test\n")
        for (bb in 1:nboots)
          {
            if (print.level > 1)
              cat("s:", bb, "\n")
            
            sindx  <- sample(1:obs, obs, replace=TRUE)
            
            X1tmp <- w[sindx[1:cutp]]
            X2tmp <- w[sindx[(cutp+1):obs]]
            
            s.ks <- ks.fast(X1tmp, X2tmp, n.x=n.x, n.y=n.y, n=obs)
            
            if (s.ks >= (fs.ks$statistic - tol) )
              bbcount  <- bbcount + 1
          }
      } else {
        if (print.level > 0)
          cat("ks.boot:",alternative,"test\n")            
        for (bb in 1:nboots)
          {
            if (print.level > 1)
              cat("s:", bb, "\n")
            
            sindx  <- sample(1:obs, obs, replace=TRUE)
            
            X1tmp <- w[sindx[1:cutp]]
            X2tmp <- w[sindx[(cutp+1):obs]]
            
            s.ks <- Mks.test(X1tmp, X2tmp, alternative=alternative)$statistic
            
            if (s.ks >= (fs.ks$statistic - tol) )
              bbcount  <- bbcount + 1                                
          }        
      }
    ks.boot.pval  <- bbcount/nboots
    
    ret  <- list(ks.boot.pvalue=ks.boot.pval, ks=fs.ks, nboots=nboots)
    class(ret)  <- "ks.boot"
    
    return(ret)
  } #end of ks.boot

summary.ks.boot <- function(object, ..., digits=5)
  {
    if(!is.list(object)) {
      warning("object not a valid 'ks.boot' object")
      return()
    }

    if (!inherits(object,  "ks.boot")) {
      warning("Object not of class 'ks.boot'")
      return()
    }    
        
    cat("\n")
    cat("Bootstrap p-value:    ", format.pval(object$ks.boot.pvalue, digits=digits), "\n")
    cat("Naive p-value:        ", format(object$ks$p.value, digits=digits), "\n")
    cat("Full Sample Statistic:", format(object$ks$statistic, digits=digits), "\n")
#    cat("nboots completed      ", object$nboots, "\n")
    cat("\n")

    z <- list()
    class(z) <- "summary.ks.boot"
    return(invisible(z))
  } #end of summary.ks.boot

print.summary.ks.boot <- function(x, ...)
  {
    invisible(NULL)
  }


RmatchLoop <- function(Y, Tr, X, Z, V, All, M, BiasAdj, Weight, Weight.matrix, Var.calc, weight,
                       SAMPLE, ccc, cdd, ecaliper=NULL, exact=NULL, caliper=NULL, restrict=NULL,
                       MatchLoopC.indx=NULL, weights.flag, replace=TRUE, ties=TRUE,
                       version="standard", MatchbyAI=FALSE)
  {
    s1 <- MatchGenoudStage1caliper(Tr=Tr, X=X, All=All, M=M, weights=weight,
                                   exact=exact, caliper=caliper,
                                   distance.tolerance=cdd,
                                   tolerance=ccc)

    sum.caliper.drops <- 0
    X.orig <- X

#are we using the restriction matrix?
    if(is.matrix(restrict)) {
      restrict.trigger <- TRUE
    } else {
      restrict.trigger <- FALSE
    }    
    
# if SATC is to be estimated the treatment indicator is reversed    
    if (All==2)
      Tr <- 1-Tr

# check on the number of matches, to make sure the number is within the limits
# feasible given the number of observations in both groups.
    if (All==1)
      {
        M <- min(M,min(sum(Tr),sum(1-Tr)));        
      } else {
        M <- min(M,sum(1-Tr));
      }

# two slippage parameters that are used to determine whether distances are equal
# distances less than ccc or cdd are interpeted as zero.
# these are passed in.  ccc, cdd


# I. set up
# I.a. vector for which observations we want the average effect
# iot_t is the vector with weights in the average treatment effects
# iot_c is the vector of indicators for potential controls

    if (All==1)
      {
        iot.t <- weight;
        iot.c <- as.matrix(rep(1,length(Tr)))
      } else {
        iot.t <- Tr*weight;
        iot.c <- 1-Tr    
      }

# I.b. determine sample and covariate vector sizes
    N  <- nrow(X)
    Kx <- ncol(X)
    Kz <- ncol(Z)

# K covariates
# N observations
#    Nt <- sum(Tr)
#    Nc <- sum(1-Tr)
#    on <- as.matrix(rep(1,N))

# I.c. normalize regressors to have mean zero and unit variance.
# If the standard deviation of a variable is zero, its normalization
# leads to a variable with all zeros.
# The matrix AA enables one to transform the user supplied weight matrix 
# to take account of this transformation.  BUT THIS IS NOT USED!!
# Mu_X and Sig_X keep track of the original mean and variances
#    AA    <- diag(Kx)
    Mu.X  <- matrix(0, Kx, 1)
    Sig.X <- matrix(0, Kx, 1)

    for (k in 1:Kx)
      {
        Mu.X[k,1] <- sum(X[,k]*weight)/sum(weight)
        eps <- X[,k]-Mu.X[k,1]
        Sig.X[k,1] <- sqrt(max(ccc, sum(X[,k]*X[,k]*weight))/sum(weight)-Mu.X[k,1]^2)
        Sig.X[k,1] <- Sig.X[k,1]*sqrt(N/(N-1))
        if(Sig.X[k,1] < ccc)
          Sig.X[k,1] <- ccc        
        X[,k]=eps/Sig.X[k,1]
#        AA[k,k]=Sig.X[k,1]
      } #end of k loop

#    Nv <- nrow(V)
    Mv <- ncol(V)
    Mu.V  <- matrix(0, Mv, 1)
    Sig.V <- matrix(0, Mv, 1)

    for (j in 1:Mv)
      {
        Mu.V[j,1]= ( t(V[,j])%*%weight ) /sum(weight)
#        dv <- V[,j]-Mu.V[j,1]
        sv <- sum(V[,j]*V[,j]*weight)/sum(weight)-Mu.V[j,1]^2
        if (sv > 0)
          {
            sv <- sqrt(sv)
          } else {
            sv <- 0
          }
        sv <- sv * sqrt(N/(N-1))
        Sig.V[j,1] <- sv
      } #end of j loop

# I.d. define weight matrix for metric, taking into account normalization of
# regressors.
# If the eigenvalues of the regressors are too close to zero the Mahalanobis metric
# is not used and we revert back to the default inverse of variances.
    if (Weight==1)
      {
        Weight.matrix=diag(Kx)
      } else if (Weight==2) {
        if (min (eigen( t(X)%*%X/N, only.values=TRUE)$values) > 0.0000001)
          {
            Weight.matrix= solve(t(X)%*%X/N)
          } else {
            Weight.matrix <- diag(Kx)
          }
      }
      # DO NOT RESCALE THE Weight.matrix!!
      #else if (Weight==3)
      #  {
      #    Weight.matrix <- AA %*% Weight.matrix %*% AA
      #  }

#    if (exact==1)
#      {
#        Weight.matrix <- cbind(Weight.matrix, matrix(0,nrow=Kx,ncol=Mv))
#        Weight.matrix <- rbind(Weight.matrix, cbind(matrix(0,nrow=Mv,ncol=Kx),
#                               1000*solve(diag(as.vector(Sig.V*Sig.V), nrow=length(Sig.V)))))
#        Weight.matrix <- as.matrix(Weight.matrix)
#        X <- cbind(X,V)
#        Mu.X  <- rbind(Mu.X, matrix(0, nrow(Mu.V), 1))
#        Sig.X <- rbind(Sig.X, matrix(1, nrow(Sig.V), 1))
#      } #end of exact

    if ( min(eigen(Weight.matrix, only.values=TRUE)$values) < ccc )
      Weight.matrix <- Weight.matrix + diag(Kx)*ccc

    ww <- chol(Weight.matrix) # so that ww*ww=w.m

    if(is.null(s1$ecaliper))
      {
        caliperflag <- 0
        use.ecaliper <- 0
      } else {
        caliperflag <- 1
        use.ecaliper <- s1$ecaliper
      }

    #if we have a diagonal matrix we can void cblas_dgemm
    if (Kx > 1)
      {
        DiagWeightMatrixFlag <- as.double(sum( (Weight.matrix!=diag(diag(Weight.matrix))) )==0)
      } else {
        DiagWeightMatrixFlag <- 1
      }

    if(is.null(MatchLoopC.indx))
      {
    #indx:
    # 1] I (unadjusted); 2] IM (unadjusted); 3] weight; 4] I (adjusted); 5] IM (adjusted)
        if(weights.flag==TRUE)
          {
            MatchLoopC.indx <- MatchLoopC(N=s1$N, xvars=Kx, All=s1$All, M=s1$M,
                                          cdd=cdd, caliperflag=caliperflag, replace=replace, ties=ties,
                                          ww=ww, Tr=s1$Tr, Xmod=s1$X,
                                          weights=weight,
                                          CaliperVec=use.ecaliper, Xorig=X.orig,
                                          restrict.trigger=restrict.trigger, restrict=restrict,
                                          DiagWeightMatrixFlag=DiagWeightMatrixFlag)
          } else {
            MatchLoopC.indx <- MatchLoopCfast(N=s1$N, xvars=Kx, All=s1$All, M=s1$M,
                                              cdd=cdd, caliperflag=caliperflag, replace=replace, ties=ties,
                                              ww=ww, Tr=s1$Tr, Xmod=s1$X,
                                              CaliperVec=use.ecaliper, Xorig=X.orig,
                                              restrict.trigger=restrict.trigger, restrict=restrict,
                                              DiagWeightMatrixFlag=DiagWeightMatrixFlag)
          }
      }

    indx <- MatchLoopC.indx

    if(indx[1,1]==0)
      {
        ret <- list()
        ret$valid <- 0
        if (caliperflag)
          {
            ret$sum.caliper.drops <- indx[1,6]
          } else {
            ret$sum.caliper.drops <- 0
          }
        return(ret)
      } 
    #we now keep going if we only have 1 valid match
    #else if (nrow(indx)< 2)
    #  {
    #    ret <- list()
    #    ret$valid <- 1
    #    if (caliperflag)
    #      {
    #        ret$sum.caliper.drops <- indx[1,6]
    #      } else {
    #        ret$sum.caliper.drops <- 0
    #      }
    #    return(ret)
    #  }
        
    if (All==2)
      {
        foo <- indx[,5]
        indx[,5] <- indx[,4]
        indx[,4] <- foo
      }

    if (caliperflag)
      {
        sum.caliper.drops <- indx[1,6]
      } else {
        sum.caliper.drops <- 0
      }

    #
    # Generate variables which we need later on
    #

    I <- indx[,1]
    IT <- Tr[indx[,1]]
    IM <- indx[,2]

#    IX <- X[indx[,1],]
#    Xt <- X[indx[,4],]
#    Xc <- X[indx[,5],]

#    IY <- Y[indx[,1]]
    Yt <- Y[indx[,4]]
    Yc <- Y[indx[,5]]

    W <- indx[,3]
    
    if(BiasAdj==1 & sum(W) < ncol(Z))
      {
        warning("Fewer (weighted) matches than variables in 'Z': BiasAdjust set to FALSE")
        BiasAdj=0
      }    

    if(BiasAdj==1)
      {
        if(sum(W) < ncol(Z))
          {
            warning("Fewer matches than variables for Bias Adjustment")
          }
        
        IZ <- Z[indx[,1],]
        Zt <- Z[indx[,4],]
        Zc <- Z[indx[,5],]
      }

    est.func <- function(N, All, Tr, indx, weight, BiasAdj, Kz)
      {
        Kcount <- as.matrix(rep(0, N))    
        KKcount <- as.matrix(rep(0, N))
        YCAUS <- matrix(0, nrow=N, ncol=1)

        if (BiasAdj==1)
          {
            ZCAUS <- matrix(0, nrow=N, ncol=Kz)
          } else {
            ZCAUS <- NULL
          }
        
        for (i in 1:N)
          {
            if ( ( Tr[i]==1 & All!=1) | All==1 )
              {

                foo.indx <- indx[,1]==i
                foo.indx2 <- foo.indx
                
                sum.foo <- sum(foo.indx)
                
                if (sum.foo < 1)
                  next;            
                
                foo <- rep(FALSE, N)
                foo.indx <- indx[foo.indx,2]
                foo[foo.indx] <- rep(TRUE,sum.foo)

                #   inner.func <- function(N, weight, indx, foo.indx2, Y, Tr, foo)
  
                Kcount <- Kcount + weight[i] * weight*foo/sum(foo*weight)
                
                KKcount <- KKcount + weight[i]*weight*weight*foo/
                  (sum(foo*weight)*sum(foo*weight))
                
                foo.indx2.2 <- indx[foo.indx2,2];
                foo.indx2.3 <- indx[foo.indx2,3];
                sum.foo.indx2.3 <- sum(foo.indx2.3)

                if(Tr[i]==1)
                  {
                    YCAUS[i] <- Y[i] - sum((Y[foo.indx2.2]*foo.indx2.3))/sum.foo.indx2.3
                  } else {
                    YCAUS[i] <- sum((Y[foo.indx2.2]*foo.indx2.3))/sum.foo.indx2.3 - Y[i]
                  }

                if (BiasAdj==1)
                  {
                    
                    if(Tr[i]==1)
                      {
                        if (sum.foo > 1)
                          {
                            ZCAUS[i,] <-
                              Z[i,] - t(Z[foo.indx2.2,]) %*% foo.indx2.3/sum.foo.indx2.3
                          } else {
                            ZCAUS[i,] <-
                              Z[i,] - Z[foo.indx2.2,]*foo.indx2.3/sum.foo.indx2.3
                          }
                      } else {
                        if (sum.foo > 1)
                          {
                            ZCAUS[i,] <- t(Z[foo.indx2.2,]) %*% foo.indx2.3/sum.foo.indx2.3 - Z[i,]
                          } else {
                            ZCAUS[i,] <- Z[foo.indx2.2,]*foo.indx2.3/sum.foo.indx2.3 - Z[i,]
                          }
                      }
                  } #endof BiasAdj
              }
          } #end of if
        return(list(YCAUS=YCAUS,ZCAUS=ZCAUS,Kcount=Kcount,KKcount=KKcount))
      } #end of est.func
    
    if(version=="standard" & BiasAdj==0)
      {
        ret <- .Call("EstFuncC", as.integer(N), as.integer(All), as.integer(nrow(indx)),
                     as.double(Y), as.double(Tr),
                     as.double(weight), as.double(indx),
                     PACKAGE="Matching")
        YCAUS <- ret[,1];
        Kcount <- ret[,2];
        KKcount <- ret[,3];
      } else if (version=="standard") {
        ret.est <- est.func(N=N, All=All, Tr=Tr, indx=indx, weight=weight, BiasAdj=BiasAdj, Kz=Kz)
        YCAUS <- ret.est$YCAUS
        ZCAUS <- ret.est$ZCAUS
        Kcount <- ret.est$Kcount
        KKcount <- ret.est$KKcount
      }

    if (All!=1)
      {
        I  <- as.matrix(I[IT==1])
        IT <- as.matrix(IT[IT==1])
        Yc <- as.matrix(Yc[IT==1])
        Yt <- as.matrix(Yt[IT==1])
        W  <- as.matrix(W[IT==1])
        if (BiasAdj==1)
          {
            if (Kz > 1)
              {
                Zc <- as.matrix(Zc[IT==1,])
                Zt <- as.matrix(Zt[IT==1,])
                IZ <- as.matrix(IZ[IT==1,])
              } else{
                Zc <- as.matrix(Zc[IT==1])
                Zt <- as.matrix(Zt[IT==1])
                IZ <- as.matrix(IZ[IT==1])            
              }
          }
#        IM <- as.matrix(IM[IT==1,])
#        IY <- as.matrix(IY[IT==1])
#        IX.u  <- as.matrix(IX.u[IT==1,])
#        Xc.u  <- as.matrix(Xc.u[IT==1,])
#        Xt.u  <- as.matrix(Xt.u[IT==1,])
#        Xc    <- as.matrix(Xc[IT==1,])
#        Xt <- as.matrix(Xt[IT==1,])
      } #end of if
    
    if (length(I) < 1)
      {
        return(list(sum.caliper.drops=N))
      }
    
    if (BiasAdj==1)
      {
        # III. Regression of outcome on covariates for matches
        if (All==1)
          {
            # number of observations
            NNt <- nrow(Z)
            # add intercept        
            ZZt <- cbind(rep(1, NNt), Z)
            # number of covariates
            Kx <- ncol(ZZt)
            xw <- ZZt*(sqrt(Tr*Kcount) %*% t(as.matrix(rep(1,Kx))))
            
            foo <- min(eigen(t(xw)%*%xw, only.values=TRUE)$values)
            foo <- as.double(foo<=ccc)
            foo2 <- apply(xw, 2, sd)

            options(show.error.messages = FALSE)
            wout <- NULL
            try(wout <- solve( t(xw) %*% xw + diag(Kx) * ccc * (foo) * max(foo2)) %*%
                (t(xw) %*% (Y*sqrt(Tr*Kcount))))
            if(is.null(wout))
              {
                wout2 <- NULL
                try(wout2 <- ginv( t(xw) %*% xw + diag(Kx) * ccc * (foo) * max(foo2)) %*%
                    (t(xw) %*% (Y*sqrt(Tr*Kcount))))
                if(!is.null(wout2))
                  {
                    wout <-wout2
                    warning("using generalized inverse to calculate Bias Adjustment probably because of singular 'Z'")
                  }
              }
            options(show.error.messages = TRUE)
            if(is.null(wout))
              {
                warning("unable to calculate Bias Adjustment probably because of singular 'Z'")
                BiasAdj <- 0
              } else {
                NW <- nrow(wout)
#                KW <- ncol(wout)
                Alphat <- wout[2:NW,1]
              }
          } else {
            Alphat <- matrix(0, nrow=Kz, ncol=1)
          } #end if ALL
      }

    if(BiasAdj==1)
      {
        # III.b.  Controls
        NNc <- nrow(Z)
        ZZc <- cbind(matrix(1, nrow=NNc, ncol=1),Z)
        Kx <- ncol(ZZc)
        
        xw <- ZZc*(sqrt((1-Tr)*Kcount) %*% matrix(1, nrow=1, ncol=Kx))
        
        foo <- min(eigen(t(xw)%*%xw, only.values=TRUE)$values)
        foo <- as.double(foo<=ccc)
        foo2 <- apply(xw, 2, sd)

        options(show.error.messages = FALSE)
        wout <- NULL        
        try(wout <- solve( t(xw) %*% xw + diag(Kx) * ccc * (foo) * max(foo2)) %*%
            (t(xw) %*% (Y*sqrt((1-Tr)*Kcount))))
        if(is.null(wout))
          {
            wout2 <- NULL
            try(wout2 <- ginv( t(xw) %*% xw + diag(Kx) * ccc * (foo) * max(foo2)) %*%
                (t(xw) %*% (Y*sqrt((1-Tr)*Kcount))))
            if(!is.null(wout2))
              {
                wout <-wout2
                warning("using generalized inverse to calculate Bias Adjustment probably because of singular 'Z'")
              }
          }        
        options(show.error.messages = TRUE)
        if(is.null(wout))
          {
            warning("unable to calculate Bias Adjustment probably because of singular 'Z'")
            BiasAdj <- 0
          } else {
            NW <- nrow(wout)
#            KW <- ncol(wout)
            Alphac <- as.matrix(wout[2:NW,1])
          }
      }

    if(BiasAdj==1)
      {
        # III.c. adjust matched outcomes using regression adjustment for bias adjusted matching estimator
        IZ <- as.matrix(IZ)
        Zc <- as.matrix(Zc)
        Zt <- as.matrix(Zt)
        Alphat <- as.matrix(Alphat)

        SCAUS <- YCAUS-Tr*(ZCAUS %*% Alphac)-(1-Tr)*(ZCAUS %*% Alphat)
        # adjusted control outcome
        Yc.adj <- Yc+BiasAdj * (IZ-Zc) %*% Alphac
        # adjusted treated outcome
        Yt.adj <- Yt+BiasAdj*(IZ-Zt) %*% Alphat
        Tau.i <- Yt.adj - Yc.adj
      } else {
        Yc.adj <- Yc
        Yt.adj <- Yt
        Tau.i <- Yt.adj - Yc.adj        
      }
       
    art.data <- cbind(I,IM)

    # III. If conditional variance is needed, initialize variance vector
    # and loop through all observations

    if (Var.calc>0)
      {
        #For R version of this function see Matching version < 4.5-0008
        Sigs <- VarCalcMatchC(N=N, xvars=ncol(X), Var.calc=Var.calc,
                              cdd=cdd, caliperflag=caliperflag, 
                              ww=ww, Tr=Tr, Xmod=s1$X,
                              CaliperVec=use.ecaliper, Xorig=X.orig,
                              restrict.trigger=restrict.trigger, restrict=restrict,
                              DiagWeightMatrixFlag=DiagWeightMatrixFlag,
                              Y=Y, weightFlag=weights.flag, weight=weight)            
      } #end of var.calc > 0

    est <- t(W) %*% Tau.i/sum(W) # matching estimator

    if(version=="standard")
      {
        if (Var.calc==0)
          {
            eps <- Tau.i - as.double(est)
            eps.sq <- eps*eps
            Sigs <- 0.5 * matrix(1, N, 1) %*% (t(eps.sq) %*% W)/sum(W) #sss <- sqrt(Sigs[1,1])
          } #end of Var.calc==0
        
        SN <- sum(iot.t)
        var.sample <- sum((Sigs*(iot.t+iot.c*Kcount)*(iot.t+iot.c*Kcount))/(SN*SN))
        
        if (All==1)
          {
            var.pop <- sum((Sigs*(iot.c*Kcount*Kcount+2*iot.c*Kcount-iot.c*KKcount))/(SN*SN))
          } else {
            var.pop=sum((Sigs*(iot.c*Kcount*Kcount-iot.c*KKcount))/(SN*SN))
          }
        
        if (BiasAdj==1)
          {
            dvar.pop <- sum(iot.t*(SCAUS-as.double(est))*(SCAUS-as.double(est)))/(SN*SN)
          } else {
            dvar.pop <- sum(iot.t*(YCAUS-as.double(est))*(YCAUS-as.double(est)))/(SN*SN)
          }
        
        var.pop <- var.pop + dvar.pop
        
        if (SAMPLE==1)
          {
            var <- var.sample
          } else {
            #var <- max(var.sample, var.pop)
            var <- var.pop
          }
        
        var.cond <- max(var.sample,var.pop)-var.sample
        se <- sqrt(var)
        se.cond <- sqrt(var.cond)
        #Sig <- sqrt(Sigs)        
      } else {
        se=NULL
        se.cond=NULL
      }
        
    if (All==2)
      est <- -1*est

#    if (exact==1)
#      {
#        Vt.u <- Xt.u[,(Kx-Mv+1):Kx]
#        Vc.u <- Xc.u[,(Kx-Mv+1):Kx]
#        Vdif <- abs(Vt.u-Vc.u)
#
#        if (Mv>1)
#          Vdif <- as.matrix(apply(t(Vdif), 2, sum))
#
#        em[1,1] <- length(Vdif)
#        em[2,1] <- sum(Vdif>0.000001)
#      }#end of exact==1

    if(!MatchbyAI)
      {
        return(list(est=est, se=se, se.cond=se.cond, W=W,
                    sum.caliper.drops=sum.caliper.drops,
                    art.data=art.data, 
                    MatchLoopC=MatchLoopC.indx))
      } else {
        if(Var.calc==0)
          Sigs <- NULL
        return(list(est=est, se=se, se.cond=se.cond, W=W,
                    sum.caliper.drops=sum.caliper.drops,
                    art.data=art.data, 
                    MatchLoopC=MatchLoopC.indx,
                    YCAUS=YCAUS, Kcount=Kcount, KKcount=KKcount,
                    Sigs=Sigs))
      }
  }# end of RmatchLoop

MatchLoopC <- function(N, xvars, All, M, cdd, caliperflag, replace, ties, ww, Tr, Xmod, weights, CaliperVec, Xorig,
                       restrict.trigger, restrict, DiagWeightMatrixFlag)
  {

    if(restrict.trigger)
      {
        restrict.nrow <- nrow(restrict)
      } else {
        restrict.nrow <- 0
      }    

    ret <- .Call("MatchLoopC", as.integer(N), as.integer(xvars), as.integer(All), as.integer(M),
                 as.double(cdd), as.integer(caliperflag), as.integer(replace), as.integer(ties),
                 as.double(ww), as.double(Tr),
                 as.double(Xmod), as.double(weights), as.double(CaliperVec), as.double(Xorig),
                 as.integer(restrict.trigger), as.integer(restrict.nrow), as.double(restrict),
                 as.double(DiagWeightMatrixFlag),
                 PACKAGE="Matching")
    return(ret)
  } #end of MatchLoopC

MatchLoopCfast <- function(N, xvars, All, M, cdd, caliperflag, replace, ties, ww, Tr, Xmod, CaliperVec,
                           Xorig, restrict.trigger, restrict, DiagWeightMatrixFlag)
  {

    if(restrict.trigger)
      {
        restrict.nrow <- nrow(restrict)
      } else {
        restrict.nrow <- 0
      }    
    
    ret <- .Call("MatchLoopCfast", as.integer(N), as.integer(xvars), as.integer(All), as.integer(M),
                 as.double(cdd), as.integer(caliperflag), as.integer(replace), as.integer(ties),
                 as.double(ww), as.double(Tr),
                 as.double(Xmod), as.double(CaliperVec), as.double(Xorig),
                 as.integer(restrict.trigger), as.integer(restrict.nrow), as.double(restrict),
                 as.double(DiagWeightMatrixFlag),
                 PACKAGE="Matching")
    return(ret)
  } #end of MatchLoopCfast

VarCalcMatchC <- function(N, xvars, Var.calc, cdd, caliperflag, ww, Tr, Xmod, CaliperVec,
                          Xorig, restrict.trigger, restrict, DiagWeightMatrixFlag,
                          Y, weightFlag, weight)
  {

    if(restrict.trigger)
      {
        restrict.nrow <- nrow(restrict)
      } else {
        restrict.nrow <- 0
      }    
    
    ret <- .Call("VarCalcMatchC", as.integer(N), as.integer(xvars), as.integer(Var.calc),
                 as.double(cdd), as.integer(caliperflag), 
                 as.double(ww), as.double(Tr),
                 as.double(Xmod), as.double(CaliperVec), as.double(Xorig),
                 as.integer(restrict.trigger), as.integer(restrict.nrow), as.double(restrict),
                 as.double(DiagWeightMatrixFlag),
                 as.double(Y),
                 as.integer(weightFlag), as.double(weight),
                 PACKAGE="Matching")
    return(ret)
  } #end of VarCalcMatchC



.onAttach <- function( ... )
{
  MatchLib <- dirname(system.file(package = "Matching"))
  version <- packageDescription("Matching", lib.loc = MatchLib)$Version
  BuildDate <- packageDescription("Matching", lib.loc = MatchLib)$Date

  foo <- paste("## \n##  Matching (Version ", version, ", Build Date: ", BuildDate, ")\n", 
               "##  See http://sekhon.berkeley.edu/matching for additional documentation.\n",
               "##  Please cite software as:\n",
               "##   Jasjeet S. Sekhon. 2011. ``Multivariate and Propensity Score Matching\n",
               "##   Software with Automated Balance Optimization: The Matching package for R.''\n",
               "##   Journal of Statistical Software, 42(7): 1-52. \n##\n",
               sep = "")
  packageStartupMessage(foo)
}

test_that("Tests GenMatch", {
suppressMessages(library(rgenoud))
suppressMessages(library(Matching))
suppressWarnings(RNGversion("3.5.3"))

set.seed(3101)
data(lalonde)
attach(lalonde)

#The covariates we want to match on
X = cbind(age, educ, black, hisp, married, nodegr, u74, u75, re75, re74)

#The covariates we want to obtain balance on
BalanceMat <- cbind(age, educ, black, hisp, married, nodegr, u74, u75, re75, re74,
                    I(re74*re75))

#
#Let's call GenMatch() to find the optimal weight to give each
#covariate in 'X' so as we have achieved balance on the covariates in
#'BalanceMat'. This is only an example so we want GenMatch to be quick
#so the population size has been set to be only 16 via the 'pop.size'
#option. This is *WAY* too small for actual problems.
#For details see http://sekhon.berkeley.edu/papers/MatchingJSS.pdf.
#
genout <- GenMatch(Tr=treat, X=X, BalanceMatrix=BalanceMat, estimand="ATE", M=1,
                   pop.size=16, max.generations=10, wait.generations=1,
                   unif.seed=3392, int.seed=8282, print.level=0)
print(genout)

#The outcome variable
Y=re78/1000

#
# Now that GenMatch() has found the optimal weights, let's estimate
# our causal effect of interest using those weights
#
mout <- Match(Y=Y, Tr=treat, X=X, estimand="ATE", Weight.matrix=genout)
summary(mout)

#
#Let's determine if balance has actually been obtained on the variables of interest
#
mb <- MatchBalance(treat~age +educ+black+ hisp+ married+ nodegr+ u74+ u75+
                     re75+ re74+ I(re74*re75),
                   match.out=mout, nboots=500)

expect_equal(nrow(genout$Weight.matrix), 10)
expect_equal(ncol(genout$Weight.matrix), 10)

# For more examples see: http://sekhon.berkeley.edu/matching/R.
before_sdiff <- unlist(lapply(mb$BeforeMatching, function(x){return(x$sdiff)}))
after_sdfiff <- unlist(lapply(mb$AfterMatching, function(x){return(x$sdiff)}))

results <- data.frame(before = before_sdiff,
                      after = after_sdfiff)

results_saved <- data.frame(before = c(10.655038555334192,  12.806026640590549,   4.476700682905446,
                                       -20.340738543987911,   8.999512178599344, -27.750943560007322,
                                       -9.189507193658958, -17.225298606861006, 8.236275506237831,
                                       -0.234370781139797,  -2.779905583506529),
                            after = c(0.698105986659637, -1.579765862957684, -0.609524006759037,
                                      0.000000000000000,  0.000000000000000,  0.000000000000000,
                                      -0.507141556044142, -0.939497016216188,  2.263685815459904,
                                      -0.280105002856819, -2.639651943919278))

# Skip the exact tests if not on MacOS
skip_if_not_mac()

# Check the sdiff before matching
expect_equal(all.equal(results[,1], results_saved[,1]), TRUE)
# Check the sdiff after matching
expect_equal(all.equal(results[,2], results_saved[,2]), TRUE)


})

#This function returns the homogeneous AI SE.
#This standalone function only works for ATT, and no weights
#The Match() function includes the version which works for everything
#  and which calculates heterogeneous variance estimates

AIse <- function(Y, Tr, index.treated, index.control, weights, est=NULL)
  {
    N <- length(Y)

    if(is.null(est))
      {
        est <- sum(Y[index.treated]*weights)/sum(weights)-
          sum(Y[index.control]*weights)/sum(weights)
      } else{
        est <- as.double(est)
      }

#    ret <- est.func(N=N, All=0, Tr=Tr,
#                    indx=cbind(index.treated,index.control,weights),
#                    weight=rep(1,N),
#                    BiasAdj=FALSE, Kz=NULL)

    ret <- .Call("EstFuncC", as.integer(N), as.integer(0), as.integer(length(index.treated)),
                 as.double(Y), as.double(Tr),
                 as.double(rep(1,N)), as.double(cbind(index.treated,index.control,weights)),
                 PACKAGE="Matching")

#    YCAUS <- ret$YCAUS
#    Kcount <- ret$Kcount
#    KKcount <- ret$KKcount

    YCAUS <- ret[,1]
    Kcount <- ret[,2]
    KKcount <- ret[,3]    

    Yt <- Y[index.treated]
    Yc <- Y[index.control]
    Tau <- Yt - Yc

    eps <- Tau - est
    eps.sq <- eps*eps
    Sigs <- 0.5 * matrix(1, N, 1) %*% (t(eps.sq) %*% weights)/sum(weights)

    SN <- sum(Tr)
    var.pop=sum((Sigs*((1-Tr)*Kcount*Kcount-(1-Tr)*KKcount))/(SN*SN))
    
    dvar.pop <- sum(Tr*(YCAUS-est)*(YCAUS-est))/(SN*SN)
    
    var <- var.pop + dvar.pop
    
    se <- sqrt(var)
    return(se)
  }

Matchby <- function(Y=NULL, Tr, X, by, estimand="ATT", M=1, ties=FALSE, replace=TRUE,
                    exact = NULL, caliper = NULL,
                    AI=FALSE, Var.calc=0,
                    Weight = 1, Weight.matrix = NULL, 
                    distance.tolerance = 1e-05, tolerance = sqrt(.Machine$double.eps), 
                    print.level=1, version="Matchby", ...)
  {

    #index for raw obs
    Tr <- as.double(Tr)    
    nobs  <- length(Tr)
    orig.treated.nobs  <- sum(Tr==1)

    if(is.null(Y))
      {
        Y <- rep(0, nobs)
      } else {
        Y  <- as.double(Y)
        if (nobs != length(Y))
          {
            stop("length(Tr) != length(Y)")
          }
      }
    
    X  <- as.matrix(X)

    if( nobs != nrow(X))
      {
        stop("length(Tr) != nrow(X)")
      }        

    index.nobs <- 1:nobs
    t.index.nobs <- split(index.nobs, by, drop=TRUE)

    nindx  <- length(t.index.nobs)

    # *OUTPUT* observation weights
    weights  <- NULL

    index.treated <- NULL
    index.control <- NULL

    if (Var.calc < 0)
      {
        warning("User set 'Var.calc' to less than 0.  Resetting to the default which is 0.")
        Var.calc <- 0
      }

    if(Var.calc > 0)
      AI <- TRUE

    if(AI & estimand!="ATT")
      {
        warning("This function can only calculate Abadie-Imbens SEs for 'ATT'.  For AI SEs and other estimands, please use 'Match()'. Setting 'AI=FALSE'")
        AI <- FALSE
        Var.calc <- 0
      }
    if(AI & ties!=TRUE)
      {
        ties=TRUE
        warning("Abadie-Imbens SEs have been requested. Setting 'ties=TRUE'")
      }
    if(AI & replace!=TRUE)
      {
        replace=TRUE
        warning("Abadie-Imbens SEs have been requested. Setting 'replace=TRUE'")
      }
    if(AI)
      version <- "MatchbyAI"

    if (replace!=FALSE & replace!=TRUE)
      {
        warning("'replace' must be TRUE or FALSE.  Setting to TRUE")
        replace <- TRUE
      }
    if(replace==FALSE)
      {
        ties <- FALSE        
      }    
    if (ties!=FALSE & ties!=TRUE)
      {
        warning("'ties' must be TRUE or FALSE.  Setting to TRUE")
        ties <- TRUE
      }

    if(AI)
      {
        Kcount <- as.matrix(rep(0, nobs))    
        KKcount <- as.matrix(rep(0, nobs))
        YCAUS <- matrix(0, nrow=nobs, ncol=1)

        if(Var.calc>0)
          Sigs <- matrix(0, nrow=nobs, ncol=1)
      }

    for (i in 1:nindx)
      {
        if(print.level > 0)
          cat(i,"of", nindx, "groups\n")

        tmp.index.nobs <- t.index.nobs[[i]]
        
        f.Tr  <- Tr[tmp.index.nobs]
        #need at least 1 Tr and 1 Co
        if(length(f.Tr) < 2)
          {
            next;
          }
        #drop if we only have Tr or Co
        if( var(f.Tr)==0 )
          {
            next;
          }

        t1  <- Match(Y=Y[tmp.index.nobs], Tr=f.Tr, X=X[tmp.index.nobs,],
                     estimand=estimand, M=M,
                     Var.calc=Var.calc,
                     exact=exact, caliper=caliper, replace=replace, ties=ties,
                     Weight=Weight, Weight.matrix=Weight.matrix,
                     tolerance=tolerance, distance.tolerance=distance.tolerance,
                     version=version, ...)
        
        if(is.na(t1[1]))
          {
            if(!is.null(exact) | !is.null(caliper))
              {
                warning("no matches found in group ",i," (probably because of the exact or caliper option) continuing")
              } else {
                warning("no matches found in group ",i," continuing")}
            next
          }

        weights  <- c(weights, t1$weights)

        index.treated <- c(index.treated,(tmp.index.nobs)[t1$index.treated])
        index.control <- c(index.control,(tmp.index.nobs)[t1$index.control])

        if(AI)
          {
            YCAUS[tmp.index.nobs]   <- t1$YCAUS
            Kcount[tmp.index.nobs]  <- t1$Kcount
            KKcount[tmp.index.nobs] <- t1$KKcount

            if(Var.calc>0)
              {
                if(is.null(t1$Sigs))
                  {
                    pfoo <- paste("Var.calc=",Var.calc," cannot be calculated (group ",i,").  Var.calc is probably set to a number larger than the possible number of matches in a subgroup",sep="")
                    stop(pfoo)
                  } else {
                    Sigs[tmp.index.nobs]    <- t1$Sigs
                  }
              }#if(Var.calc>0)
          }#if(AI)
      }#i:nindx

    if(is.null(index.treated))
      {
        warning("no valid matches were found")
        z <- NA
        class(z)  <- "Matchby"    
        return(z)
      }

    Yt  <- Y[index.treated]
    Yc  <- Y[index.control]
    
    sum.tw  <- sum(weights)
    est  <- sum(Yt*weights)/sum.tw  - sum(Yc*weights)/sum.tw
    Tau  <- Yt - Yc
    varest  <- sum( ((Tau-est)^2)*weights)/(sum.tw*sum.tw)
    se.standard  <- sqrt(varest)
    
    ret  <- list(est=est, se=NULL, se.standard=se.standard,
                 index.treated=index.treated,
                 index.control=index.control,
                 weights=weights)
    ret$orig.nobs  <- nobs
    ret$orig.wnobs <- nobs
    ret$nobs  <- length(Yt)
    ret$wnobs  <- sum.tw
    ret$orig.treated.nobs  <- orig.treated.nobs
    ret$ndrops  <- orig.treated.nobs-ret$wnobs
    ret$estimand  <- estimand
    ret$version  <- version
    class(ret)  <- "Matchby"
    
    if(AI)
      {
        if(Var.calc==0)
          {
            eps <- Tau - est
            eps.sq <- eps*eps
            Sigs <- 0.5 * matrix(1, nobs, 1) %*% (t(eps.sq) %*% weights)/sum(weights)
          }
        
        SN <- orig.treated.nobs
        var.pop=sum((Sigs*((1-Tr)*Kcount*Kcount-(1-Tr)*KKcount))/(SN*SN))
        
        dvar.pop <- sum(Tr*(YCAUS-est)*(YCAUS-est))/(SN*SN)
        
        var <- var.pop + dvar.pop
        
        se <- sqrt(var)
        ret$se <- se
      }
    
    invisible(return(ret))
  }
  

summary.Matchby  <- function(object, ..., digits=5)
  {
    if(!is.list(object)) {
      warning("'Matchby' object contains less than two valid matches.  Cannot proceed.")
      return(invisible(NULL))
    }
    
    if (!inherits(object, "Matchby")) {
      warning("Object not of class 'Matchby'")
      return(invisible(NULL))
    }

    cat("\n")
    cat("Estimate... ",format(object$est,digits=digits),"\n")

    if(!is.null(object$se))
      {
        cat("AI SE...... ",format(object$se,digits=digits),"\n")
        cat("AI T-stat.. ",format(object$est/object$se,digits=digits),"\n")
        cat("AI p.val... ",format.pval((1-pnorm(abs(object$est/object$se)))*2,digits=digits),"\n\n")
      }
    cat("SE......... ",format(object$se.standard,digits=digits),"\n")
    cat("T-stat..... ",format(object$est/object$se.standard,digits=digits),"\n")
    cat("p.val...... ",format.pval((1-pnorm(abs(object$est/object$se.standard)))*2,digits=digits),"\n")
    cat("\n")        

    cat("Original number of observations.............. ", object$orig.nobs,"\n")
    if(object$estimand!="ATC")
      {
        cat("Original number of treated obs............... ", object$orig.treated.nobs,"\n")
      } else {
        cat("Original number of control obs............... ",
            object$orig.nobs- object$orig.treated.nobs,"\n")        
      }
    cat("Matched number of observations............... ", round(object$wnobs, 3),"\n")
    cat("Matched number of observations  (unweighted). ", object$nobs,"\n")
    cat("\n")
        
    if ( object$ndrops > 0)
        {
          cat("Number of treated observations dropped....... ", object$ndrops, "\n")
          cat("\n")
        }

    z <- list()
    class(z) <- "summary.Matchby"
    return(invisible(z))
  } #end of summary.Matchby

print.summary.Matchby <- function(x, ...)
  {
    invisible(NULL)
  }

suppressMessages(library(rgenoud))
suppressMessages(library(Matching))
suppressWarnings(RNGversion("3.5.3"))

set.seed(3101)
data(lalonde)
attach(lalonde)

#The covariates we want to match on
X = cbind(age, educ, black, hisp, married, nodegr, u74, u75, re75, re74)

#The covariates we want to obtain balance on
BalanceMat <- cbind(age, educ, black, hisp, married, nodegr, u74, u75, re75, re74,
                    I(re74*re75))

#
#Let's call GenMatch() to find the optimal weight to give each
#covariate in 'X' so as we have achieved balance on the covariates in
#'BalanceMat'. This is only an example so we want GenMatch to be quick
#so the population size has been set to be only 16 via the 'pop.size'
#option. This is *WAY* too small for actual problems.
#For details see http://sekhon.berkeley.edu/papers/MatchingJSS.pdf.
#
genout <- GenMatch(Tr=treat, X=X, BalanceMatrix=BalanceMat, estimand="ATE", M=1,
                   pop.size=16, max.generations=10, wait.generations=1,
                   unif.seed=3392, int.seed=8282, print.level=0)
print(genout)

#The outcome variable
Y=re78/1000

#
# Now that GenMatch() has found the optimal weights, let's estimate
# our causal effect of interest using those weights
#
mout <- Match(Y=Y, Tr=treat, X=X, estimand="ATE", Weight.matrix=genout)
summary(mout)

#                        
#Let's determine if balance has actually been obtained on the variables of interest
#                        
mb <- MatchBalance(treat~age +educ+black+ hisp+ married+ nodegr+ u74+ u75+
                   re75+ re74+ I(re74*re75),
                   match.out=mout, nboots=500)

# For more examples see: http://sekhon.berkeley.edu/matching/R.

MATCHpt <- function(q, df, ...)
  {
    #work around for strange windows-R bug found by example(GenMatch) when MatchBalance is hit.
    #don't know how general it is so let's try to work around it
    ret = suppressWarnings(pt(q,df, ...))
    
    if (is.na(ret)) {
      ret   <- pt(q, df, ...)
      if(is.na(ret))
        warning("pt() generated NaN. q:",q," df:",df,"\n")
    }
    
    return(ret)
  } #end of MATCHpt

Mt.test.pvalue  <- function(Tr, Co, weights)
  {
    v1  <- Tr-Co
    estimate  <- sum(v1*weights)/sum(weights)
    var1  <- sum( ((v1-estimate)^2)*weights )/( sum(weights)*sum(weights) )

    if (estimate==0 & var1==0)
      {
        return(1)
      }
    
    statistic  <- estimate/sqrt(var1)
#    p.value    <- (1-pnorm(abs(statistic)))*2
    p.value    <- (1-MATCHpt(abs(statistic), df=sum(weights)-1))*2

    return(p.value)
  } #end of Mt.test.pvalue

Mt.test.unpaired.pvalue  <- function(Tr, Co, weights)
  {
    obs <- sum(weights)
    
    mean.Tr <- sum(Tr*weights)/obs
    mean.Co <- sum(Co*weights)/obs
    estimate <- mean.Tr-mean.Co
    var.Tr  <- sum( ( (Tr - mean.Tr)^2 )*weights)/(obs-1)
    var.Co  <- sum( ( (Co - mean.Co)^2 )*weights)/(obs-1)
    dim <- sqrt(var.Tr/obs + var.Co/obs)

    if (estimate==0 & dim==0)
      {
        return(1)
      }
    
    statistic  <- estimate/dim

    a1 <- var.Tr/obs
    a2 <- var.Co/obs
    dof <- ((a1 + a2)^2)/( (a1^2)/(obs - 1) + (a2^2)/(obs - 1) )    
    p.value    <- (1-MATCHpt(abs(statistic), df=dof))*2    

    return(p.value)
  } #end of Mt.test.unpaired.pvalue

ks.fast <- function(x, y, n.x, n.y, n)
  {
    w <- c(x, y)
    z <- cumsum(ifelse(order(w) <= n.x, 1/n.x, -1/n.y))
    z <- z[c(which(diff(sort(w)) != 0), n.x + n.y)]        
    
    return( max(abs(z)) )
  } #ks.fast

qqstats <- function (x, y, standardize = TRUE, summary.func=NULL)
{
  if (standardize) {
    vals <- sort(unique(c(x, y)))
    x <- ecdf(x)
    sx <- x(vals)
    y <- ecdf(y)
    sy <- y(vals)
  }  else {
    sx <- sort(x)
    sy <- sort(y)
    lenx <- length(sx)
    leny <- length(sy)
    
    if (leny < lenx) 
      sx <- approx(1:lenx, sx, n = leny, method = "constant")$y
    if (leny > lenx) 
      sy <- approx(1:leny, sy, n = lenx, method = "constant")$y
  }
  diffxy <- abs(sx-sy)
  meandiff <- mean(diffxy)
  maxdiff <- max(diffxy)  
  mediandiff <- median(diffxy)
  if(!is.null(summary.func))
    {
      summarydiff <- summary.func(diffxy)
      
      return(list(meandiff=meandiff, mediandiff=mediandiff, maxdiff=maxdiff,
                  summarydiff=summarydiff, summary.func=summary.func))      
    } else {
      return(list(meandiff=meandiff, mediandiff=mediandiff, maxdiff=maxdiff))
    }
}#qqstats

GenBalanceQQ <- function(rr, X, summarystat="mean", summaryfunc="mean")
  {
    index.treated <- rr[,1]
    index.control <- rr[,2]
    
    nvars <- ncol(X)
    qqsummary   <- c(rep(NA,nvars))
    
    for (i in 1:nvars)
      {    
        
        qqfoo <- qqstats(X[,i][index.treated], X[,i][index.control], standardize=TRUE)
        
        if (summarystat=="median")
          {
            qqsummary[i] <- qqfoo$mediandiff
          } else if (summarystat=="max") {
            qqsummary[i] <- qqfoo$maxdiff    
          } else {
            qqsummary[i] <- qqfoo$meandiff
          }
        
      } #end of for loop
    
    
    if (summaryfunc=="median")
      {
        return(median(qqsummary))
      } else if (summaryfunc=="max")  {
        return(sort(qqsummary, decreasing=TRUE))
      } else if (summaryfunc=="sort")  {
        return(sort(qqsummary, decreasing=TRUE))
      } else {
        return(mean(qqsummary))
      }    
  } #end of GenBalanceQQ

GenBalance <-
  function(rr, X, nvars=ncol(X), nboots = 0, ks=TRUE, verbose = FALSE, paired=TRUE)
  {
    index.treated <- rr[,1]
    index.control <- rr[,2]
    weights <- rr[,3]

    tol  <- sqrt(.Machine$double.eps)
    storage.t <- c(rep(9,nvars))
    storage.k <- c(rep(9,nvars))
    fs.ks     <- matrix(nrow=nvars, ncol=1)
    s.ks      <- matrix(nrow=nvars, ncol=1)  
    bbcount   <- matrix(0, nrow=nvars, ncol=1)
    dummy.indx  <- matrix(0, nrow=nvars, ncol=1)
    
    w  <- c(X[,1][index.treated], X[,1][index.control])
    obs <- length(w)
    n.x  <- length(X[,1][index.treated])
    n.y  <- length(X[,1][index.control])
    cutp <- n.x
    w  <- matrix(nrow=obs, ncol=nvars)

    for (i in 1:nvars)
      {
        w[,i] <- c(X[,i][index.treated], X[,i][index.control])

        if(paired)
          {
            t.out <- Mt.test.pvalue(X[,i][index.treated],
                                    X[,i][index.control],
                                    weights = weights)
          } else {
            t.out <- Mt.test.unpaired.pvalue(X[,i][index.treated],
                                             X[,i][index.control],
                                             weights = weights)
          }
        
        storage.t[i] <- t.out            
        
                                        #      print(length(unique(X[,i])) < 3)

        
        dummy.indx[i]  <- length(unique(X[,i])) < 3

        if (!dummy.indx[i] & ks & nboots > 9)
          {
            fs.ks[i]  <- ks.fast(X[,i][index.treated], X[,i][index.control],
                                 n.x=n.x, n.y=n.y, n=obs)
          } else if(!dummy.indx[i] & ks)
            {
                                        #storage.k[i] <- ks.test(X[,i][index.treated], X[,i][index.control])$p.value
              
              storage.k[i] <- Mks.test(X[,i][index.treated], X[,i][index.control])$p.value

            }
      }#end of i loop

    
    if (ks & nboots > 9)
      {
        for (b in 1:nboots)
          {
            sindx  <- sample(1:obs, obs, replace = TRUE)
            
            for (i in 1:nvars)
              {
                
                if (dummy.indx[i])
                  next;
                
                X1tmp <- w[sindx[1:cutp],i ]
                X2tmp <- w[sindx[(cutp + 1):obs], i]
                s.ks[i] <- ks.fast(X1tmp, X2tmp, n.x=n.x, n.y=n.y, n=obs)
                if (s.ks[i] >= (fs.ks[i] - tol) )
                  bbcount[i]  <-  bbcount[i] + 1
              }#end of i loop
          } #end of b loop
        
        for (i in 1:nvars)
          {
            
            if (dummy.indx[i])
              {
                storage.k[i]  <- 9
                next;
              }
            
            storage.k[i]  <- bbcount[i]/nboots
            
          }
        storage.k[storage.k==9]=storage.t[storage.k==9]
        output <- c(storage.t, storage.k)
      } else if(ks){
        storage.k[storage.k==9]=storage.t[storage.k==9]                
        output <- c(storage.t, storage.k)
      } else {
        output <- storage.t
      }

    if(sum(is.na(output)) > 0) {
      output[is.na(output)] = 2
      warning("output has NaNs")
    }
    
    if (verbose == TRUE)
      {
        cat("\n")
        for (i in 1:nvars)
          {
            cat("\n", i, " t-test p-val  =", storage.t[i], "\n" )
            if(ks)
              cat(" ", i, "  ks-test p-val = ", storage.k[i], " \n",sep="")
          }
        cat("\nsorted return vector:\n", sort(output), "\n")
        cat("number of return values:", length(output), "\n")
      }

    return(output)
  } #end of GenBalance


#
# writing fast KS test
#

KSbootBalanceSummary <- function(index.treated, index.control, X, 
                                 nboots = 1000)
{
  X <- as.matrix(X)
  nvars <- ncol(X)

  tol  <- sqrt(.Machine$double.eps)
  storage.k <- c(rep(NA,nvars))
  storage.k.naive <- c(rep(NA,nvars))  
  fs.ks     <- matrix(nrow=nvars, ncol=1)
  s.ks      <- matrix(nrow=nvars, ncol=1)  
  bbcount   <- matrix(0, nrow=nvars, ncol=1)
  dummy.indx  <- matrix(0, nrow=nvars, ncol=1)
  
  w  <- c(X[,1][index.treated], X[,1][index.control])
  obs <- length(w)
  n.x  <- length(X[,1][index.treated])
  n.y  <- length(X[,1][index.control])
  cutp <- n.x
  w  <- matrix(nrow=obs, ncol=nvars)

  for (i in 1:nvars)
    {
      w[,i] <- c(X[,i][index.treated], X[,i][index.control])
      
      dummy.indx[i]  <- length(unique(X[,i])) < 3

      if (!dummy.indx[i])
        {
          foo <- Mks.test(X[,i][index.treated], X[,i][index.control])
          fs.ks[i] <- foo$statistic
          storage.k.naive[i] <- foo$p.value
        } 
    }#end of i loop

  for (b in 1:nboots)
    {
      sindx  <- sample(1:obs, obs, replace = TRUE)
      
      for (i in 1:nvars)
        {
          
          if (dummy.indx[i])
            next;
          
          X1tmp <- w[sindx[1:cutp],i ]
          X2tmp <- w[sindx[(cutp + 1):obs], i]
          s.ks[i] <- ks.fast(X1tmp, X2tmp, n.x=n.x, n.y=n.y, n=obs)
          if (s.ks[i] >= (fs.ks[i] - tol) )
            bbcount[i]  <-  bbcount[i] + 1
        }#end of i loop
    } #end of b loop

  for (i in 1:nvars)
    {
      if (dummy.indx[i])
        {
          storage.k[i]  <- NA
          next;
        }
      
      storage.k[i]  <- bbcount[i]/nboots
    }

  ret = list(ks.boot.pval=storage.k, ks.naive.pval=storage.k.naive, ks.stat=fs.ks)

  return(ret)
} #end of KSbootBalanceSummary

suppressMessages(library(Matching))
suppressWarnings(RNGversion("3.5.3"))

data(lalonde)

X  <- cbind(lalonde$black, lalonde$age, lalonde$educ)
Y  <- lalonde$re78
Tr  <- lalonde$treat

rr2  <- Matchby(Y=Y, Tr=Tr, X=X, M=1, exact=TRUE, by=X[,1],
                ties=TRUE, replace=TRUE, AI=TRUE)
summary(rr2)

rr  <- Match(Y=Y, Tr=Tr, X=X, M=1, exact=TRUE)
summary(rr, full=TRUE)

abs(rr$est-rr2$est) < 1e-10
abs(rr$se-rr2$se) < 1e-10
abs(rr$se.standard-rr2$se.standard) < 1e-10

is_mac <- function() {
  return(grepl('darwin', R.Version()$platform, ignore.case = TRUE) && 
           (grepl('x86', R.Version()$platform, ignore.case = TRUE) || grepl('aarch', R.Version()$platform, ignore.case = TRUE)))
}

skip_if_not_mac <- function() {
  if(is_mac()) {
    return(invisible(TRUE))
  }

  skip('Not run on non-mac platforms due to non-deterministic randomness in implementation of std::uniform_int_distribution across different compilers. Test cases were written only for macOS (clang). We have not had time to write tests for other compilers like gcc yet.')
}

FastMatchC <- function(N, xvars, All, M, cdd, ww, Tr, Xmod, weights)
  {
    ret <- .Call("FastMatchC", as.integer(N), as.integer(xvars), as.integer(All), as.integer(M),
                 as.double(cdd), as.double(ww), as.double(Tr),
                 as.double(Xmod), as.double(weights),
                 PACKAGE="Matching")
    return(ret)
  }

MatchGenoudStage1 <- function(Tr=Tr, X=X, All=All, M=M, weights=weights,
                              tolerance)
  {
    N  <- nrow(X)
    xvars <- ncol(X)
    
# if SATC is to be estimated the treatment indicator is reversed    
    if (All==2)
      Tr <- 1-Tr

# check on the number of matches, to make sure the number is within the limits
# feasible given the number of observations in both groups.
    if (All==1)
      {
        M <- min(M,min(sum(Tr),sum(1-Tr)));        
      } else {
        M <- min(M,sum(1-Tr));
      }

# I.c. normalize regressors to have mean zero and unit variance.
# If the standard deviation of a variable is zero, its normalization
# leads to a variable with all zeros.
    Mu.X  <- matrix(0, xvars, 1)
    Sig.X <- matrix(0, xvars, 1)

    weights.sum <- sum(weights)

    for (k in 1:xvars)
      {
        Mu.X[k,1] <- sum(X[,k]*weights)/weights.sum;
        eps <- X[,k]-Mu.X[k,1]
        Sig.X[k,1] <- sqrt(max(tolerance, sum(X[,k]*X[,k]*weights))/weights.sum-Mu.X[k,1]^2)
        Sig.X[k,1] <- Sig.X[k,1]*sqrt(N/(N-1))
        if(Sig.X[k,1] < tolerance)
          Sig.X[k,1] <- tolerance
        X[,k]=eps/Sig.X[k,1]
      } #end of k loop

    ret <- list(Tr=Tr, X=X, All=All, M=M, N=N)
    return(ret)
  } #end of MatchGenoudStage1


###############################################################################
## For Caliper!
##
###############################################################################

MatchGenoudStage1caliper <- function(Tr=Tr, X=X, All=All, M=M, weights=weights,
                                     exact=exact, caliper=caliper,
                                     distance.tolerance, tolerance)
  {
    N  <- nrow(X)
    xvars <- ncol(X)
    weights.orig  <- as.matrix(weights)

    if (!is.null(exact))
      {
        exact = as.vector(exact)
        nexacts = length(exact)
        if ( (nexacts > 1) & (nexacts != xvars) )
          {
            warning("length of exact != ncol(X). Ignoring exact option")
            exact <- NULL
          } else if (nexacts==1 & (xvars > 1) ){
            exact <- rep(exact, xvars)
          }
      }

    if (!is.null(caliper))
      {
        caliper = as.vector(caliper)
        ncalipers = length(caliper)
        if ( (ncalipers > 1) & (ncalipers != xvars) )
          {
            warning("length of caliper != ncol(X). Ignoring caliper option")
            caliper <- NULL
          } else if (ncalipers==1 & (xvars > 1) ){
            caliper <- rep(caliper, xvars)
          }
      }

    if (!is.null(caliper))
      {
        ecaliper <- vector(mode="numeric", length=xvars)
        sweights  <- sum(weights.orig)
        for (i in 1:xvars)
          {
            meanX  <- sum( X[,i]*weights.orig )/sweights
            sdX  <- sqrt(sum( (X[,i]-meanX)^2 )/sweights)
            ecaliper[i]  <- caliper[i]*sdX
          }
      } else {
        ecaliper <- NULL
      }

    if (!is.null(exact))
      {
        if(is.null(caliper))
          {
            max.diff <- abs(max(X)-min(X) + distance.tolerance * 100)
            ecaliper <- matrix(max.diff, nrow=xvars, ncol=1)
          }
        
        for (i in 1:xvars)
          {
            if (exact[i])
              ecaliper[i] <- distance.tolerance;
          }
      }        
    
# if SATC is to be estimated the treatment indicator is reversed    
    if (All==2)
      Tr <- 1-Tr

# check on the number of matches, to make sure the number is within the limits
# feasible given the number of observations in both groups.
    if (All==1)
      {
        M <- min(M,min(sum(Tr),sum(1-Tr)));        
      } else {
        M <- min(M,sum(1-Tr));
      }

# I.c. normalize regressors to have mean zero and unit variance.
# If the standard deviation of a variable is zero, its normalization
# leads to a variable with all zeros.
    Mu.X  <- matrix(0, xvars, 1)
    Sig.X <- matrix(0, xvars, 1)

    weights.sum <- sum(weights)

    for (k in 1:xvars)
      {
        Mu.X[k,1] <- sum(X[,k]*weights)/weights.sum;
        eps <- X[,k]-Mu.X[k,1]
        Sig.X[k,1] <- sqrt(max(tolerance, sum(X[,k]*X[,k]*weights))/weights.sum-Mu.X[k,1]^2)
        Sig.X[k,1] <- Sig.X[k,1]*sqrt(N/(N-1))
        if(Sig.X[k,1] < tolerance)
          Sig.X[k,1] <- tolerance        
        X[,k]=eps/Sig.X[k,1]
      } #end of k loop

    ret <- list(Tr=Tr, X=X, All=All, M=M, N=N, ecaliper=ecaliper)
    return(ret)
  } #end of MatchGenoudStage1caliper


###############################################################################
## GenMatch
##
###############################################################################

GenMatch <- function(Tr, X, BalanceMatrix=X, estimand="ATT", M=1,
                     weights=NULL,
                     pop.size = 100, max.generations=100,
                     wait.generations=4, hard.generation.limit=FALSE,
                     starting.values=rep(1,ncol(X)),
                     fit.func="pvals",
                     MemoryMatrix=TRUE,
                     exact=NULL, caliper=NULL, replace=TRUE, ties=TRUE,
                     CommonSupport=FALSE,nboots=0, ks=TRUE, verbose=FALSE,
                     distance.tolerance=0.00001,
                     tolerance=sqrt(.Machine$double.eps),
                     min.weight=0,
                     max.weight=1000,
                     Domains=NULL,
                     print.level=2,
                     project.path=NULL,
                     paired=TRUE,
                     loss=1,
                     data.type.integer=FALSE,
                     restrict=NULL,
                     cluster=FALSE,
                     balance=TRUE, ...)
  {

    requireNamespace("rgenoud")

    Tr <- as.double(Tr)
    X  <- as.matrix(X)
    BalanceMatrix  <- as.matrix(BalanceMatrix)

    if(length(Tr) != nrow(X))
      {
        stop("length(Tr) != nrow(X)")
      }
    if(!is.function(fit.func))
      {
        if(nrow(BalanceMatrix) != length(Tr))
          {
            stop("nrow(BalanceMatrix) != length(Tr)")
          }
      }

    if (is.null(weights))
      {
        weights <- rep(1,length(Tr))
        weights.flag <- FALSE
      } else {
        weights.flag <- TRUE
        weights <- as.double(weights)
        if( length(Tr) != length(weights))
          {
            stop("length(Tr) != length(weights)")
          }            
      }

    isna  <- sum(is.na(Tr)) + sum(is.na(X)) + sum(is.na(weights)) + sum(is.na(BalanceMatrix))
    if (isna!=0)
      {
        stop("GenMatch(): input includes NAs")
        return(invisible(NULL))
      }    

    #check inputs
    if (sum(Tr !=1 & Tr !=0) > 0) {
      stop("Treatment indicator must be a logical variable---i.e., TRUE (1) or FALSE (0)")
    }
    if (var(Tr)==0) {
      stop("Treatment indicator ('Tr') must contain both treatment and control observations")
    }    
    if (distance.tolerance < 0)
      {
        warning("User set 'distance.tolerance' to less than 0.  Resetting to the default which is 0.00001.")
        distance.tolerance <- 0.00001
      }
    #CommonSupport
    if (CommonSupport !=1 & CommonSupport !=0) {
      stop("'CommonSupport' must be a logical variable---i.e., TRUE (1) or FALSE (0)")
    }    
    if(CommonSupport==TRUE)
      {
        tr.min <- min(X[Tr==1,1])
        tr.max <- max(X[Tr==1,1])

        co.min <- min(X[Tr==0,1])
        co.max <- max(X[Tr==0,1])

        if(tr.min >= co.min)
          {
            indx1 <- X[,1] < (tr.min-distance.tolerance)
          } else {
            indx1 <- X[,1] < (co.min-distance.tolerance)            
          }

        if(co.max <= tr.max)
          {        
            indx2 <- X[,1] > (co.max+distance.tolerance)
          } else {
            indx2 <- X[,1] > (tr.max+distance.tolerance)            
          }

        indx3 <- indx1==0 & indx2==0
        Tr <- as.double(Tr[indx3])
        X  <- as.matrix(X[indx3,])
        BalanceMatrix <- as.matrix(BalanceMatrix[indx3,])
        weights <- as.double(weights[indx3])
      }#end of CommonSupport
    
    if (pop.size < 0 | pop.size!=round(pop.size) )
      {
        warning("User set 'pop.size' to an illegal value.  Resetting to the default which is 100.")
        pop.size <- 100
      }
    if (max.generations < 0 | max.generations!=round(max.generations) )
      {
        warning("User set 'max.generations' to an illegal value.  Resetting to the default which is 100.")
        max.generations <-100
      }
    if (wait.generations < 0 | wait.generations!=round(wait.generations) )
      {
        warning("User set 'wait.generations' to an illegal value.  Resetting to the default which is 4.")
        wait.generations <- 4
      }
    if (hard.generation.limit != 0 & hard.generation.limit !=1 )
      {
        warning("User set 'hard.generation.limit' to an illegal value.  Resetting to the default which is FALSE.")
        hard.generation.limit <- FALSE
      }
    if (data.type.integer != 0 & data.type.integer !=1 )
      {
        warning("User set 'data.type.integer' to an illegal value.  Resetting to the default which is TRUE.")
        data.type.integer <- TRUE
      }
    if (MemoryMatrix != 0 & MemoryMatrix !=1 )
      {
        warning("User set 'MemoryMatrix' to an illegal value.  Resetting to the default which is TRUE.")
        MemoryMatrix <- TRUE
      }                
    if (nboots < 0 | nboots!=round(nboots) )
      {
        warning("User set 'nboots' to an illegal value.  Resetting to the default which is 0.")
        nboots <- 0
      }
    if (ks != 0 & ks !=1 )
      {
        warning("User set 'ks' to an illegal value.  Resetting to the default which is TRUE.")
        ks <- TRUE
      }
    if (verbose != 0 & verbose !=1 )
      {
        warning("User set 'verbose' to an illegal value.  Resetting to the default which is FALSE.")
        verbose <- FALSE
      }
    if (min.weight < 0)
      {
        warning("User set 'min.weight' to an illegal value.  Resetting to the default which is 0.")
        min.weight <- 0
      }
    if (max.weight < 0)
      {
        warning("User set 'max.weight' to an illegal value.  Resetting to the default which is 1000.")
        max.weight <- 1000
      }
    if (print.level != 0 & print.level !=1 & print.level !=2 & print.level !=3)
      {
        warning("User set 'print.level' to an illegal value.  Resetting to the default which is 2.")
        print.level <- 2
      }    
    if (paired != 0 & paired !=1 )
      {
        warning("User set 'paired' to an illegal value.  Resetting to the default which is TRUE.")
        paired <- FALSE
      }    
    ##from Match()
    if (tolerance < 0)
      {
        warning("User set 'tolerance' to less than 0.  Resetting to the default which is 0.00001.")
        tolerance <- 0.00001
      }
    if (M < 1)
      {
        warning("User set 'M' to less than 1.  Resetting to the default which is 1.")
        M <- 1
      }
    if ( M!=round(M) )
      {
        warning("User set 'M' to an illegal value.  Resetting to the default which is 1.")
        M <- 1        
      }
    if (replace!=FALSE & replace!=TRUE)
      {
        warning("'replace' must be TRUE or FALSE.  Setting to TRUE")
        replace <- TRUE
      }
    if(replace==FALSE)
      ties <- FALSE
    if (ties!=FALSE & ties!=TRUE)
      {
        warning("'ties' must be TRUE or FALSE.  Setting to TRUE")
        ties <- TRUE
      }

    #print warning if pop.size, max.generations and wait.generations are all set to their original values
    if(pop.size==100 & max.generations==100 & wait.generations==4)
      {
        warning("The key tuning parameters for optimization were are all left at their default values.  The 'pop.size' option in particular should probably be increased for optimal results.  For details please see the help page and http://sekhon.berkeley.edu/papers/MatchingJSS.pdf")
      }

    #loss function
    if (is.double(loss))
      {
        if (loss==1)  {
          loss.func=sort
          lexical=ncol(BalanceMatrix)
          if(ks)
            lexical=lexical+lexical
        } else if(loss==2) {
          loss.func=min
          lexical=0
        } else{
          stop("unknown loss function")
        }
      } else if (is.function(loss)) {
        loss.func=loss
        lexical=1
      } else {
        stop("unknown loss function")
      }

    #set lexical for fit.func
    if (is.function(fit.func))
      {
        lexical = 1
      } else if (fit.func=="qqmean.max" | fit.func=="qqmedian.max" | fit.func=="qqmax.max")   {
        lexical=ncol(BalanceMatrix)
      } else if (fit.func!="qqmean.mean" & fit.func!="qqmean.max" &
                 fit.func!="qqmedian.median" & fit.func!="qqmedian.max"
                 & fit.func!="pvals") {
        stop("invalid 'fit.func' argument")
      } else  if (!fit.func=="pvals") {
        lexical = 0
      }

    if(replace==FALSE)
      {
        #replace==FALE, needs enough observation
        #ATT
        orig.weighted.control.nobs <- sum(weights[Tr!=1])
        orig.weighted.treated.nobs <- sum(weights[Tr==1])
        if(estimand=="ATC")
          {
            if (orig.weighted.treated.nobs < orig.weighted.control.nobs)
              {
                warning("replace==FALSE, but there are more (weighted) control obs than treated obs.  Some obs will be dropped.  You may want to estimate ATC instead")
              }
          } else if(estimand=="ATE") 
            {
              #ATE
              if (orig.weighted.treated.nobs > orig.weighted.control.nobs)
                {
                  warning("replace==FALSE, but there are more (weighted) treated obs than control obs.  Some treated obs will not be matched.  You may want to estimate ATC instead.")
                }              
              if (orig.weighted.treated.nobs < orig.weighted.control.nobs)
                {
                  warning("replace==FALSE, but there are more (weighted) control obs than treated obs.  Some control obs will not be matched.  You may want to estimate ATT instead.")
                }
            } else {
              #ATT
              if (orig.weighted.treated.nobs > orig.weighted.control.nobs)
                {
                  warning("replace==FALSE, but there are more (weighted) treated obs than control obs.  Some treated obs will not be matched.  You may want to estimate ATC instead.")
                }
            }
        
        #we need a restrict matrix if we are going to not do replacement
        if(is.null(restrict))
          {
            restrict <- t(as.matrix(c(0,0,0)))
          }
      }#end of replace==FALSE    

    

    #check the restrict matrix input
    if(!is.null(restrict))
      {
        if(!is.matrix(restrict))
          stop("'restrict' must be a matrix of restricted observations rows and three columns: c(i,j restriction)")

        if(ncol(restrict)!=3 )
          stop("'restrict' must be a matrix of restricted observations rows and three columns: c(i,j restriction)")

        restrict.trigger <- TRUE
      }  else {
        restrict.trigger <- FALSE
      }

    if(!is.null(caliper) | !is.null(exact) | restrict.trigger | !ties)
      {
        GenMatchCaliper.trigger <- TRUE
      } else {
        GenMatchCaliper.trigger <- FALSE
      }

    isunix  <- .Platform$OS.type=="unix"
    if (is.null(project.path))
      {
        if (print.level < 3 & isunix)
          {
            project.path="/dev/null"
          } else {
            project.path=paste(tempdir(),"/genoud.pro",sep="")
            
            #work around for rgenoud bug
            #if (print.level==3)
            #print.level <- 2
          }
      } 
    
    nvars <- ncol(X)
    balancevars <- ncol(BalanceMatrix)

    if (is.null(Domains))
      {
        Domains <- matrix(min.weight, nrow=nvars, ncol=2)
        Domains[,2] <- max.weight
      } else {
        indx <- (starting.values < Domains[,1]) | (starting.values > Domains[,2])
        starting.values[indx] <- round( (Domains[indx,1]+Domains[indx,2])/2 )
      }

    # create All
    if (estimand=="ATT")
      {
        All  <- 0
      } else if(estimand=="ATE") {
        All  <- 1
      } else if(estimand=="ATC") {
        All  <- 2
      } else {
        All  <- 0
        warning("User set 'estimand' to an illegal value.  Resetting to the default which is 'ATT'")
      }

    #stage 1 Match, only needs to be called once    
    if(!GenMatchCaliper.trigger)
      {

        s1 <- MatchGenoudStage1(Tr=Tr, X=X, All=All, M=M, weights=weights,
                                tolerance=tolerance);
        s1.Tr <- s1$Tr
        s1.X <- s1$X
        s1.All <- s1$All
        s1.M <- s1$M
        s1.N <- s1$N
        rm(s1)
      } else {
        s1 <- MatchGenoudStage1caliper(Tr=Tr, X=X, All=All, M=M, weights=weights,
                                       exact=exact, caliper=caliper,
                                       distance.tolerance=distance.tolerance,
                                       tolerance=tolerance)
        s1.Tr <- s1$Tr
        s1.X <- s1$X
        s1.All <- s1$All
        s1.M <- s1$M
        s1.N <- s1$N
        s1.ecaliper  <- s1$ecaliper
        
        if (is.null(s1.ecaliper))
          {
            caliperFlag  <- 0
            Xorig  <- 0
            CaliperVec  <- 0
          } else {
            caliperFlag  <- 1
            Xorig  <- X
            CaliperVec  <- s1$ecaliper
          }
        rm(s1)        
      } #GenMatchCaliper.trigger

    genoudfunc  <- function(x)
      {
        wmatrix <- diag(x, nrow=nvars)
        if ( min(eigen(wmatrix, symmetric=TRUE, only.values=TRUE)$values) < tolerance )
            wmatrix <- wmatrix + diag(nvars)*tolerance
        
        ww <- chol(wmatrix)

        if(!GenMatchCaliper.trigger)
          {
            
            if (weights.flag==TRUE)
              {
                FastMatchC.internal <- function(N, xvars, All, M, cdd, ww, Tr, Xmod, weights)
                  {
                    ret <- .Call("FastMatchC", as.integer(N), as.integer(xvars), as.integer(All), as.integer(M),
                                 as.double(cdd), as.double(ww), as.double(Tr),
                                 as.double(Xmod), as.double(weights),
                                 PACKAGE="Matching")
                    return(ret)
                  }
                
                rr <- FastMatchC.internal(N=s1.N, xvars=nvars, All=s1.All, M=s1.M,
                                          cdd=distance.tolerance, ww=ww, Tr=s1.Tr, Xmod=s1.X,
                                          weights=weights)
              } else {
                FasterMatchC.internal <- function(N, xvars, All, M, cdd, ww, Tr, Xmod, weights)
                  {
                    ret <- .Call("FasterMatchC", as.integer(N), as.integer(xvars), as.integer(All), as.integer(M),
                                 as.double(cdd), as.double(ww), as.double(Tr),
                                 as.double(Xmod), 
                                 PACKAGE="Matching")
                    return(ret)
                  }
                
                rr <- FasterMatchC.internal(N=s1.N, xvars=nvars, All=s1.All, M=s1.M,
                                             cdd=distance.tolerance, ww=ww, Tr=s1.Tr, Xmod=s1.X)
              } #end of weights.flag
          } else {
            if (weights.flag==TRUE)
              {
                MatchLoopC.internal <- function(N, xvars, All, M, cdd, caliperflag, replace, ties, ww, Tr, Xmod, weights, CaliperVec,
                                                Xorig, restrict.trigger, restrict)
                  {
                    
                    if(restrict.trigger)
                      {
                        restrict.nrow <- nrow(restrict)
                      } else {
                        restrict.nrow <- 0
                      }    
                    
                    ret <- .Call("MatchLoopC", as.integer(N), as.integer(xvars), as.integer(All), as.integer(M),
                                 as.double(cdd), as.integer(caliperflag), as.integer(replace), as.integer(ties), as.double(ww), as.double(Tr),
                                 as.double(Xmod), as.double(weights), as.double(CaliperVec), as.double(Xorig),
                                 as.integer(restrict.trigger), as.integer(restrict.nrow), as.double(restrict),
                                 #next line is sets the DiagWeightMatrixFlag
                                 as.double(1),                             
                                 PACKAGE="Matching")
                    return(ret)
                  } #end of MatchLoopC.internal
                
                rr <- MatchLoopC.internal(N=s1.N, xvars=nvars, All=s1.All, M=s1.M,
                                          cdd=distance.tolerance,
                                          caliperflag=caliperFlag,
                                          replace=replace, ties=ties,
                                          ww=ww, Tr=s1.Tr, Xmod=s1.X, weights=weights,
                                          CaliperVec=CaliperVec, Xorig=Xorig,
                                          restrict.trigger=restrict.trigger, restrict=restrict)
              } else {
                MatchLoopCfast.internal <- function(N, xvars, All, M, cdd, caliperflag, replace, ties, ww, Tr, Xmod, CaliperVec, Xorig,
                                                    restrict.trigger, restrict)
                  {
                    
                    if(restrict.trigger)
                      {
                        restrict.nrow <- nrow(restrict)
                      } else {
                        restrict.nrow <- 0
                      }    
                    
                    ret <- .Call("MatchLoopCfast", as.integer(N), as.integer(xvars), as.integer(All), as.integer(M),
                                 as.double(cdd), as.integer(caliperflag), as.integer(replace), as.integer(ties), as.double(ww), as.double(Tr),
                                 as.double(Xmod), as.double(CaliperVec), as.double(Xorig),
                                 as.integer(restrict.trigger), as.integer(restrict.nrow), as.double(restrict),
                                 #next line is the DiagWeightMatrixFlag
                                 as.double(1),                              
                                 PACKAGE="Matching")
                    return(ret)
                  } #end of MatchLoopCfast.internal
                
                
                rr <- MatchLoopCfast.internal(N=s1.N, xvars=nvars, All=s1.All, M=s1.M,
                                              cdd=distance.tolerance,
                                              caliperflag=caliperFlag,
                                              replace=replace, ties=ties,
                                              ww=ww, Tr=s1.Tr, Xmod=s1.X, 
                                              CaliperVec=CaliperVec, Xorig=Xorig,
                                              restrict.trigger=restrict.trigger, restrict=restrict)
              } #end of weights.flag
            
            #no matches
            if(rr[1,1]==0) {
              warning("no valid matches found in GenMatch evaluation") 
              return(rep(-9999, balancevars*2))
            }
            
            rr <- rr[,c(4,5,3)]            
          } #Caliper.Trigger

        #should be the same as GenBalance() in GenBalance.R but we need to include it here because of
        #cluster scoping issues.
        GenBalance.internal <-
          function(rr, X, nvars=ncol(X), nboots = 0, ks=TRUE, verbose = FALSE, paired=TRUE)
          {

            #CUT-AND-PASTE from GenBalance.R, the functions before GenBalance. but get rid of warn *switch*
            MATCHpt <- function(q, df, ...)
              {
                #don't know how general it is so let's try to work around it.
                ret=pt(q,df, ...)
                
                if (is.na(ret)) {
                  ret   <- pt(q, df, ...)
                  if(is.na(ret))
                    warning("pt() generated NaN. q:",q," df:",df,"\n",date())
                }
                
                return(ret)
              } #end of MATCHpt

            Mt.test.pvalue  <- function(Tr, Co, weights)
              {
                v1  <- Tr-Co
                estimate  <- sum(v1*weights)/sum(weights)
                var1  <- sum( ((v1-estimate)^2)*weights )/( sum(weights)*sum(weights) )
                
                if (estimate==0 & var1==0)
                  {
                    return(1)
                  }
                
                statistic  <- estimate/sqrt(var1)
                #    p.value    <- (1-pnorm(abs(statistic)))*2
                p.value    <- (1-MATCHpt(abs(statistic), df=sum(weights)-1))*2
                
                return(p.value)
              } #end of Mt.test.pvalue

            Mt.test.unpaired.pvalue  <- function(Tr, Co, weights)
              {
                obs <- sum(weights)
                
                mean.Tr <- sum(Tr*weights)/obs
                mean.Co <- sum(Co*weights)/obs
                estimate <- mean.Tr-mean.Co
                var.Tr  <- sum( ( (Tr - mean.Tr)^2 )*weights)/(obs-1)
                var.Co  <- sum( ( (Co - mean.Co)^2 )*weights)/(obs-1)
                dim <- sqrt(var.Tr/obs + var.Co/obs)
                
                if (estimate==0 & dim==0)
                  {
                    return(1)
                  }
                
                statistic  <- estimate/dim
                
                a1 <- var.Tr/obs
                a2 <- var.Co/obs
                dof <- ((a1 + a2)^2)/( (a1^2)/(obs - 1) + (a2^2)/(obs - 1) )    
                p.value    <- (1-MATCHpt(abs(statistic), df=dof))*2    
                
                return(p.value)
              } #end of Mt.test.unpaired.pvalue

            ks.fast <- function(x, y, n.x, n.y, n)
              {
                w <- c(x, y)
                z <- cumsum(ifelse(order(w) <= n.x, 1/n.x, -1/n.y))
                z <- z[c(which(diff(sort(w)) != 0), n.x + n.y)]        
                
                return( max(abs(z)) )
              } #ks.fast

            index.treated <- rr[,1]
            index.control <- rr[,2]
            weights <- rr[,3]
            
            tol  <- .Machine$double.eps*100  
            storage.t <- c(rep(9,nvars))
            storage.k <- c(rep(9,nvars))
            fs.ks     <- matrix(nrow=nvars, ncol=1)
            s.ks      <- matrix(nrow=nvars, ncol=1)  
            bbcount   <- matrix(0, nrow=nvars, ncol=1)
            dummy.indx  <- matrix(0, nrow=nvars, ncol=1)
            
            w  <- c(X[,1][index.treated], X[,1][index.control])
            obs <- length(w)
            n.x  <- length(X[,1][index.treated])
            n.y  <- length(X[,1][index.control])
            cutp <- round(obs/2)  
            w  <- matrix(nrow=obs, ncol=nvars)
            
            for (i in 1:nvars)
              {
                w[,i] <- c(X[,i][index.treated], X[,i][index.control])
                
                if(paired)
                  {
                    t.out <- Mt.test.pvalue(X[,i][index.treated],
                                            X[,i][index.control],
                                            weights = weights)
                  } else {
                    t.out <- Mt.test.unpaired.pvalue(X[,i][index.treated],
                                                     X[,i][index.control],
                                                     weights = weights)
                  }
                
                storage.t[i] <- t.out            
        
                dummy.indx[i]  <- length(unique(X[,i])) < 3
                
                if (!dummy.indx[i] & ks & nboots > 9)
                  {
                    fs.ks[i]  <- ks.fast(X[,i][index.treated], X[,i][index.control],
                                         n.x=n.x, n.y=n.y, n=obs)
                  } else if(!dummy.indx[i] & ks)
                    {
                      
                      storage.k[i] <- Mks.test(X[,i][index.treated], X[,i][index.control])$p.value

                    }
              }#end of i loop

            
            if (ks & nboots > 9)
              {
                n.x  <- cutp
                n.y  <- obs-cutp
                for (b in 1:nboots)
                  {
                    sindx  <- sample(1:obs, obs, replace = TRUE)
                    
                    for (i in 1:nvars)
                      {
                        
                        if (dummy.indx[i])
                          next;
                        
                        X1tmp <- w[sindx[1:cutp],i ]
                        X2tmp <- w[sindx[(cutp + 1):obs], i]
                        s.ks[i] <- ks.fast(X1tmp, X2tmp, n.x=n.x, n.y=n.y, n=obs)
                        if (s.ks[i] >= (fs.ks[i] - tol) )
                          bbcount[i]  <-  bbcount[i] + 1
                      }#end of i loop
                  } #end of b loop
                
                for (i in 1:nvars)
                  {
                    
                    if (dummy.indx[i])
                      {
                        storage.k[i]  <- 9
                        next;
                      }
                    
                    storage.k[i]  <- bbcount[i]/nboots
                    
                  }
                storage.k[storage.k==9]=storage.t[storage.k==9]
                output <- c(storage.t, storage.k)
              } else if(ks){
                storage.k[storage.k==9]=storage.t[storage.k==9]                
                output <- c(storage.t, storage.k)
              } else {
                output <- storage.t
              }
            
            if(sum(is.na(output)) > 0) {
              output[is.na(output)] = 2
              warning("output has NaNs")
            }
            
            if (verbose == TRUE)
              {
                cat("\n")
                for (i in 1:nvars)
                  {
                    cat("\n", i, " t-test p-val  =", storage.t[i], "\n" )
                    if(ks)
                      cat(" ", i, "  ks-test p-val = ", storage.k[i], " \n",sep="")
                  }
                cat("\nsorted return vector:\n", sort(output), "\n")
                cat("number of return values:", length(output), "\n")
              }
            
            return(output)
          } #end of GenBalance.internal

        GenBalanceQQ.internal <- function(rr, X, summarystat="mean", summaryfunc="mean")
          {
            index.treated <- rr[,1]
            index.control <- rr[,2]
    
            nvars <- ncol(X)
            qqsummary   <- c(rep(NA,nvars))
            
            for (i in 1:nvars)
              {    
                
                qqfoo <- qqstats(X[,i][index.treated], X[,i][index.control], standardize=TRUE)
                
                if (summarystat=="median")
                  {
                    qqsummary[i] <- qqfoo$mediandiff
                  } else if (summarystat=="max")  {
                    qqsummary[i] <- qqfoo$maxdiff    
                  } else {
                    qqsummary[i] <- qqfoo$meandiff
                  }
                
              } #end of for loop
            
            
            if (summaryfunc=="median")
              {
                return(median(qqsummary))
              } else if (summaryfunc=="max")  {
                return(sort(qqsummary, decreasing=TRUE))
              } else if (summaryfunc=="sort")  {
                return(sort(qqsummary, decreasing=TRUE))
              } else {
                return(mean(qqsummary))
              }    
          } #end of GenBalanceQQ.internal


        if (is.function(fit.func)) {
          a <- fit.func(rr, BalanceMatrix)
          return(a)
        } else if (fit.func=="pvals")
          {
            a <- GenBalance.internal(rr=rr, X=BalanceMatrix, nvars=balancevars, nboots=nboots,
                                     ks=ks, verbose=verbose, paired=paired)
            a <- loss.func(a)
            return(a)
          } else if (fit.func=="qqmean.mean") {
            a <- GenBalanceQQ.internal(rr=rr, X=BalanceMatrix, summarystat="mean", summaryfunc="mean")
            return(a)
          } else if (fit.func=="qqmean.max") {
            a <- GenBalanceQQ.internal(rr=rr, X=BalanceMatrix, summarystat="mean", summaryfunc="max")
            return(a)
          } else if (fit.func=="qqmax.mean") {
            a <- GenBalanceQQ.internal(rr=rr, X=BalanceMatrix, summarystat="max", summaryfunc="mean")
            return(a)            
          } else if (fit.func=="qqmax.max") {
            a <- GenBalanceQQ.internal(rr=rr, X=BalanceMatrix, summarystat="max", summaryfunc="max")
            return(a)                        
          } else if (fit.func=="qqmedian.median") {
            a <- GenBalanceQQ.internal(rr=rr, X=BalanceMatrix, summarystat="median", summaryfunc="median")
            return(a)
          } else if (fit.func=="qqmedian.max") {
            a <- GenBalanceQQ.internal(rr=rr, X=BalanceMatrix, summarystat="median", summaryfunc="max")
            return(a)
          } 
      } #end genoudfunc

    #cluster info
    clustertrigger=1
    if (is.logical(cluster))
      {
        if (cluster==FALSE)  {
          clustertrigger=0
        } else {
          stop("cluster option must be either FALSE, an object of the 'cluster' class (from the 'parallel' package) or a list of machines so 'genoud' can create such an object")
        }
      }
    
    if(clustertrigger) {
      parallel.exists = requireNamespace("parallel")
      if (!parallel.exists) {
        stop("The 'cluster' feature cannot be used unless the package 'parallel' can be loaded.")
      }
    } 

    if(clustertrigger)
      {

        GENclusterExport <- function (cl, list, envir = .GlobalEnv) 
          {
            gets <- function(n, v) {
              assign(n, v, envir = envir)
              NULL
            }
            for (name in list) {
              parallel::clusterCall(cl, gets, name, get(name))
            }
          }        

        if (class(cluster)[1]=="SOCKcluster" | class(cluster)[1]=="PVMcluster" | class(cluster)[1]=="spawnedMPIcluster" | class(cluster)[1]=="MPIcluster") {
          clustertrigger=1
          cl <- cluster
          cl.genoud <- cl
        } else {
          clustertrigger=2
          cluster <- as.vector(cluster)
          cat("Initializing Cluster\n")
          cl <- parallel::makePSOCKcluster(cluster)
          cl.genoud <- cl
        }      
      } else {
        cl.genoud <- FALSE
      }#end of clustertrigger

    if (clustertrigger > 0)
      {
        #create restrict.summary, because passing the entire restrict matrix is too much
        
        parallel::clusterEvalQ(cl, library("Matching"))
        GENclusterExport(cl, c("s1.N", "s1.All", "s1.M", "s1.Tr", "s1.X", "nvars",
                               "tolerance", "distance.tolerance", "weights",
                               "BalanceMatrix", "balancevars", "nboots", "ks", "verbose", "paired", "loss.func",
                               "fit.func"))

        if(GenMatchCaliper.trigger) {
          GENclusterExport(cl, c("caliperFlag", "CaliperVec", "Xorig", "restrict.trigger", "restrict","replace"))
        }
        
        GENclusterExport(cl, "genoudfunc")
      }

    do.max <- FALSE    
    if(!is.function(fit.func))
      {
        if (fit.func=="pvals") 
          do.max <- TRUE
      }

    rr <- rgenoud::genoud(genoudfunc, nvars=nvars, starting.values=starting.values,
                 pop.size=pop.size, max.generations=max.generations,
                 wait.generations=wait.generations, hard.generation.limit=hard.generation.limit,
                 Domains=Domains,
                 MemoryMatrix=MemoryMatrix,
                 max=do.max, gradient.check=FALSE, data.type.int=data.type.integer,
                 hessian=FALSE,
                 BFGS=FALSE, project.path=project.path, print.level=print.level,
                 lexical=lexical,
                 cluster=cl.genoud,
                 balance=balance,
                 ...)
    
    wmatrix <- diag(rr$par, nrow=nvars)
    
    if ( min(eigen(wmatrix, symmetric=TRUE, only.values=TRUE)$values) < tolerance )
      wmatrix <- wmatrix + diag(nvars)*tolerance
        
    ww <- chol(wmatrix)

    if(!GenMatchCaliper.trigger)
      {
        mout <- FastMatchC(N=s1.N, xvars=nvars, All=s1.All, M=s1.M,
                           cdd=distance.tolerance, ww=ww, Tr=s1.Tr, Xmod=s1.X,
                           weights=weights)
        rr2 <- list(value=rr$value, par=rr$par, Weight.matrix=wmatrix, matches=mout, ecaliper=NULL)
      } else {
        mout <- MatchLoopC(N=s1.N, xvars=nvars, All=s1.All, M=s1.M,
                           cdd=distance.tolerance,
                           caliperflag=caliperFlag,
                           replace=replace, ties=ties,
                           ww=ww, Tr=s1.Tr, Xmod=s1.X, weights=weights,
                           CaliperVec=CaliperVec, Xorig=Xorig,
                           restrict.trigger=restrict.trigger, restrict=restrict,
                           DiagWeightMatrixFlag=1)

        #no matches
        if(mout[1,1]==0) {
          warning("no valid matches found by GenMatch") 
        }        
        
        rr2 <- list(value=rr$value, par=rr$par, Weight.matrix=wmatrix, matches=mout, ecaliper=CaliperVec)
      }

    if (clustertrigger==2)
      parallel::stopCluster(cl)    
    
    class(rr2) <- "GenMatch"
    return(rr2)
  } #end of GenMatch

#
# Replication of Dehejia and Wahba psid3 model
#
# Dehejia, Rajeev and Sadek Wahba. 1999.``Causal Effects in Non-Experimental Studies: Re-Evaluating the
# Evaluation of Training Programs.''Journal of the American Statistical Association 94 (448): 1053-1062.
#

suppressMessages(library(Matching))
suppressWarnings(RNGversion("3.5.3"))

# Replication of Dehejia and Wahba psid3 model.  

# Dehejia, Rajeev and Sadek Wahba. 1999.``Causal Effects in
# Non-Experimental Studies: Re-Evaluating the # Evaluation of Training
# Programs.''Journal of the American Statistical Association 94 (448):
# 1053-1062.  

set.seed(10391)
data(lalonde)

#
# Estimate the propensity model
#
glm1  <- glm(treat~age + I(age^2) + educ + I(educ^2) + black +
             hisp + married + nodegr + re74  + I(re74^2) + re75 + I(re75^2) +
             u74 + u75, family=binomial, data=lalonde)

#
#save data objects
#
X  <- glm1$fitted
Y  <- lalonde$re78
Tr  <- lalonde$treat

#
# one-to-one matching with replacement (the "M=1" option).
# Estimating the treatment effect on the treated (the "estimand" option which defaults ATT).
#
rr  <- Match(Y=Y,Tr=Tr,X=X,M=1);
summary(rr)

#
# Let's check for balance
#
mb  <- MatchBalance(treat~age + I(age^2) + educ + I(educ^2) + black +
                    hisp + married + nodegr + re74  + I(re74^2) + re75 + I(re75^2) +
                    u74 + u75, data=lalonde, match.out=rr, nboots=0)

\name{lalonde}
\alias{lalonde}
\docType{data}
\title{Lalonde Dataset}
\description{
  Dataset used by Dehejia and Wahba (1999) to evaluate propensity score matching.
}
\usage{data(lalonde)}
\format{
  A data frame with 445 observations on the following 12 variables.
  \describe{
    \item{age}{age in years.}
    \item{educ}{years of schooling.}
    \item{black}{indicator variable for blacks.}
    \item{hisp}{indicator variable for Hispanics.}
    \item{married}{indicator variable for martial status.}
    \item{nodegr}{indicator variable for high school diploma.}
    \item{re74}{real earnings in 1974.}
    \item{re75}{real earnings in 1975.}
    \item{re78}{real earnings in 1978.}
    \item{u74}{indicator variable for earnings in 1974 being zero.}
    \item{u75}{indicator variable for earnings in 1975 being zero.}
    \item{treat}{an indicator variable for treatment status.}
  }
}
\details{
  Two demos are provided which use this dataset.  The first,
  \code{DehejiaWahba}, replicates one of the models from Dehejia and
  Wahba (1999).  The second demo, \code{AbadieImbens}, replicates the
  models produced by Abadie and Imbens in their Matlab code. 
  Many of these models are found to produce good balance for the Lalonde
  data.
  }
\references{
  Dehejia, Rajeev and Sadek Wahba. 1999.``Causal Effects in
  Non-Experimental Studies: Re-Evaluating the
  Evaluation of Training Programs.'' \emph{Journal of the American Statistical
  Association} 94 (448): 1053-1062.

  LaLonde, Robert. 1986.  ``Evaluating the Econometric Evaluations of
  Training Programs.''  \emph{American Economic Review} 76:604-620.
} 
  \seealso{ Also see \code{\link{Match}}, \code{\link{GenMatch}},
  \code{\link{MatchBalance}},
   \code{\link{balanceUV}},
   \code{\link{ks.boot}}, \code{\link{GerberGreenImai}} }
\keyword{datasets}

%  LocalWords:  lalonde docType Dataset Dehejia Wahba hisp nodegr dataset url
%  LocalWords:  DehejiaWahba AbadieImbens Abadie Imbens Rajeev Sadek emph
%  LocalWords:  seealso GenMatch MatchBalance balanceUV datasets
%  LocalWords:  GerberGreenImai

\name{GenMatch}
\alias{GenMatch}
\title{Genetic Matching}
\description{
  This function finds optimal balance using multivariate matching where
  a genetic search algorithm determines the weight each covariate is
  given.  Balance is determined by examining cumulative probability
  distribution functions of a variety of standardized statistics.  By
  default, these statistics include t-tests and Kolmogorov-Smirnov
  tests. A variety of descriptive statistics based on empirical-QQ
  (eQQ) plots can also be used or any user provided measure of balance.
  The statistics are not used to conduct formal hypothesis tests,
  because no measure of balance is a monotonic function of bias and
  because balance should be maximized without limit. The object
  returned by \code{GenMatch} can be supplied to the \code{\link{Match}}
  function (via the \code{Weight.matrix} option) to obtain causal
  estimates.  \code{GenMatch} uses \code{\link[rgenoud]{genoud}} to
  perform the genetic search.  Using the \code{cluster} option, one may
  use multiple computers, CPUs or cores to perform parallel
  computations.
}
\usage{
GenMatch(Tr, X, BalanceMatrix=X, estimand="ATT", M=1, weights=NULL,
         pop.size = 100, max.generations=100,
         wait.generations=4, hard.generation.limit=FALSE,
         starting.values=rep(1,ncol(X)),
         fit.func="pvals",
         MemoryMatrix=TRUE,
         exact=NULL, caliper=NULL, replace=TRUE, ties=TRUE,
         CommonSupport=FALSE, nboots=0, ks=TRUE, verbose=FALSE,
         distance.tolerance=1e-05,
         tolerance=sqrt(.Machine$double.eps),
         min.weight=0, max.weight=1000,
         Domains=NULL, print.level=2,
         project.path=NULL,
         paired=TRUE, loss=1,
         data.type.integer=FALSE,
         restrict=NULL,
         cluster=FALSE, balance=TRUE, ...)
}
\arguments{
  \item{Tr}{ A vector indicating the observations which are in the
    treatment regime and those which are not.  This can either be a
    logical vector or a real vector where 0 denotes control and 1 denotes
    treatment.}
  \item{X}{ A matrix containing the variables we wish to match on.
    This matrix may contain the actual observed covariates or the
    propensity score or a combination of both.}
  \item{BalanceMatrix}{ A matrix containing the variables we wish
    to achieve balance on.  This is by default equal to \code{X}, but it can
    in principle be a matrix which contains more or less variables than
    \code{X} or variables which are transformed in various ways.  See
    the examples.}
  \item{estimand}{ A character string for the estimand.  The default
    estimand is "ATT", the sample average treatment effect for the
    treated. "ATE" is the sample average treatment effect, and
    "ATC" is the sample average treatment effect for the controls.}
  \item{M}{A scalar for the number of matches which should be
    found. The default is one-to-one matching. Also see the \code{ties}
    option.}
  \item{weights}{ A vector the same length as \code{Y} which
    provides observation specific weights.}
  \item{pop.size}{Population Size.  This is the number of individuals
    \code{\link[rgenoud]{genoud}} uses to solve the optimization problem.
    The theorems proving that genetic algorithms find good solutions are
    asymptotic in population size.  Therefore, it is important that this value not
    be small.  See \code{\link[rgenoud]{genoud}} for more details.}
  \item{max.generations}{ Maximum Generations.  This is the maximum
    number of generations that \code{\link[rgenoud]{genoud}} will run when
    optimizing.  This is a \emph{soft} limit.  The maximum generation
    limit will be binding only if \code{hard.generation.limit} has been
    set equal to \emph{TRUE}.  Otherwise, \code{wait.generations} controls
    when optimization stops. See \code{\link[rgenoud]{genoud}} for more
    details.}
  \item{wait.generations}{If there is no improvement in the objective
    function in this number of generations, optimization will stop.  The
    other options controlling termination are \code{max.generations} and
    \code{hard.generation.limit}.}
  \item{hard.generation.limit}{ This logical variable determines if the
    \code{max.generations} variable is a binding constraint.  If
    \code{hard.generation.limit} is \emph{FALSE}, then
    the algorithm may exceed the \code{max.generations}
    count if the objective function has improved within a given number of
    generations (determined by \code{wait.generations}).}
  \item{starting.values}{
    This vector's length is equal to the number of variables in \code{X}.  This
    vector contains the starting weights each of the variables is
    given. The \code{starting.values} vector is a way for the user
    to insert \emph{one} individual into the starting population.
    \code{\link[rgenoud]{genoud}} will randomly create the other individuals.  These values
    correspond to the diagonal of the \code{Weight.matrix} as described
    in detail in the \code{\link{Match}} function.}
  \item{fit.func}{The balance metric \code{GenMatch} should optimize.
    The user may choose from the following or provide a function:\cr
    \code{pvals}: maximize the p.values from (paired) t-tests and
    Kolmogorov-Smirnov tests conducted for each column in
    \code{BalanceMatrix}.  Lexical optimization is conducted---see the
    \code{loss} option for details.\cr
    \code{qqmean.mean}: calculate the mean standardized difference in the eQQ
    plot for each variable.  Minimize the mean of these differences across
    variables.\cr
    \code{qqmean.max}:  calculate the mean standardized difference in the eQQ
    plot for each variable.  Minimize the maximum of these differences across
    variables.  Lexical optimization is conducted.\cr
    \code{qqmedian.mean}: calculate the median standardized difference in the eQQ
    plot for each variable.  Minimize the median of these differences across
    variables.\cr
    \code{qqmedian.max}:  calculate the median standardized difference in the eQQ
    plot for each variable.  Minimize the maximum of these differences across
    variables.  Lexical optimization is conducted.\cr
    \code{qqmax.mean}: calculate the maximum standardized difference in the eQQ
    plot for each variable.  Minimize the mean of these differences across
    variables.\cr
    \code{qqmax.max}:  calculate the maximum standardized difference in the eQQ
    plot for each variable.  Minimize the maximum of these differences across
    variables.  Lexical optimization is conducted.\cr
    Users may provide their own \code{fit.func}. The name of the user
    provided function should not be backquoted or quoted.  This function needs
    to return a fit value that will be minimized, by lexical
    optimization if more than one fit value is returned.  The function
    should expect two arguments.  The first being the \code{matches} object
    returned by \code{GenMatch}---see
    below.  And the second being a matrix which contains the variables to
    be balanced---i.e., the \code{BalanceMatrix} the user provided to
    \code{GenMatch}. For an example see
    \url{http://sekhon.berkeley.edu/matching/R/my_fitfunc.R}.}
  \item{MemoryMatrix}{
    This variable controls if \code{\link[rgenoud]{genoud}} sets up a memory matrix.  Such a
    matrix ensures that \code{\link[rgenoud]{genoud}} will request the fitness evaluation
    of a given set of parameters only once. The variable may be
    \emph{TRUE} or \emph{FALSE}.  If it is \emph{FALSE}, \code{\link[rgenoud]{genoud}}
    will be aggressive in
    conserving memory.  The most significant negative implication of
    this variable being set to \emph{FALSE} is that \code{\link[rgenoud]{genoud}} will no
    longer maintain a memory
    matrix of all evaluated individuals.  Therefore, \code{\link[rgenoud]{genoud}} may request
    evaluations which it has  previously requested.  When
    the number variables in \code{X} is large, the memory matrix
    consumes a large amount of RAM.\cr
    
    \code{\link[rgenoud]{genoud}}'s memory matrix will require \emph{significantly} less
    memory if the user sets \code{hard.generation.limit} equal
    to \emph{TRUE}.  Doing this is a good way of conserving
    memory while still making use of the memory matrix structure.}
  \item{exact}{ A logical scalar or vector for whether exact matching
    should be done.  If a logical scalar is
    provided, that logical value is applied to all covariates in
    \code{X}.  If a logical vector is provided, a logical value should
    be provided for each covariate in \code{X}. Using a logical vector
    allows the user to specify exact matching for some but not other
    variables.  When exact matches are not found, observations are
    dropped.  \code{distance.tolerance} determines what is considered to
    be an exact match. The \code{exact} option takes precedence over the
    \code{caliper} option.  Obviously, if \code{exact} matching is done
    using \emph{all} of the covariates, one should not be using
    \code{GenMatch} unless the \code{distance.tolerance} has been set
    unusually high.}
  \item{caliper}{ A scalar or vector denoting the caliper(s) which
    should be used when matching.  A caliper is the distance which is
    acceptable for any match.  Observations which are outside of the
    caliper are dropped. If a scalar caliper is provided, this caliper is
    used for all covariates in \code{X}.  If a vector of calipers is
    provided, a caliper value should be provided for each covariate in
    \code{X}. The caliper is interpreted to be in standardized units.  For
    example, \code{caliper=.25} means that all matches not equal to or
    within .25 standard deviations of each covariate in \code{X} are
    dropped.  The \code{ecaliper} object which is returned by
    \code{GenMatch} shows the enforced caliper on the scale of the
    \code{X} variables. Note that dropping observations generally changes
    the quantity being estimated.}
  \item{replace}{A logical flag for whether matching should be done with
    replacement.  Note that if \code{FALSE}, the order of matches
    generally matters.  Matches will be found in the same order as the
    data are sorted.  Thus, the match(es) for the first observation will
    be found first, the match(es) for the second observation will be found second, etc.
    Matching without replacement will generally increase bias.
    Ties are randomly broken when \code{replace==FALSE}---see the
    \code{ties} option for details.}
  \item{ties}{A logical flag for whether ties should be handled deterministically.  By
    default \code{ties==TRUE}. If, for example, one treated observation
    matches more than one control observation, the matched dataset will
    include the multiple matched control observations and the matched data
    will be weighted to reflect the multiple matches.  The sum of the
    weighted observations will still equal the original number of
    observations. If \code{ties==FALSE}, ties will be randomly broken.
    \emph{If the dataset is large and there are many ties, setting
    \code{ties=FALSE} often results in a large speedup.} Whether two
    potential matches are close enough to be considered tied, is
    controlled by the \code{distance.tolerance}
    option.}
  \item{CommonSupport}{This logical flag implements the usual procedure
    by which observations outside of the common support of a variable
    (usually the propensity score) across treatment and control groups are
    discarded.  The \code{caliper} option is to
    be preferred to this option because \code{CommonSupport}, consistent
    with the literature, only drops \emph{outliers} and leaves
    \emph{inliers} while the caliper option drops both.
    If \code{CommonSupport==TRUE}, common support will be enforced on
    the first variable in the \code{X} matrix.  Note that dropping
    observations generally changes the quantity being estimated.  Use of
    this option renders it impossible to use the returned
    object \code{matches} to reconstruct the matched dataset.
    Seriously, don't use this option; use the \code{caliper} option instead.}
  \item{nboots}{The number of bootstrap samples to be run for the
    \code{ks} test.  By default this option is set to zero so no
    bootstraps are done.  See \code{\link{ks.boot}} for additional
    details.}
  \item{ks}{ A logical flag for if the univariate bootstrap
    Kolmogorov-Smirnov (KS) test should be calculated.  If the ks option
    is set to true, the univariate KS test is calculated for all
    non-dichotomous variables.  The bootstrap KS test is consistent even
    for non-continuous variables.  By default, the bootstrap KS test is
    not used. To change this see the \code{nboots} option. If a given
    variable is dichotomous, a t-test is used even if the KS test is requested.  See
    \code{\link{ks.boot}} for additional details.}
  \item{verbose}{A logical flag for whether details of each
    fitness evaluation should be printed.  Verbose is set to FALSE if
    the \code{cluster} option is used.}
  \item{distance.tolerance}{This is a scalar which is used to determine
    if distances between two observations are different from zero.  Values
    less than \code{distance.tolerance} are deemed to be equal to zero.
    This option can be used to perform a type of optimal matching.}
  \item{tolerance}{ This is a scalar which is used to determine
    numerical tolerances.  This option is used by numerical routines
    such as those used to determine if a matrix is singular.}
  \item{min.weight}{This is the minimum weight any variable may be
    given.}  
  \item{max.weight}{This is the maximum weight any variable may be
    given.}
  \item{Domains}{This is a \code{ncol(X)} \eqn{\times 2}{*2} matrix.
    The first column is the lower bound, and the second column is the
    upper bound for each variable over which \code{\link[rgenoud]{genoud}} will
    search for weights.  If the user does not provide this matrix, the
    bounds for each variable will be determined by the \code{min.weight}
    and \code{max.weight} options.}
  \item{print.level}{
    This option controls the level of printing.  There
    are four possible levels: 0 (minimal printing), 1 (normal), 2
    (detailed), and 3 (debug).  If level 2 is selected, \code{GenMatch} will
    print details about the population at each generation, including the
    best individual found so far. If debug
    level printing is requested, details of the \code{\link[rgenoud]{genoud}}
    population are printed in the "genoud.pro" file which is located in
    the temporary \code{R} directory returned by the \code{\link{tempdir}}
    function.  See the \code{project.path} option for more details.
    Because \code{GenMatch} runs may take a long time, it is important for the
    user to receive feedback.  Hence, print level 2 has been set as the
    default.}
  \item{project.path}{ This is the path of the
    \code{\link[rgenoud]{genoud}} project file.  By default no file is
    produced unless \code{print.level=3}.  In that case,
    \code{\link[rgenoud]{genoud}} places its output in a file called
    "genoud.pro" located in the temporary directory provided by
    \code{\link{tempdir}}.  If a file path is provided to the
    \code{project.path} option, a file will be created regardless of the
    \code{print.level}. The behavior of the project file, however, will
    depend on the \code{print.level} chosen.  If the \code{print.level}
    variable is set to 1, then the project file is rewritten after each
    generation.  Therefore, only the currently fully completed generation
    is included in the file.  If the \code{print.level} variable is set to
    2 or higher, then each new generation is simply appended to the
    project file. No project file is generated for
    \code{print.level=0}.}
  \item{paired}{A flag for whether the paired \code{\link{t.test}} should be
    used when determining balance.}
  \item{loss}{The loss function to be optimized.  The default value, \code{1},
    implies "lexical" optimization: all of the balance statistics will
    be sorted from the most discrepant to the least and weights will be
    picked which minimize the maximum discrepancy. If multiple sets of
    weights result in the same maximum discrepancy, then the second
    largest discrepancy is examined to choose the best weights.  The
    processes continues iteratively until ties are broken.  \cr

    If the value of \code{2} is used, then only the maximum discrepancy
    is examined.  This was the default behavior prior to version 1.0.  The
    user may also pass in any function she desires. Note that the
    option 1 corresponds to the \code{\link{sort}} function and option 2
    to the \code{\link{min}} function.  Any user specified function
    should expect a vector of balance statistics ("p-values") and it
    should return either a vector of values (in which case "lexical"
    optimization will be done) or a scalar value (which will be
    maximized). Some possible alternative functions are
    \code{\link{mean}} or \code{\link{median}}.}
  \item{data.type.integer}{
    By default, floating-point weights are considered. If this option is
    set to \code{TRUE}, search will be done over integer weights. Note
    that before version 4.1, the default was to use integer weights.}
  \item{restrict}{A matrix which restricts the possible matches. This
    matrix has one row for each restriction and three
    columns.  The first two columns contain the two observation numbers
    which are to be restricted (for example 4 and 20), and the third
    column is the restriction imposed on the observation-pair.
    Negative numbers in the third column imply that the two observations
    cannot be matched under any circumstances, and positive numbers are
    passed on as the distance between the two observations for the
    matching algorithm.  The most commonly used positive restriction is
    \code{0} which implies that the two observations will always
    be matched.  \cr

    Exclusion restriction are even more common.  For example, if we want
    to exclude the observation pair 4 and 20 and the pair 6 and 55 from
    being matched, the restrict matrix would be:
    \code{restrict=rbind(c(4,20,-1),c(6,55,-1))}} \item{cluster}{This
    can either be an object of the 'cluster' class returned by one of
    the \code{\link[parallel]{makeCluster}} commands in the
    \code{parallel} package or a vector of machine names so that
    \code{GenMatch} can setup the cluster automatically. If it is the
    latter, the vector should look like: \cr
    \code{c("localhost","musil","musil","deckard")}.\cr This vector
    would create a cluster with four nodes: one on the localhost another
    on "deckard" and two on the machine named "musil".  Two nodes on a
    given machine make sense if the machine has two or more chips/cores.
    \code{GenMatch} will setup a SOCK cluster by a call to
    \code{\link[parallel]{makePSOCKcluster}}.  This will require the
    user to type in her password for each node as the cluster is by
    default created via \code{ssh}.  One can add on usernames to the
    machine name if it differs from the current shell: "username@musil".
    Other cluster types, such as PVM and MPI, which do not require
    passwords, can be created by directly calling
    \code{\link[parallel]{makeCluster}}, and then passing the returned
    cluster object to \code{GenMatch}. For an example of how to manually
    setup up a cluster with a direct call to
    \code{\link[parallel]{makeCluster}} see
    \url{http://sekhon.berkeley.edu/matching/R/cluster_manual.R}.  For
    an example of how to get around a firewall by ssh tunneling see:
    \url{http://sekhon.berkeley.edu/matching/R/cluster_manual_tunnel.R}.}
    \item{balance}{This logical flag controls if load balancing is done
    across the cluster.  Load balancing can result in better cluster
    utilization; however, increased communication can reduce
    performance.  This option is best used if each individual call to
    \code{\link[Matching]{Match}} takes at least several minutes to
    calculate or if the nodes in the cluster vary significantly in their
    performance. If cluster==FALSE, this option has no effect.}
    \item{...}{Other options which are passed on to
    \code{\link[rgenoud]{genoud}}.}  } \value{ \item{value}{The fit
    values at the solution.  By default, this is a vector of p-values
    sorted from the smallest to the largest.  There will generally be
    twice as many p-values as there are variables in
    \code{BalanceMatrix}, unless there are dichotomous variables in this
    matrix.  There is one p-value for each covariate in
    \code{BalanceMatrix} which is the result of a paired t-test and
    another p-value for each non-dichotomous variable in
    \code{BalanceMatrix} which is the result of a Kolmogorov-Smirnov
    test. Recall that these p-values cannot be interpreted as hypothesis
    tests.  They are simply measures of balance.}  \item{par}{A vector
    of the weights given to each variable in \code{X}.}
    \item{Weight.matrix}{A matrix whose diagonal corresponds to the
    weight given to each variable in \code{X}.  This object corresponds
    to the \code{Weight.matrix} in the \code{\link{Match}} function.}
    \item{matches}{A matrix where the first column contains the row
    numbers of the treated observations in the matched dataset. The
    second column contains the row numbers of the control
    observations. And the third column contains the weight that each
    matched pair is given.  These objects may not correspond
    respectively to the \code{index.treated}, \code{index.control} and
    \code{weights} objects which are returned by \code{\link{Match}}
    because they may be ordered in a different way. Therefore, end users
    should use the objects returned by \code{\link{Match}} because those
    are ordered in the way that users expect.}
  \item{ecaliper }{The
    size of the enforced caliper on the scale of the \code{X} variables.
    This object has the same length as the number of covariates in
    \code{X}.}  } \references{ Sekhon, Jasjeet S. 2011.  "Multivariate
    and Propensity Score Matching Software with Automated Balance
    Optimization.''  \emph{Journal of Statistical Software} 42(7): 1-52.
    \url{http://www.jstatsoft.org/v42/i07/}
  
  Diamond, Alexis and Jasjeet S. Sekhon. 2013. "Genetic
  Matching for Estimating Causal Effects: A General Multivariate
  Matching Method for Achieving Balance in Observational Studies.''
  \emph{Review of Economics and Statistics}.  95 (3): 932--945.
  \url{http://sekhon.berkeley.edu/papers/GenMatch.pdf}

  Sekhon, Jasjeet Singh and Walter R. Mebane, Jr. 1998.  "Genetic
  Optimization Using Derivatives: Theory and Application to Nonlinear
  Models.'' \emph{Political Analysis}, 7: 187-210.
  \url{http://sekhon.berkeley.edu/genoud/genoud.pdf} 
}
\author{
  Jasjeet S. Sekhon, UC Berkeley, \email{sekhon@berkeley.edu},
  \url{http://sekhon.berkeley.edu/}.  
}
\seealso{ Also see \code{\link{Match}}, \code{\link{summary.Match}},
  \code{\link{MatchBalance}}, \code{\link[rgenoud]{genoud}},
  \code{\link{balanceUV}}, \code{\link{qqstats}},
  \code{\link{ks.boot}}, \code{\link{GerberGreenImai}}, \code{\link{lalonde}}
} 
\examples{
data(lalonde)
attach(lalonde)

#The covariates we want to match on
X = cbind(age, educ, black, hisp, married, nodegr, u74, u75, re75, re74)

#The covariates we want to obtain balance on
BalanceMat <- cbind(age, educ, black, hisp, married, nodegr, u74, u75, re75, re74,
                    I(re74*re75))

#
#Let's call GenMatch() to find the optimal weight to give each
#covariate in 'X' so as we have achieved balance on the covariates in
#'BalanceMat'. This is only an example so we want GenMatch to be quick
#so the population size has been set to be only 16 via the 'pop.size'
#option. This is *WAY* too small for actual problems.
#For details see http://sekhon.berkeley.edu/papers/MatchingJSS.pdf.
#
genout <- GenMatch(Tr=treat, X=X, BalanceMatrix=BalanceMat, estimand="ATE", M=1,
                   pop.size=16, max.generations=10, wait.generations=1)

#The outcome variable
Y=re78/1000

#
# Now that GenMatch() has found the optimal weights, let's estimate
# our causal effect of interest using those weights
#
mout <- Match(Y=Y, Tr=treat, X=X, estimand="ATE", Weight.matrix=genout)
summary(mout)

#                        
#Let's determine if balance has actually been obtained on the variables of interest
#                        
mb <- MatchBalance(treat~age +educ+black+ hisp+ married+ nodegr+ u74+ u75+
                   re75+ re74+ I(re74*re75),
                   match.out=mout, nboots=500)

# For more examples see: http://sekhon.berkeley.edu/matching/R.
}
\keyword{nonparametric}

%  LocalWords:  GenMatch Kolmogorov multinomial CPUs BalanceMatrix estimand ATT
%  LocalWords:  ncol MemoryMatrix nboots ATC rgenoud genoud emph cr ecaliper UC
%  LocalWords:  eqn tempdir rbind makeCluster localhost musil deckard usernames
%  LocalWords:  makePSOCKcluster PVM MPI url dataset seealso MatchBalance cbind
%  LocalWords:  balanceUV GerberGreenImai lalonde hisp nodegr genout
%  LocalWords:  BalanceMat mout mb mv pvals qqmean QQ qqmedian qqmax Smirnov Tr
%  LocalWords:  func CommonSupport min Abadie Imbens boostrap ssh username educ
%  LocalWords:  ealso qqstats nonparametric eQQ rep ks inliers es

\name{qqstats}
\alias{qqstats}
\title{QQ Summary Statistics}
\description{
  This function calculates a set of summary statistics for the QQ
  plot of two samples of data.  The summaries are useful for determining
  if the two samples are from the same distribution.  If
  \code{standardize==TRUE}, the empirical CDF is used instead of the
  empirical-QQ plot.  The later retains the scale of the variable.
}
\usage{
qqstats(x, y, standardize=TRUE, summary.func)
}
\arguments{
  \item{x}{The first sample.}
  \item{y}{The second sample.}
  \item{standardize}{A logical flag for whether the statistics should be
  standardized by the empirical cumulative distribution functions of the
  two samples.}
  \item{summary.func}{A user provided function to summarize the
    difference between the two distributions.  The function should
    expect a vector of the differences as an argument and return summary
    statistic.  For example, the \code{\link{quantile}} function is a
    legal function to pass in.}
}
\value{
  \item{meandiff}{The mean difference between the QQ plots of the two
    samples.}
  \item{mediandiff}{The median difference between the QQ plots of the
    two samples.}
  \item{maxdiff}{The maximum difference between the QQ plots of the two
    samples.}
  \item{summarydiff}{If the user provides a \code{summary.func}, the
    user requested summary difference is returned.}
  \item{summary.func}{If the user provides a \code{summary.func}, the
    function is returned.}
}
\author{
  Jasjeet S. Sekhon, UC Berkeley, \email{sekhon@berkeley.edu},
  \url{http://sekhon.berkeley.edu/}.  
}
\references{
  Sekhon, Jasjeet S. 2011.  "Multivariate and Propensity Score
  Matching Software with Automated Balance Optimization.''
  \emph{Journal of Statistical Software} 42(7): 1-52.
  \url{http://www.jstatsoft.org/v42/i07/}

  Diamond, Alexis and Jasjeet S. Sekhon. Forthcoming. "Genetic Matching for
  Estimating Causal Effects: A General Multivariate Matching Method for
  Achieving Balance in Observational Studies.'' \emph{Review of Economics and Statistics}.
  \url{http://sekhon.berkeley.edu/papers/GenMatch.pdf}
  }
  \seealso{ Also see \code{\link{ks.boot}},
    \code{\link{balanceUV}}, \code{\link{Match}},
    \code{\link{GenMatch}},
    \code{\link{MatchBalance}},
    \code{\link{GerberGreenImai}}, \code{\link{lalonde}} }
\examples{
#
# Replication of Dehejia and Wahba psid3 model
#
# Dehejia, Rajeev and Sadek Wahba. 1999.``Causal Effects in
# Non-Experimental Studies: Re-Evaluating the Evaluation of Training
# Programs.''Journal of the American Statistical Association 94 (448):
# 1053-1062.
#

data(lalonde)

#
# Estimate the propensity model
#
glm1  <- glm(treat~age + I(age^2) + educ + I(educ^2) + black +
             hisp + married + nodegr + re74  + I(re74^2) + re75 + I(re75^2) +
             u74 + u75, family=binomial, data=lalonde)


#
#save data objects
#
X  <- glm1$fitted
Y  <- lalonde$re78
Tr  <- lalonde$treat

#
# one-to-one matching with replacement (the "M=1" option).
# Estimating the treatment effect on the treated (the "estimand" option which defaults to 0).
#
rr  <- Match(Y=Y,Tr=Tr,X=X,M=1);
summary(rr)

#
# Do we have balance on 1975 income after matching?
#
qqout  <- qqstats(lalonde$re75[rr$index.treated], lalonde$re75[rr$index.control])
print(qqout)
}
\keyword{htest}
\keyword{distribution}

%  LocalWords:  Kolmogorov nboots ksboot pvalue UC url Abadie emph seealso psid
%  LocalWords:  balanceUV GenMatch MatchBalance GerberGreenImai Wahba
%  LocalWords:  lalonde Dehejia Rajeev Sadek glm hisp nodegr estimand rr htest
%  LocalWords:  maxdiff QQ summarydiff func qqout qqstats meandiff mediandiff

\name{Match}
\alias{Match}
\title{Multivariate and Propensity Score Matching Estimator for Causal Inference}
\description{
  \code{Match} implements a variety of algorithms for multivariate
  matching including propensity score, Mahalanobis and inverse variance
  matching.  The function is intended to be used in conjunction with the
  \code{MatchBalance} function which determines the extent to which
  \code{Match} has been able to achieve covariate balance.  In order to
  do propensity score matching, one should estimate the propensity model
  before calling \code{Match}, and then send \code{Match} the propensity
  score to use.  \code{Match} enables a wide variety of matching
  options including matching with or without replacement, bias
  adjustment, different methods for handling ties, exact and caliper
  matching, and a method for the user to fine tune the matches via a
  general restriction matrix.  Variance estimators include the usual
  Neyman standard errors, Abadie-Imbens standard errors, and robust
  variances which do not assume a homogeneous causal effect. The
  \code{\link{GenMatch}} function can be used to \emph{automatically
  find balance} via a genetic search algorithm which determines the
  optimal weight to give each covariate.
}
\usage{
Match(Y=NULL, Tr, X, Z = X, V = rep(1, length(Y)), estimand = "ATT", M = 1,
      BiasAdjust = FALSE, exact = NULL, caliper = NULL, replace=TRUE, ties=TRUE,
      CommonSupport=FALSE,Weight = 1, Weight.matrix = NULL, weights = NULL,
      Var.calc = 0, sample = FALSE, restrict=NULL, match.out = NULL,
      distance.tolerance = 1e-05, tolerance=sqrt(.Machine$double.eps),
      version="standard")
}
\arguments{
  \item{Y}{ A vector containing the outcome of interest.
    Missing values are not allowed.  An outcome vector is not required
    because the matches generated will be the same regardless of the
    outcomes.  Of course, without any outcomes no causal effect
    estimates will be produced, only a matched dataset. }
  \item{Tr}{ A vector indicating the observations which are
    in the treatment regime and those which are not.  This can either be a
    logical vector or a real vector where 0 denotes control and 1 denotes
    treatment.}
  \item{X}{ A matrix containing the variables we wish to match on.
    This matrix may contain the actual observed covariates or the
    propensity score or a combination of both. All columns of this
    matrix must have positive variance or \code{Match} will return an
    error.}
  \item{Z}{ A matrix containing the covariates for which we wish to make
    bias adjustments.}
  \item{V}{ A matrix containing the covariates for which the variance
    of the causal effect may vary. Also see the \code{Var.calc} option,
    which takes precedence.}
  \item{estimand}{ A character string for the estimand.  The default
    estimand is "ATT", the sample average treatment effect for the
    treated. "ATE" is the sample average treatment effect, and
    "ATC" is the sample average treatment effect for the controls.}
  \item{M}{A scalar for the number of matches which should be
    found. The default is one-to-one matching. Also see the \code{ties}
    option.}
  \item{BiasAdjust}{ A logical scalar for whether regression adjustment
    should be used. See the \code{Z} matrix.}
  \item{exact}{ A logical scalar or vector for whether exact matching
    should be done. If a logical scalar is provided, that logical value is
      applied to all covariates in
    \code{X}.  If a logical vector is provided, a logical value should
    be provided for each covariate in \code{X}. Using a logical vector
    allows the user to specify exact matching for some but not other
    variables.  When exact matches are not found, observations are
    dropped.  \code{distance.tolerance} determines what is considered to be an
    exact match. The \code{exact} option takes precedence over the
    \code{caliper} option.}
  \item{caliper}{ A scalar or vector denoting the caliper(s) which
    should be used when matching.  A caliper is the distance which is
    acceptable for any match.  Observations which are outside of the
    caliper are dropped. If a scalar caliper is provided, this caliper is
    used for all covariates in \code{X}.  If a vector of calipers is
    provided, a caliper value should be provided for each covariate in
    \code{X}. The caliper is interpreted to be in standardized units.  For
    example, \code{caliper=.25} means that all matches not equal to or
    within .25 standard deviations of each covariate in \code{X} are
    dropped. Note that dropping observations generally changes the
    quantity being estimated.}
  \item{replace}{A logical flag for whether matching should be done with
    replacement.  Note that if \code{FALSE}, the order of matches
    generally matters.  Matches will be found in the same order as the
    data are sorted.  Thus, the match(es) for the first observation will
    be found first, the match(es) for the second observation will be found second, etc.
    Matching without replacement will generally increase bias.
    Ties are randomly broken when \code{replace==FALSE}
    ---see the \code{ties} option for details.}
  \item{ties}{A logical flag for whether ties should be handled deterministically.  By
    default \code{ties==TRUE}. If, for example, one treated observation
    matches more than one control observation, the matched dataset will
    include the multiple matched control observations and the matched data
    will be weighted to reflect the multiple matches.  The sum of the
    weighted observations will still equal the original number of
    observations. If \code{ties==FALSE}, ties will be randomly broken.
    \emph{If the dataset is large and there are many ties, setting
    \code{ties=FALSE} often results in a large speedup.} Whether two
    potential matches are close enough to be considered tied, is
    controlled by the \code{distance.tolerance}
    option.}
  \item{CommonSupport}{This logical flag implements the usual procedure
    by which observations outside of the common support of a variable
    (usually the propensity score) across treatment and control groups are
    discarded.  The \code{caliper} option is to
    be preferred to this option because \code{CommonSupport}, consistent
    with the literature, only drops \emph{outliers} and leaves
    \emph{inliers} while the caliper option drops both.
    If \code{CommonSupport==TRUE}, common support will be enforced on
    the first variable in the \code{X} matrix.  Note that dropping
    observations generally changes the quantity being estimated. Use of
    this option renders it impossible to use the returned
    objects \code{index.treated} and \code{index.control} to
    reconstruct the matched dataset.  The returned object \code{mdata}
    will, however, still contain the matched dataset.  Seriously, don't
    use this option; use the \code{caliper} option instead.}
  \item{Weight}{ A scalar for the type of weighting scheme the matching
    algorithm should use when weighting each of the covariates in
    \code{X}.  The default value of 1 denotes that weights are equal to
    the inverse of the variances. 2 denotes the Mahalanobis distance
    metric, and 3 denotes that the user will supply a weight matrix
    (\code{Weight.matrix}).  Note that if the user supplies a
    \code{Weight.matrix}, \code{Weight} will be automatically set to be
    equal to 3.}
  \item{Weight.matrix}{ This matrix denotes the weights the matching
    algorithm uses when weighting each of the covariates in \code{X}---see
    the \code{Weight} option. This square matrix should have as many
    columns as the number of columns of the \code{X} matrix. This matrix
    is usually provided by a call to the \code{\link{GenMatch}} function
    which finds the optimal weight each variable should be given so as to
    achieve balance on the covariates. \cr

    For most uses, this matrix has zeros in the off-diagonal
    cells.  This matrix can be used to weight some variables more than
    others.  For
    example, if \code{X} contains three variables and we want to
    match as best as we can on the first, the following would work well:
    \cr
    \code{> Weight.matrix <- diag(3)}\cr
    \code{> Weight.matrix[1,1] <- 1000/var(X[,1])} \cr
    \code{> Weight.matrix[2,2] <- 1/var(X[,2])} \cr
    \code{> Weight.matrix[3,3] <- 1/var(X[,3])} \cr
    This code changes the weights implied by the
    inverse of the variances by multiplying the first variable by a 1000
    so that it is highly weighted.  In order to enforce exact matching
    see the \code{exact} and \code{caliper} options.}
  \item{weights}{ A vector the same length as \code{Y} which
    provides observation specific weights.}
  \item{Var.calc}{ A scalar for the variance estimate
    that should be used.  By default \code{Var.calc=0} which means that
    homoscedasticity is assumed.  For values of  \code{Var.calc > 0},
    robust variances are calculated using \code{Var.calc} matches. }
  \item{sample}{ A logical flag for whether the population or sample
    variance is returned. }
  \item{distance.tolerance}{This is a scalar which is used to determine
    if distances between two observations are different from zero.  Values
    less than \code{distance.tolerance} are deemed to be equal to zero.
    This option can be used to perform a type of optimal matching}
  \item{tolerance}{ This is a scalar which is used to determine
    numerical tolerances.  This option is used by numerical routines
    such as those used to determine if a matrix is singular.}
  \item{restrict}{A matrix which restricts the possible matches.  This
    matrix has one row for each restriction and three
    columns.  The first two columns contain the two observation numbers
    which are to be restricted (for example 4 and 20), and the third
    column is the restriction imposed on the observation-pair.
    Negative numbers in the third column imply that the two observations
    cannot be matched under any circumstances, and positive numbers are
    passed on as the distance between the two observations for the
    matching algorithm.  The most commonly used positive restriction is
    \code{0} which implies that the two observations will always
    be matched.  \cr

    Exclusion restrictions are even more common.  For example, if we want
    to exclude the observation pair 4 and 20 and
    the pair 6 and 55 from being matched, the restrict matrix would be:
    \code{restrict=rbind(c(4,20,-1),c(6,55,-1))}}
  \item{match.out}{ The return object from a previous call to
    \code{Match}.  If this object is provided, then \code{Match} will
    use the matches found by the previous invocation of the function.
    Hence, \code{Match} will run faster.  This is
    useful when the treatment does not vary across calls to
    \code{Match} and one wants to use the same set of matches as found
    before.  This often occurs when one is trying to estimate the causal
    effect of the same treatment (\code{Tr}) on different outcomes
    (\code{Y}). When using this option, be careful to use the same
    arguments as used for the previous invocation of \code{Match} unless
    you know exactly what you are doing.}
  \item{version}{The version of the code to be used.  The "fast" C/C++
    version of the code does not calculate Abadie-Imbens standard errors.
    Additional speed can be obtained by setting \code{ties=FALSE} or
    \code{replace=FALSE} if the dataset is large and/or has many ties.
    The "legacy" version of the code does not make a call to an optimized
    C/C++ library and is included only for historical compatibility.  The
    "fast" version of the code is significantly faster than the "standard"
    version for large datasets, and the "legacy" version is much slower
    than either of the other two.}
}
\details{

  This function is intended to be used in conjunction with the
  \code{MatchBalance} function which checks if the results of this
  function have actually achieved balance.  The results of this function
  can be summarized by a call to the \code{\link{summary.Match}}
  function. If one wants to do propensity score matching, one should estimate the
  propensity model before calling \code{Match}, and then place the
  fitted values in the \code{X} matrix---see the provided example.  \cr

  The \code{\link{GenMatch}} function can be used to \emph{automatically
  find balance} by the use of a genetic search algorithm which determines
  the optimal weight to give each covariate. The object returned by
  \code{\link{GenMatch}} can be supplied to the \code{Weight.matrix}
  option of \code{Match} to obtain estimates.\cr

  \code{Match} is often much faster with large datasets if
  \code{ties=FALSE} or \code{replace=FALSE}---i.e., if matching is done
  by randomly breaking ties or without replacement.  Also see the
  \code{\link{Matchby}} function.  It provides a wrapper for
  \code{Match} which is much faster for large datasets when it can be
  used.\cr

  Three demos are included: \code{GerberGreenImai}, \code{DehejiaWahba},
  and \code{AbadieImbens}.  These can be run by calling the
  \code{\link{demo}} function such as by \code{demo(DehejiaWahba)}. \cr
}

\value{
  \item{est }{The estimated average causal effect.}
  \item{se }{The Abadie-Imbens standard error.  This standard error has
    correct coverage if \code{X} consists of either covariates or a known
    propensity score because it takes into account the uncertainty of the
    matching procedure.  If an estimated propensity score is used, the
    uncertainty involved in its estimation is not accounted for although
    the uncertainty of the matching procedure itself still is.}
  \item{est.noadj }{The estimated average causal effect without any
    \code{BiasAdjust}.  If \code{BiasAdjust} is not requested, this is the
    same as \code{est}.}
  \item{se.standard }{The usual standard error.  This is the standard error
    calculated on the matched data using the usual method of calculating
    the difference of means (between treated and control) weighted by the
    observation weights provided by \code{weights}.  Note that the
    standard error provided by \code{se} takes into account the uncertainty
    of the matching procedure while \code{se.standard} does not.  Neither
    \code{se} nor \code{se.standard} take into account the uncertainty of
    estimating a propensity score.  \code{se.standard} does
    not take into account any \code{BiasAdjust}.  Summary of both types
    of standard error results can be requested by setting the
    \code{full=TRUE} flag when using the \code{\link{summary.Match}}
    function on the object returned by \code{Match}.}
  \item{se.cond }{The conditional standard error.  The practitioner
    should not generally use this.}
  \item{mdata }{A list which contains the matched datasets produced by
    \code{Match}.  Three datasets are included in this list: \code{Y},
    \code{Tr} and \code{X}.}
  \item{index.treated }{A vector containing the observation numbers from
    the original dataset for the treated observations in the
    matched dataset.  This index in conjunction with \code{index.control}
    can be used to recover the matched dataset produced by
    \code{Match}.  For example, the \code{X} matrix used by \code{Match}
    can be recovered by
    \code{rbind(X[index.treated,],X[index.control,])}. The user should
    generally just examine the output of \code{mdata}.}
  \item{index.control }{A vector containing the observation numbers from
    the original data for the control observations in the
    matched data.  This index in conjunction with \code{index.treated}
    can be used to recover the matched dataset produced by
    \code{Match}.  For example, the \code{X} matrix used by \code{Match}
    can be recovered by
    \code{rbind(X[index.treated,],X[index.control,])}. The user should
    generally just examine the output of \code{mdata}.}
  \item{index.dropped}{A vector containing the observation numbers from
    the original data which were dropped (if any) in the matched dataset
    because of various options such as \code{caliper} and
    \code{exact}.  If no observations were dropped, this
    index will be \code{NULL}.}
  \item{weights}{A vector of weights.  There is one weight for each
    matched-pair in the matched dataset.  If all of the observations had a
    weight of 1 on input, then each matched-pair will have a weight of 1
    on output if there are no ties.}
  \item{orig.nobs }{The original number of observations in the dataset.}
  \item{orig.wnobs }{The original number of weighted observations in the
    dataset.}
  \item{orig.treated.nobs}{The original number of treated observations (unweighted).}
  \item{nobs }{The number of observations in the matched dataset.}
  \item{wnobs }{The number of weighted observations in the matched dataset.}
  \item{caliper }{The \code{caliper} which was used.}
  \item{ecaliper }{The size of the enforced caliper on the scale of the
    \code{X} variables.  This object has the same length as the number of
    covariates in \code{X}.}
  \item{exact}{The value of the \code{exact} function argument.}
  \item{ndrops}{The number of weighted observations which were dropped
    either because of caliper or exact matching.  This number, unlike
    \code{ndrops.matches}, takes into account observation specific
    weights which the user may have provided via the \code{weights}
    argument.}
  \item{ndrops.matches}{The number of matches which were dropped either because of
    caliper or exact matching.}
}
\references{
  Sekhon, Jasjeet S. 2011.  "Multivariate and Propensity Score
  Matching Software with Automated Balance Optimization.''
  \emph{Journal of Statistical Software} 42(7): 1-52.
  \doi{10.18637/jss.v042.i07}

  Diamond, Alexis and Jasjeet S. Sekhon. 2013. "Genetic
  Matching for Estimating Causal Effects: A General Multivariate
  Matching Method for Achieving Balance in Observational Studies.''
  \emph{Review of Economics and Statistics}. 95 (3): 932--945.
  \url{http://sekhon.berkeley.edu/papers/GenMatch.pdf}

  Abadie, Alberto and Guido Imbens. 2006.
  ``Large Sample Properties of Matching Estimators for Average
  Treatment Effects.'' \emph{Econometrica} 74(1): 235-267.

  Imbens, Guido. 2004. Matching Software for Matlab and
  Stata.

}
\author{Jasjeet S. Sekhon, UC Berkeley, \email{sekhon@berkeley.edu},
  \url{http://sekhon.berkeley.edu/}.
}
\seealso{ Also see \code{\link{summary.Match}},
  \code{\link{GenMatch}},
  \code{\link{MatchBalance}},
  \code{\link{Matchby}},
  \code{\link{balanceUV}},
  \code{\link{qqstats}}, \code{\link{ks.boot}},
  \code{\link{GerberGreenImai}}, \code{\link{lalonde}}
}
\examples{

# Replication of Dehejia and Wahba psid3 model
#
# Dehejia, Rajeev and Sadek Wahba. 1999.``Causal Effects in
# Non-Experimental Studies: Re-Evaluating the Evaluation of Training
# Programs.''Journal of the American Statistical Association 94 (448):
# 1053-1062.

data(lalonde)

#
# Estimate the propensity model
#
glm1  <- glm(treat~age + I(age^2) + educ + I(educ^2) + black +
             hisp + married + nodegr + re74  + I(re74^2) + re75 + I(re75^2) +
             u74 + u75, family=binomial, data=lalonde)


#
#save data objects
#
X  <- glm1$fitted
Y  <- lalonde$re78
Tr  <- lalonde$treat

#
# one-to-one matching with replacement (the "M=1" option).
# Estimating the treatment effect on the treated (the "estimand" option defaults to ATT).
#
rr  <- Match(Y=Y, Tr=Tr, X=X, M=1);
summary(rr)

# Let's check the covariate balance
# 'nboots' is set to small values in the interest of speed.
# Please increase to at least 500 each for publication quality p-values.
mb  <- MatchBalance(treat~age + I(age^2) + educ + I(educ^2) + black +
                    hisp + married + nodegr + re74  + I(re74^2) + re75 + I(re75^2) +
                    u74 + u75, data=lalonde, match.out=rr, nboots=10)
}
\keyword{nonparametric}


%  LocalWords:  MatchBalance GenMatch emph estimand ATT BiasAdjust calc dataset
%  LocalWords:  ATC ecaliper cr diag homoscedasticity rbind GerberGreenImai se
%  LocalWords:  DehejiaWahba AbadieImbens noadj cond mdata datasets wnobs url
%  LocalWords:  ndrops Abadie Imbens Econometrica Matlab Stata UC seealso Wahba
%  LocalWords:  balanceUV lalonde Dehejia psid Rajeev Sadek glm hisp
%  LocalWords:  nodegr rr nboots nmc mb Neyman Tr CommonSupport Var var Matchby
%  LocalWords:  est orig nobs Guido ealso qqstats educ nonparametric rep es ks
%  LocalWords:  inliers Mahalanobis

\name{balanceUV}
\alias{balanceUV}
\title{Univariate Balance Tests}
\description{
  This function provides a number of univariate balance metrics.
  Generally, users should call \code{\link{MatchBalance}} and not this function
  directly.
}
\usage{
balanceUV(Tr, Co, weights = rep(1, length(Co)), exact = FALSE, ks=FALSE,
          nboots = 1000, paired=TRUE, match=FALSE,
          weights.Tr=rep(1,length(Tr)), weights.Co=rep(1,length(Co)),
          estimand="ATT")
}
\arguments{
  \item{Tr}{A vector containing the treatment observations.}
  \item{Co}{A vector containing the control observations.}
  \item{weights}{A vector containing the observation specific
    weights. Only use this option when the treatment and control
    observations are paired (as they are after matching).}
  \item{exact}{A logical flag indicating if the exact Wilcoxon test
    should be used instead of the test with a correction.  See
    \code{\link{wilcox.test}} for details.}
  \item{ks}{ A logical flag for if the univariate bootstrap
    Kolmogorov-Smirnov (KS) test should be calculated.  If the ks option
    is set to true, the univariate KS test is calculated for all
    non-dichotomous variables.  The bootstrap KS test is consistent even
    for non-continuous variables.  See \code{\link{ks.boot}} for more
    details.}
  \item{nboots}{The number of bootstrap samples to be run for the
    \code{ks} test.  If zero, no bootstraps are done.  Bootstrapping is
    highly recommended because the bootstrapped Kolmogorov-Smirnov test
    only provides correct coverage even for non-continuous covariates. At
    least 500 \code{nboots} (preferably 1000) are
    recommended for publication quality p-values.}
  \item{paired}{A flag for if the paired \code{\link{t.test}} should be used.}
  \item{match}{A flag for if the \code{Tr} and \code{Co} objects are the result
    of a call to \code{\link{Match}}.}
  \item{weights.Tr}{A vector of weights for the treated observations.}
  \item{weights.Co}{A vector of weights for the control observations.}
  \item{estimand}{This determines if the standardized mean difference
    returned by the \code{sdiff} object is standardized by the variance of
    the treatment observations (which is done if the estimand is either
    "ATE" or "ATT") or by the variance of the control observations (which
    is done if the estimand is "ATC").}
}
\value{
  \item{sdiff}{This is the standardized difference between the treated
    and control units multiplied by 100.  That is, 100 times the mean
    difference  between treatment and control units divided by the standard
    deviation of the treatment
    observations alone if the estimand is either \code{ATT} or
    \code{ATE}.  The variance of the control observations are used if
    the estimand is \code{ATC}.}
  \item{sdiff.pooled}{This is the standardized difference between the
    treated and control units multiplied by 100 using the pooled
    variance.  That is, 100 times the mean
    difference between treatment and control units divided by the pooled standard deviation as in
    Rosenbaum and Rubin (1985).}
  \item{mean.Tr}{The mean of the treatment group.}
  \item{mean.Co}{The mean of the control group.}
  \item{var.Tr}{The variance of the treatment group.}
  \item{var.Co}{The variance of the control group.}
  \item{p.value}{The p-value from the two-sided weighted \code{\link{t.test}}.}
  \item{var.ratio}{var.Tr/var.Co.}
  \item{ks}{The object returned by \code{\link{ks.boot}}.}
  \item{tt}{The object returned by two-sided weighted
    \code{\link{t.test}}.}
  \item{qqsummary}{The return object from a call to
    \code{\link{qqstats}} with standardization---i.e., balance test
    based on the empirical CDF.}
  \item{qqsummary.raw}{The return object from a call to
    \code{\link{qqstats}} without standardization--i.e., balance tests
    based on the empirical QQ-plot which retain the scale of the
    variable.}
}
\author{
  Jasjeet S. Sekhon, UC Berkeley, \email{sekhon@berkeley.edu},
  \url{http://sekhon.berkeley.edu/}.
}
\references{
  Sekhon, Jasjeet S. 2011.  "Multivariate and Propensity Score
  Matching Software with Automated Balance Optimization.''
  \emph{Journal of Statistical Software} 42(7): 1-52.
  \doi{10.18637/jss.v042.i07}

  Diamond, Alexis and Jasjeet S. Sekhon. 2013. "Genetic
  Matching for Estimating Causal Effects: A General Multivariate
  Matching Method for Achieving Balance in Observational Studies.''
  \emph{Review of Economics and Statistics}.  95 (3): 932--945.
  \url{http://sekhon.berkeley.edu/papers/GenMatch.pdf}

  Rosenbaum, Paul R. and Donald B. Rubin. 1985. ``Constructing a Control
  Group Using Multivariate Matched Sampling Methods That Incorporate the
  Propensity Score.'' \emph{The American Statistician} 39:1 33-38.

  Hollander, Myles and Douglas A. Wolfe. 1973. \emph{Nonparametric
    statistical inference}. New York: John Wiley & Sons.
  }
  \seealso{ Also see \code{\link{summary.balanceUV}}, \code{\link{qqstats}}
    \code{\link{ks.boot}}, \code{\link{Match}}, \code{\link{GenMatch}},
    \code{\link{MatchBalance}},
    \code{\link{GerberGreenImai}}, \code{\link{lalonde}} }
\examples{
data(lalonde)
attach(lalonde)

foo  <- balanceUV(re75[treat==1],re75[treat!=1])
summary(foo)

}
\keyword{univar}

%  LocalWords:  balanceUV Univariate MatchBalance nboots Wilcoxon wilcox sdiff
%  LocalWords:  Kolmogorov tt UC url emph seealso GenMatch lalonde
%  LocalWords:  GerberGreenImai univar qqstats

\name{summary.Match}
\alias{summary.Match}
\alias{print.summary.Match}
\title{Summarizing output from Match}
\description{
  \code{\link{summary}} method for class \code{\link{Match}}
}
\usage{
  \method{summary}{Match}(object, ... , full=FALSE, digits=5)
}
\arguments{
  \item{object}{An object of class "\code{Match}", usually, a
    result of a call to \code{\link{Match}}.}
  \item{full}{A flag for whether the unadjusted estimates and naive
    standard errors should also be summarized.}
  \item{digits}{The number of significant digits that should be
    displayed.}
  \item{...}{Other options for the generic summary function.}
}
\author{
  Jasjeet S. Sekhon, UC Berkeley, \email{sekhon@berkeley.edu},
  \url{http://sekhon.berkeley.edu/}.  
}
\seealso{ Also see \code{\link{Match}}, \code{\link{GenMatch}},
  \code{\link{MatchBalance}}, 
  \code{\link{balanceUV}}, \code{\link{qqstats}}, \code{\link{ks.boot}},
  \code{\link{GerberGreenImai}}, \code{\link{lalonde}}
}
\keyword{htest}

%  LocalWords:  UC url seealso GenMatch MatchBalance balanceUV htest
%  LocalWords:  GerberGreenImai lalonde

\name{MatchBalance}
\alias{MatchBalance}
\title{Tests for Univariate and Multivariate Balance}
\description{
  This function provides a variety of balance statistics useful for
  determining if balance exists in any unmatched dataset and
  in matched datasets produced by the \code{\link{Match}}
  function.  Matching is performed by the \code{\link{Match}} function,
  and \code{MatchBalance} is used to determine if \code{\link{Match}}
  was successful in achieving balance on the observed covariates.
}
\usage{
MatchBalance(formul, data = NULL, match.out = NULL, ks = TRUE,
             nboots=500, weights=NULL, digits=5, paired=TRUE, print.level=1)
}
\arguments{
  \item{formul}{
    This formula does \emph{not} estimate any model.  The formula is
    simply an efficient way to use the R modeling language to list the
    variables we wish to obtain univariate balance statistics for.  The
    dependent variable in the formula is usually the treatment
    indicator.  One should include many functions of the observed
    covariates.  Generally, one should request balance statistics on
    more higher-order terms and interactions than were used to conduct
    the matching itself.}
  \item{data}{ A data frame which contains all of the variables in the
    formula.  If a data frame is not provided, the variables are
    obtained via lexical scoping.}
  \item{match.out}{ The output object from the \code{\link{Match}}
    function.  If this output is included, \code{\link{MatchBalance}} will provide
    balance statistics for both before and after matching. Otherwise
    balance statistics will only be reported for the raw unmatched
    data.}
  \item{ks}{ A logical flag for whether the univariate bootstrap
    Kolmogorov-Smirnov (KS) test should be calculated.  If the ks option
    is set to true, the univariate KS test is calculated for all
    non-dichotomous variables.  The bootstrap KS test is consistent even
    for non-continuous variables.  See \code{\link{ks.boot}} for more
    details.}
  \item{weights}{An optional vector of observation specific weights.}
  \item{nboots}{The number of bootstrap samples to be run.  If zero, no
    bootstraps are done.  Bootstrapping is highly recommended because
    the bootstrapped Kolmogorov-Smirnov test provides correct coverage
    even when the distributions being compared are not continuous.  At
    least 500 \code{nboots} (preferably 1000) are recommended for
    publication quality p-values.}
  \item{digits}{The number of significant digits that should be
    displayed.}
  \item{paired}{A flag for whether the paired \code{\link{t.test}} should be
    used after matching.  Regardless of the value of this option, an
    unpaired \code{\link{t.test}} is done for the unmatched data because
    it is assumed that the unmatched data were not generated by a paired
    experiment.}
  \item{print.level}{The amount of printing to be done.  If zero, there is
    no printing.  If one, the results are summarized.  If
    two, details of the computations are printed.}
}
\value{
  \item{BeforeMatching}{A list containing the before matching univariate
    balance statistics. That is, a list containing the results of
    the  \code{\link{balanceUV}} function applied to all of the 
    covariates described in \code{formul}.  Note that the univariate
    test results for all of the variables in \code{formul} are printed
    if \code{verbose > 0}.}
  \item{AfterMatching}{A list containing the after matching univariate
    balance statistics. That is, a list containing the results of
    the  \code{\link{balanceUV}} function applied to all of the 
    covariates described in \code{formul}.  Note that the univariate
    test results for all of the variables in \code{formul} are printed
    if \code{verbose > 0}.  This object is \code{NULL}, if no matched
    dataset was provided.}
  \item{BMsmallest.p.value}{The smallest p.value found across all of the
    \emph{before} matching balance tests (including t-tests and KS-tests.}
  \item{BMsmallestVarName}{The name of the variable with the
    \code{BMsmallest.p.value} (a vector in case of ties).}
  \item{BMsmallestVarNumber}{The number of the variable with the
    \code{BMsmallest.p.value} (a vector in case of ties).}
  \item{AMsmallest.p.value}{The smallest p.value found across all of the
    \emph{after} matching balance tests (including t-tests and
    KS-tests.}
  \item{AMsmallestVarName}{The name of the variable with the
    \code{AMsmallest.p.value} (a vector in case of ties).}
  \item{AMsmallestVarNumber}{The number of the variable with the
    \code{AMsmallest.p.value} (a vector in case of ties).}  
}
\details{
  This function can be used to determine if there is balance in the pre-
  and/or post-matching datasets.  Difference of means between treatment
  and control groups are provided as well as a variety of summary
  statistics for the empirical CDF (eCDF) and empirical-QQ (eQQ) plot
  between the two groups.  The eCDF results are the standardized mean,
  median and maximum differences in the empirical CDF. The eQQ results
  are summaries of the raw differences in the empirical-QQ plot.\cr

  Two univariate tests are also provided: the t-test and the bootstrap
  Kolmogorov-Smirnov (KS) test. These tests should not be treated as
  hypothesis tests in the usual fashion because we wish to maximize
  balance without limit.  The bootstrap KS test is highly
  recommended (see the \code{ks} and \code{nboots} options) because the
  bootstrap KS is consistent even for non-continuous distributions.
  Before matching, the two sample t-test is used; after matching, the
  paired t-test is used.\cr

  Two multivariate tests are provided.  The KS and Chi-Square null
  deviance tests.  The KS test is to be preferred over the Chi-Square
  test because the Chi-Square test is not testing the relevant
  hypothesis.  The null hypothesis for the KS test is equal balance
  in the estimated probabilities between treated and control.  The null
  hypothesis for the Chi-Square test, however, is all of the
  parameters being insignificant; a comparison of residual versus null
  deviance.  If the covariates being considered are discrete, this KS
  test is asymptotically nonparametric as long as the logit model does
  not produce zero parameter estimates.

  \code{NA}'s are handled by the \code{\link{na.action}} option.  But it
  is highly recommended that \code{NA}'s not simply be deleted, but
  one should check to make sure that missingness is balanced.
  }
  \references{
  Sekhon, Jasjeet S. 2011.  "Multivariate and Propensity Score
  Matching Software with Automated Balance Optimization.''
  \emph{Journal of Statistical Software} 42(7): 1-52.
  \url{http://www.jstatsoft.org/v42/i07/}

  Diamond, Alexis and Jasjeet S. Sekhon. 2013. "Genetic
  Matching for Estimating Causal Effects: A General Multivariate
  Matching Method for Achieving Balance in Observational Studies.''
  \emph{Review of Economics and Statistics}.  95 (3): 932--945.
  \url{http://sekhon.berkeley.edu/papers/GenMatch.pdf}

  Abadie, Alberto.  2002.  ``Bootstrap Tests for Distributional Treatment
  Effects in Instrumental Variable Models.'' \emph{Journal of the
  American Statistical Association}, 97:457 (March) 284-292.  

  Hall, Peter. 1992. \emph{The Bootstrap and Edgeworth Expansion}.  New
  York: Springer-Verlag.

  Wilcox, Rand R. 1997. \emph{Introduction to Robust Estimation}.  San
  Diego, CA: Academic Press.

  William J. Conover (1971), \emph{Practical nonparametric statistics}.
  New York: John Wiley & Sons. Pages 295-301 (one-sample
  "Kolmogorov" test), 309-314 (two-sample "Smirnov" test).

  Shao, Jun and Dongsheng Tu. 1995. \emph{The Jackknife and Bootstrap}.
  New York: Springer-Verlag.
}
\author{
  Jasjeet S. Sekhon, UC Berkeley, \email{sekhon@berkeley.edu},
  \url{http://sekhon.berkeley.edu/}.  
}
\seealso{
  Also see \code{\link{Match}}, \code{\link{GenMatch}}, 
  \code{\link{balanceUV}}, \code{\link{qqstats}}, \code{\link{ks.boot}}, 
  \code{\link{GerberGreenImai}}, \code{\link{lalonde}}
}
\examples{
#
# Replication of Dehejia and Wahba psid3 model
#
# Dehejia, Rajeev and Sadek Wahba. 1999.``Causal Effects in
# Non-Experimental Studies: Re-Evaluating the Evaluation of Training
# Programs.''Journal of the American Statistical Association 94 (448):
# 1053-1062.

data(lalonde)

#
# Estimate the propensity model
#
glm1  <- glm(treat~age + I(age^2) + educ + I(educ^2) + black +
             hisp + married + nodegr + re74  + I(re74^2) + re75 + I(re75^2) +
             u74 + u75, family=binomial, data=lalonde)


#
#save data objects
#
X  <- glm1$fitted
Y  <- lalonde$re78
Tr  <- lalonde$treat

#
# one-to-one matching with replacement (the "M=1" option).
# Estimating the treatment effect on the treated (the "estimand" option which defaults to 0).
#
rr  <- Match(Y=Y,Tr=Tr,X=X,M=1);

#Let's summarize the output
summary(rr)

# Let's check the covariate balance
# 'nboots' is set to small values in the interest of speed.
# Please increase to at least 500 each for publication quality p-values.  
mb  <- MatchBalance(treat~age + I(age^2) + educ + I(educ^2) + black +
                    hisp + married + nodegr + re74  + I(re74^2) + re75 + I(re75^2) +
                    u74 + u75, data=lalonde, match.out=rr, nboots=10)
}
\keyword{nonparametric}
\keyword{htest}

%  LocalWords:  MatchBalance formul nboots nrow regressors glm uv
%  LocalWords:  datasets Kolmogorov balanceUV logit url Abadie emph
%  LocalWords:  Edgeworth Verlag Conover Shao Dongsheng UC seealso GenMatch rr
%  LocalWords:  GerberGreenImai lalonde Dehejia Wahba psid Rajeev Sadek hisp mb
%  LocalWords:  nodegr estimand htest QQ eQQ cr Monte Carlo Smirnov ks rep Chi
%  LocalWords:  pre nonparametric San CA Jun Tu ealso qqstats educ Tr

\name{GerberGreenImai}
\alias{GerberGreenImai}
\docType{data}
\title{Gerber and Green Dataset used by Imai}
\description{

  This is the dataset used by Imai (2005) to replicate and evaluate the
  field experiment done by Gerber and Green (2000). The accompanying
  demo replicates Imai's propensity score model which is then used to
  estimate the causal effect of get-out-the-vote telephone calls on
  turnout. 
  
}
\usage{data(GerberGreenImai)}
\format{
  A data frame with 10829 observations on the following 26 variables.
  \describe{
    \item{PERSONS }{Number persons in household}
    \item{WARD    }{Ward of residence}
    \item{QUESTION}{Asked to commit to voting}
    \item{MAILGRP }{Sent mail}
    \item{PHONEGRP}{Phone batch \#1}
    \item{PERSNGRP}{Personal contact attempted}
    \item{APPEAL  }{Content of message}
    \item{CONTACT }{Personal contact occurred}
    \item{MAILINGS}{Number of mailings sent}
    \item{AGE     }{Age of respondent}
    \item{MAJORPTY}{Democratic or Republican}
    \item{VOTE96.0}{Abstained in 1996}
    \item{VOTE96.1}{Voted in 1996}
    \item{MAILCALL}{Phone batch \#2}
    \item{VOTED98 }{Voted in 1998}
    \item{PHNSCRPT}{Script read to phone respondents}
    \item{DIS.MC  }{Contacted by phone in batch \#2}
    \item{DIS.PHN }{Contacted by phone in batch \#1}
    \item{PHN.C   }{Contacted by phone}
    \item{PHNTRT1 }{Phone contact attempted (no blood or blood/civic)}
    \item{PHNTRT2 }{Phone contact attempted (no blood)}
    \item{PHN.C1  }{Contact occurred in phntrt1}
    \item{PHN.C2  }{Contact occurred in phntrt2}
    \item{NEW     }{New voter}
    \item{phone   }{Contacted by phone}
    \item{AGE2    }{Age squared}
  }
}
\details{

  The demo provided, entitled \code{GerberGreenImai}, uses Imai's
  propensity score model to estimate the causal effect of
  get-out-the-vote telephone calls on turnout. The propensity score
  model fails to balance age.

}
\references{

  Gerber, Alan S. and Donald P. Green. 2000. ``The Effects of
  Canvassing, Telephone Calls, and Direct Mail on Voter Turnout: A Field
  Experiment.'' \emph{American Political Science Review} 94: 653-663.

  Gerber, Alan S. and Donald P. Green. 2005. ``Correction to Gerber
  and Green (2000), replication of disputed findings, and reply to
  Imai (2005).''  \emph{American Political Science Review} 99:
  301-313.

  Imai, Kosuke. 2005. ``Do Get-Out-The-Vote Calls Reduce Turnout?
  The Importance of Statistical Methods for Field Experiments.'' 
  \emph{American Political Science Review} 99: 283-300.

  Hansen, Ben B. Hansen and Jake Bowers. forthcoming. ``Attributing
  Effects to a Cluster Randomized Get-Out-The-Vote Campaign.''
  \emph{Journal of the American Statistical Association}.
} 
\seealso{ Also see \code{\link{Match}} and \code{\link{MatchBalance}},
  \code{\link{GenMatch}}, \code{\link{balanceUV}}, \code{\link{ks.boot}}
  \code{\link{lalonde}}
}
\keyword{datasets}


%  LocalWords:  GerberGreenImai docType Dataset dataset Imai's MAILGRP PHONEGRP
%  LocalWords:  PERSNGRP MAJORPTY MAILCALL PHNSCRPT PHN PHNTRT phntrt emph
%  LocalWords:  Kosuke seealso MatchBalance GenMatch balanceUV
%  LocalWords:  lalonde datasets

\name{ks.boot}
\alias{ks.boot}
\title{Bootstrap Kolmogorov-Smirnov}
\description{
  This function executes a bootstrap version of the univariate
  Kolmogorov-Smirnov test which provides correct coverage even when
  the distributions being compared are not entirely continuous.  Ties
  are allowed with this test unlike the traditional Kolmogorov-Smirnov
  test.
}
\usage{
ks.boot(Tr, Co, nboots=1000, alternative = c("two.sided","less","greater"),
        print.level=0)
}
\arguments{
  \item{Tr}{A vector containing the treatment observations.}
  \item{Co}{A vector containing the control observations.}
  \item{nboots}{The number of bootstraps to be performed.  These are, in
    fact, really Monte Carlo simulations which are preformed in order to
    determine the proper p-value from the empiric.}
  \item{alternative}{indicates the alternative hypothesis and must be one of
    '"two.sided"' (default), '"less"', or '"greater"'.  You can
    specify just the initial letter.  See \code{\link{ks.test}} for details.}
  \item{print.level}{If this is greater than 1, then the simulation count is
    printed out while the simulations are being done.}
}
\value{
  \item{ks.boot.pvalue}{The bootstrap p-value of the Kolmogorov-Smirnov test
    for the hypothesis that the probability densities for both the
    treated and control groups are the same.}
  \item{ks}{Return object from \code{\link{ks.test}}.}
  \item{nboots}{The number of bootstraps which were completed.}
}
\author{
  Jasjeet S. Sekhon, UC Berkeley, \email{sekhon@berkeley.edu},
  \url{http://sekhon.berkeley.edu/}.
}
\references{
  Sekhon, Jasjeet S. 2011.  "Multivariate and Propensity Score
  Matching Software with Automated Balance Optimization.''
  \emph{Journal of Statistical Software} 42(7): 1-52.
  \doi{10.18637/jss.v042.i07}

  Diamond, Alexis and Jasjeet S. Sekhon. 2013. "Genetic Matching for
  Estimating Causal Effects: A General Multivariate Matching Method for
  Achieving Balance in Observational Studies.'' \emph{Review of
  Economics and Statistics}.  95 (3): 932--945.
  \url{http://sekhon.berkeley.edu/papers/GenMatch.pdf}

  Abadie, Alberto.  2002.  ``Bootstrap Tests for Distributional Treatment
  Effects in Instrumental Variable Models.'' \emph{Journal of the
  American Statistical Association}, 97:457 (March) 284-292.
  }
  \seealso{ Also see \code{\link{summary.ks.boot}},
    \code{\link{qqstats}}, \code{\link{balanceUV}}, \code{\link{Match}},
    \code{\link{GenMatch}}, \code{\link{MatchBalance}},
    \code{\link{GerberGreenImai}}, \code{\link{lalonde}} }
\examples{
#
# Replication of Dehejia and Wahba psid3 model
#
# Dehejia, Rajeev and Sadek Wahba. 1999.``Causal Effects in
# Non-Experimental Studies: Re-Evaluating the Evaluation of Training
# Programs.''Journal of the American Statistical Association 94 (448):
# 1053-1062.
#
data(lalonde)

#
# Estimate the propensity model
#
glm1  <- glm(treat~age + I(age^2) + educ + I(educ^2) + black +
             hisp + married + nodegr + re74  + I(re74^2) + re75 + I(re75^2) +
             u74 + u75, family=binomial, data=lalonde)


#
#save data objects
#
X  <- glm1$fitted
Y  <- lalonde$re78
Tr  <- lalonde$treat

#
# one-to-one matching with replacement (the "M=1" option).
# Estimating the treatment effect on the treated (the "estimand" option which defaults to 0).
#
rr  <- Match(Y=Y,Tr=Tr,X=X,M=1);
summary(rr)

#
# Do we have balance on 1975 income after matching?
#
ks  <- ks.boot(lalonde$re75[rr$index.treated], lalonde$re75[rr$index.control], nboots=500)
summary(ks)
}
\keyword{htest}

%  LocalWords:  Kolmogorov nboots ksboot pvalue UC url Abadie emph seealso psid
%  LocalWords:  balanceUV GenMatch MatchBalance GerberGreenImai Wahba
%  LocalWords:  lalonde Dehejia Rajeev Sadek glm hisp nodegr estimand rr htest

\name{Matchby}
\alias{Matchby}
\title{Grouped Multivariate and Propensity Score Matching}
\description{
  This function is a wrapper for the \code{\link{Match}} function which
  separates the matching problem into subgroups defined by a factor.
  This is equivalent to conducting exact matching on each level of a factor.
  Matches within each level are found as determined by the
  usual matching options.  This function is much faster for large
  datasets than the \code{\link{Match}} function itself.  For additional
  speed, consider doing matching without replacement---see the
  \code{replace} option.  This function is more limited than the
  \code{\link{Match}} function.  For example, \code{Matchby} cannot be
  used if the user wishes to provide observation specific weights.
}

\usage{
Matchby(Y, Tr, X, by, estimand = "ATT", M = 1, ties=FALSE, replace=TRUE,
        exact = NULL, caliper = NULL, AI=FALSE, Var.calc=0,
        Weight = 1, Weight.matrix = NULL, distance.tolerance = 1e-05,
        tolerance = sqrt(.Machine$double.eps), print.level=1, version="Matchby", ...)
}
\arguments{
  \item{Y}{ A vector containing the outcome of interest. Missing values are not allowed.}
  \item{Tr}{ A vector indicating the observations which are
    in the treatment regime and those which are not.  This can either be a
    logical vector or a real vector where 0 denotes control and 1 denotes
    treatment.}
  \item{X}{ A matrix containing the variables we wish to match on.
    This matrix may contain the actual observed covariates or the
    propensity score or a combination of both.}
  \item{by}{A "factor" in the sense that \code{as.factor(by)} defines the
    grouping, or a list of such factors in which case their
    interaction is used for the grouping.}
  \item{estimand}{ A character string for the estimand.  The default
    estimand is "ATT", the sample average treatment effect for the
    treated. "ATE" is the sample average treatment effect (for all), and
    "ATC" is the sample average treatment effect for the controls.}
  \item{M}{ A scalar for the number of matches which should be
    found. The default is one-to-one matching. Also see the
    \code{ties} option.}
  \item{ties}{A logical flag for whether ties should be handled
    deterministically.  By default \code{ties==TRUE}. If, for example, one
    treated observation matches more than one control observation, the
    matched dataset will include the multiple matched control observations
    and the matched data will be weighted to reflect the multiple matches.
    The sum of the weighted observations will still equal the original
    number of observations. If \code{ties==FALSE}, ties will be randomly
    broken.  \emph{If the dataset is large and there are many ties,
      setting \code{ties=FALSE} often results in a large speedup.} Whether
    two potential matches are close enough to be considered tied, is
    controlled by the \code{distance.tolerance} option.}
  \item{replace}{Whether matching should be done with replacement.  Note
    that if \code{FALSE}, the order of matches generally matters.  Matches
    will be found in the same order as the data is sorted.  Thus, the
    match(es) for the first observation will be found first and then for
    the second etc. Matching without replacement will generally increase
    bias so it is not recommended.  \emph{But if the dataset is large and
      there are many potential matches, setting \code{replace=false} often
      results in a large speedup and negligible or no bias.} Ties are
    randomly broken when \code{replace==FALSE}---see the \code{ties}
    option for details.}
  \item{exact}{ A logical scalar or vector for whether exact matching
    should be done.  If a logical scalar is provided, that logical value is
    applied to all covariates of
    \code{X}.  If a logical vector is provided, a logical value should
    be provided for each covariate in \code{X}. Using a logical vector
    allows the user to specify exact matching for some but not other
    variables.  When exact matches are not found, observations are
    dropped.  \code{distance.tolerance} determines what is considered to be an
    exact match. The \code{exact} option takes precedence over the
    \code{caliper} option.}
  \item{caliper}{ A scalar or vector denoting the caliper(s) which
    should be used when matching.  A caliper is the distance which is
    acceptable for any match.  Observations which are outside of the
    caliper are dropped. If a scalar caliper is provided, this caliper is
    used for all covariates in \code{X}.  If a vector of calipers is
    provided, a caliper value should be provide for each covariate in
    \code{X}. The caliper is interpreted to be in standardized units.  For
    example, \code{caliper=.25} means that all matches not equal to or
    within .25 standard deviations of each covariate in \code{X} are
    dropped.}
  \item{AI}{A logical flag for if the Abadie-Imbens standard error
    should be calculated. It is computationally expensive to calculate
    with large datasets. \code{Matchby} can only calculate AI SEs for ATT.
    To calculate AI errors with other estimands, please use the
    \code{\link{Match}} function.  See the \code{Var.calc} option if one
    does not want to assume homoscedasticity.}
  \item{Var.calc}{A scalar for the variance estimate
    that should be used.  By default \code{Var.calc=0} which means that
    homoscedasticity is assumed.  For values of  \code{Var.calc > 0},
    robust variances are calculated using \code{Var.calc} matches.}
  \item{Weight}{ A scalar for the type of
    weighting scheme the matching algorithm should use when weighting
    each of the covariates in \code{X}.  The default value of
    1 denotes that weights are equal to the inverse of the variances. 2
    denotes the Mahalanobis distance metric, and 3 denotes
    that the user will supply a weight matrix (\code{Weight.matrix}).  Note that
    if the user supplies a \code{Weight.matrix}, \code{Weight} will be automatically
    set to be equal to 3.}
  \item{Weight.matrix}{ This matrix denotes the weights the matching
    algorithm uses when weighting each of the covariates in \code{X}---see
    the \code{Weight} option. This square matrix should have as many
    columns as the number of columns of the \code{X} matrix. This matrix
    is usually provided by a call to the \code{\link{GenMatch}} function
    which finds the optimal weight each variable should be given so as to
    achieve balance on the covariates. \cr

    For most uses, this matrix has zeros in the off-diagonal
    cells.  This matrix can be used to weight some variables more than
    others.  For
    example, if \code{X} contains three variables and we want to
    match as best as we can on the first, the following would work well:
    \cr
    \code{> Weight.matrix <- diag(3)}\cr
    \code{> Weight.matrix[1,1] <- 1000/var(X[,1])} \cr
    \code{> Weight.matrix[2,2] <- 1/var(X[,2])} \cr
    \code{> Weight.matrix[3,3] <- 1/var(X[,3])} \cr
    This code changes the weights implied by the
    inverse of the variances by multiplying the first variable by a 1000
    so that it is highly weighted.  In order to enforce exact matching
    see the \code{exact} and \code{caliper} options.}
  \item{distance.tolerance}{This is a scalar which is used to determine if distances
    between two observations are different from zero.  Values less than
    \code{distance.tolerance} are deemed to be equal to zero.  This
    option can be used to perform a type of optimal matching}
  \item{tolerance}{ This is a scalar which is used to determine
    numerical tolerances.  This option is used by numerical routines
    such as those used to determine if a matrix is singular.}
  \item{print.level}{The level of printing. Set to '0' to turn off printing.}
  \item{version}{The version of the code to be used.  The "Matchby" C/C++
    version of the code is the fastest, and the end-user should not
    change this option.}
  \item{...}{Additional arguments passed on to \code{\link{Match}}.}
}
\details{
  \code{Matchby} is much faster for large datasets than
  \code{\link{Match}}.  But \code{Matchby} only implements a subset of
  the functionality of \code{\link{Match}}.  For example, the
  \code{restrict} option cannot be used, Abadie-Imbens standard errors
  are not provided and bias adjustment cannot be requested.
  \code{Matchby} is a wrapper for the \code{\link{Match}} function which
  separates the matching problem into subgroups defined by a factor.  This
  is the equivalent to doing exact matching on each factor, and the
  way in which matches are found within each factor is determined by the
  usual matching options. \cr

  \emph{Note that by default \code{ties=FALSE} although the default for
    the \code{Match} in \code{GenMatch} functions is \code{TRUE}.  This is
    done because randomly breaking ties in large datasets often results in
    a great speedup.}  For additional speed, consider doing matching
  without replacement which is often much faster when the dataset is
  large---see the \code{replace} option. \cr

  There will be slight differences in the matches produced by
  \code{Matchby} and \code{\link{Match}} because of how the covariates
  are weighted.  When the data is broken up into separate groups (via
  the \code{by} option), Mahalanobis distance and inverse variance
  will imply different weights than when the data is taken as whole.
}

\value{
  \item{est}{The estimated average causal effect.}
  \item{se.standard }{The usual standard error.  This is the standard error
    calculated on the matched data using the usual method of calculating
    the difference of means (between treated and control) weighted so
    that ties are taken into account.}
  \item{se }{The Abadie-Imbens standard error. This is only calculated
    if the \code{AI} option is \code{TRUE}. This standard error has
    correct coverage if \code{X} consists of either covariates or a
    known propensity score because it takes into account the uncertainty
    of the matching
    procedure.  If an estimated propensity score is used, the
    uncertainty involved in its estimation is not accounted for although the
    uncertainty of the matching procedure itself still is.}
  \item{index.treated }{A vector containing the observation numbers from
    the original dataset for the treated observations in the
    matched dataset.  This index in conjunction with \code{index.control}
    can be used to recover the matched dataset produced by
    \code{Matchby}.  For example, the \code{X} matrix used by \code{Matchby}
    can be recovered by
    \code{rbind(X[index.treated,],X[index.control,])}.}
  \item{index.control }{A vector containing the observation numbers from
    the original data for the control observations in the
    matched data.  This index in conjunction with \code{index.treated}
    can be used to recover the matched dataset produced by
    \code{Matchby}.  For example, the \code{Y} matrix for the matched dataset
    can be recovered by
    \code{c(Y[index.treated],Y[index.control])}.}
  \item{weights}{The weights for each observation in the matched
    dataset.}
  \item{orig.nobs }{The original number of observations in the dataset.}
  \item{nobs }{The number of observations in the matched dataset.}
  \item{wnobs }{The number of weighted observations in the matched dataset.}
  \item{orig.treated.nobs}{The original number of treated observations.}
  \item{ndrops}{The number of matches which were dropped because there
    were not enough observations in a given group and because of
    caliper and exact matching.}
  \item{estimand}{The estimand which was estimated.}
  \item{version}{The version of \code{\link{Match}} which was used.}
}
\references{
  Sekhon, Jasjeet S. 2011.  "Multivariate and Propensity Score
  Matching Software with Automated Balance Optimization.''
  \emph{Journal of Statistical Software} 42(7): 1-52.
  \doi{10.18637/jss.v042.i07}

  Diamond, Alexis and Jasjeet S. Sekhon. 2013. "Genetic
  Matching for Estimating Causal Effects: A General Multivariate
  Matching Method for Achieving Balance in Observational Studies.''
  \emph{Review of Economics and Statistics}.  95 (3): 932--945.
  \url{http://sekhon.berkeley.edu/papers/GenMatch.pdf}

  Abadie, Alberto and Guido Imbens. 2006.
  ``Large Sample Properties of Matching Estimators for Average
  Treatment Effects.'' \emph{Econometrica} 74(1): 235-267.

  Imbens, Guido. 2004. Matching Software for Matlab and
  Stata.

}
\author{Jasjeet S. Sekhon, UC Berkeley, \email{sekhon@berkeley.edu},
  \url{http://sekhon.berkeley.edu/}.
}
\seealso{ Also see \code{\link{Match}},
    \code{\link{summary.Matchby}},
    \code{\link{GenMatch}},
    \code{\link{MatchBalance}},
    \code{\link{balanceUV}},
    \code{\link{qqstats}}, \code{\link{ks.boot}},
    \code{\link{GerberGreenImai}}, \code{\link{lalonde}}
}
\examples{
#
# Match exactly by racial groups and then match using the propensity score within racial groups
#

data(lalonde)

#
# Estimate the Propensity Score
#
glm1  <- glm(treat~age + I(age^2) + educ + I(educ^2) +
             hisp + married + nodegr + re74  + I(re74^2) + re75 + I(re75^2) +
             u74 + u75, family=binomial, data=lalonde)


#save data objects
#
X  <- glm1$fitted
Y  <- lalonde$re78
Tr <- lalonde$treat

# one-to-one matching with replacement (the "M=1" option) after exactly
# matching on race using the 'by' option.  Estimating the treatment
# effect on the treated (the "estimand" option defaults to ATT).
rr  <- Matchby(Y=Y, Tr=Tr, X=X, by=lalonde$black, M=1);
summary(rr)

# Let's check the covariate balance
# 'nboots' is set to small values in the interest of speed.
# Please increase to at least 500 each for publication quality p-values.
mb  <- MatchBalance(treat~age + I(age^2) + educ + I(educ^2) + black +
                    hisp + married + nodegr + re74  + I(re74^2) + re75 + I(re75^2) +
                    u74 + u75, data=lalonde, match.out=rr, nboots=10)

}
\keyword{nonparametric}


%  LocalWords:  MatchBalance GenMatch emph estimand ATT BiasAdjust calc dataset
%  LocalWords:  ATC ecaliper cr diag homoscedasticity rbind GerberGreenImai se
%  LocalWords:  DehejiaWahba AbadieImbens noadj cond mdata datasets wnobs url
%  LocalWords:  ndrops Abadie Imbens Econometrica Matlab Stata UC seealso Wahba
%  LocalWords:  balanceUV lalonde Dehejia psid Rajeev Sadek glm hisp
%  LocalWords:  nodegr rr nboots nmc mb Matchby ret qqstats Mahalanobis SEs
%  LocalWords:  estimands
